@@ -5,8 +5,8 @@
 
 #ifndef FUNCTIONS_BOTH_LIBS_H
 #define FUNCTIONS_BOTH_LIBS_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
 void both_libs_build_impl_tbl(void);
-extern struct func_impl_name impl_tbl_both_libs[];
+extern struct func_impl impl_tbl_both_libs[];
 #endif
