@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_CONFIGURATION_DATA_H
 #define MUON_FUNCTIONS_CONFIGURATION_DATA_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_configuration_data[];
+extern const struct func_impl impl_tbl_configuration_data[];
 #endif
