@@ -5,7 +5,8 @@
 
 #ifndef MUON_FUNCTIONS_DICT_H
 #define MUON_FUNCTIONS_DICT_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_dict[];
+extern const struct func_impl impl_tbl_dict[];
+extern const struct func_impl impl_tbl_dict_internal[];
 #endif
