@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_FEATURE_OPT_H
 #define MUON_FUNCTIONS_FEATURE_OPT_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_feature_opt[];
+extern const struct func_impl impl_tbl_feature_opt[];
 #endif
