@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_KERNEL_CUSTOM_TARGET_H
 #define MUON_FUNCTIONS_KERNEL_CUSTOM_TARGET_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
 struct make_custom_target_opts {
 	obj name;
@@ -23,8 +23,7 @@ struct make_custom_target_opts {
 	bool extra_args_valid, extra_args_used;
 };
 
-bool make_custom_target(struct workspace *wk,
-	struct make_custom_target_opts *opts, obj *res);
+bool make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, obj *res);
 
 struct process_custom_target_commandline_opts {
 	uint32_t err_node;
@@ -40,9 +39,11 @@ struct process_custom_target_commandline_opts {
 };
 
 bool process_custom_target_commandline(struct workspace *wk,
-	struct process_custom_target_commandline_opts *opts, obj arr, obj *res);
+	struct process_custom_target_commandline_opts *opts,
+	obj arr,
+	obj *res);
 
-bool func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res);
+bool func_custom_target(struct workspace *wk, obj _, obj *res);
 
-bool func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res);
+bool func_vcs_tag(struct workspace *wk, obj _, obj *res);
 #endif
