@@ -5,7 +5,7 @@
 
 #ifndef MUON_FUNCTIONS_KERNEL_DEPENDENCY_H
 #define MUON_FUNCTIONS_KERNEL_DEPENDENCY_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
 void dep_process_deps(struct workspace *wk, obj deps, struct build_dep *dest);
 bool dep_process_link_with(struct workspace *wk, uint32_t err_node, obj arr, struct build_dep *dest);
@@ -14,6 +14,6 @@ void dep_process_includes(struct workspace *wk, obj arr, enum include_type inclu
 
 void build_dep_init(struct workspace *wk, struct build_dep *dep);
 
-bool func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res);
-bool func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *res);
+bool func_dependency(struct workspace *wk, obj self, obj *res);
+bool func_declare_dependency(struct workspace *wk, obj _, obj *res);
 #endif
