@@ -5,8 +5,8 @@
 
 #ifndef MUON_FUNCTIONS_MODULES_FS_H
 #define MUON_FUNCTIONS_MODULES_FS_H
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 
-extern const struct func_impl_name impl_tbl_module_fs[];
-extern const struct func_impl_name impl_tbl_module_fs_internal[];
+extern const struct func_impl impl_tbl_module_fs[];
+extern const struct func_impl impl_tbl_module_fs_internal[];
 #endif
