@@ -13,21 +13,22 @@
 #include <stdio.h>
 
 #include "compilers.h"
-#include "data/bucket_array.h"
-#include "iterator.h"
+#include "datastructures/bucket_arr.h"
+#include "datastructures/iterator.h"
 #include "lang/types.h"
+#include "machines.h"
 
 enum obj_type {
 	/* singleton object types */
 	obj_null,
 	obj_meson,
 	obj_disabler,
-	obj_machine, // this won't be a singleton object when cross compilaton is implemented
 
 	/* simple object types */
 	obj_bool,
 	obj_file,
 	obj_feature_opt,
+	obj_machine,
 
 	/* complex object types */
 	_obj_aos_start,
@@ -41,6 +42,7 @@ enum obj_type {
 	obj_subproject,
 	obj_dependency,
 	obj_external_program,
+	obj_python_installation,
 	obj_run_result,
 	obj_configuration_data,
 	obj_test,
@@ -55,95 +57,51 @@ enum obj_type {
 	obj_both_libs,
 	obj_source_set,
 	obj_source_configuration,
+	obj_iterator,
+
+	/* muon-specific objects */
+	obj_func,
+	obj_capture,
 	obj_typeinfo,
 
 	obj_type_count,
 };
 
-typedef uint64_t type_tag;
-
-#define ARG_TYPE_NULL (obj_type_count + 1)
-
-#define ARG_TYPE_GLOB             (((type_tag)1) << 61)
-#define ARG_TYPE_ARRAY_OF         (((type_tag)1) << 62)
-#define obj_typechecking_type_tag (((type_tag)1) << 63)
-
-#define tc_meson              (obj_typechecking_type_tag | (((type_tag)1) << 0))
-#define tc_disabler           (obj_typechecking_type_tag | (((type_tag)1) << 1))
-#define tc_machine            (obj_typechecking_type_tag | (((type_tag)1) << 2))
-#define tc_bool               (obj_typechecking_type_tag | (((type_tag)1) << 3))
-#define tc_file               (obj_typechecking_type_tag | (((type_tag)1) << 4))
-#define tc_feature_opt        (obj_typechecking_type_tag | (((type_tag)1) << 5))
-#define tc_number             (obj_typechecking_type_tag | (((type_tag)1) << 6))
-#define tc_string             (obj_typechecking_type_tag | (((type_tag)1) << 7))
-#define tc_array              (obj_typechecking_type_tag | (((type_tag)1) << 8))
-#define tc_dict               (obj_typechecking_type_tag | (((type_tag)1) << 9))
-#define tc_compiler           (obj_typechecking_type_tag | (((type_tag)1) << 10))
-#define tc_build_target       (obj_typechecking_type_tag | (((type_tag)1) << 11))
-#define tc_custom_target      (obj_typechecking_type_tag | (((type_tag)1) << 12))
-#define tc_subproject         (obj_typechecking_type_tag | (((type_tag)1) << 13))
-#define tc_dependency         (obj_typechecking_type_tag | (((type_tag)1) << 14))
-#define tc_external_program   (obj_typechecking_type_tag | (((type_tag)1) << 15))
-#define tc_run_result         (obj_typechecking_type_tag | (((type_tag)1) << 16))
-#define tc_configuration_data (obj_typechecking_type_tag | (((type_tag)1) << 17))
-#define tc_test               (obj_typechecking_type_tag | (((type_tag)1) << 18))
-#define tc_module             (obj_typechecking_type_tag | (((type_tag)1) << 19))
-#define tc_install_target     (obj_typechecking_type_tag | (((type_tag)1) << 20))
-#define tc_environment        (obj_typechecking_type_tag | (((type_tag)1) << 21))
-#define tc_include_directory  (obj_typechecking_type_tag | (((type_tag)1) << 22))
-#define tc_option             (obj_typechecking_type_tag | (((type_tag)1) << 23))
-#define tc_generator          (obj_typechecking_type_tag | (((type_tag)1) << 24))
-#define tc_generated_list     (obj_typechecking_type_tag | (((type_tag)1) << 25))
-#define tc_alias_target       (obj_typechecking_type_tag | (((type_tag)1) << 26))
-#define tc_both_libs          (obj_typechecking_type_tag | (((type_tag)1) << 27))
-#define tc_source_set         (obj_typechecking_type_tag | (((type_tag)1) << 28))
-#define tc_source_configuration (obj_typechecking_type_tag | (((type_tag)1) << 29))
-#define tc_type_count         30
-
-#define tc_any                (tc_bool | tc_file | tc_number | tc_string | tc_array | tc_dict \
-			       | tc_compiler | tc_build_target | tc_custom_target \
-			       | tc_subproject | tc_dependency | tc_feature_opt \
-			       | tc_external_program | tc_run_result \
-			       | tc_configuration_data | tc_test | tc_module \
-			       | tc_install_target | tc_environment | tc_include_directory \
-			       | tc_option | tc_generator | tc_generated_list \
-			       | tc_alias_target | tc_both_libs | tc_disabler \
-			       | tc_meson | tc_machine | tc_source_set | tc_source_configuration)
-
-#define tc_exe                (tc_string | tc_file | tc_external_program | tc_build_target | tc_custom_target | tc_both_libs)
-
-#define tc_coercible_env      (tc_environment | tc_string | tc_array | tc_dict)
-#define tc_coercible_files    (tc_string | tc_custom_target | tc_build_target | tc_file | tc_both_libs)
-#define tc_coercible_inc      (tc_string | tc_include_directory)
-#define tc_command_array      (ARG_TYPE_ARRAY_OF | tc_exe)
-#define tc_depends_kw         (ARG_TYPE_ARRAY_OF | tc_build_target | tc_custom_target | tc_both_libs)
-#define tc_install_mode_kw    (ARG_TYPE_ARRAY_OF | tc_string | tc_number | tc_bool)
-#define tc_required_kw        (tc_bool | tc_feature_opt)
-/* XXX: tc_file should not really be in tc_link_with_kw, however this is
- * how muon represents custom_target outputs, which are valid link_with
- * arguments...
- */
-#define tc_link_with_kw       (ARG_TYPE_ARRAY_OF | tc_build_target | tc_custom_target | tc_file | tc_both_libs)
-#define tc_message            (ARG_TYPE_GLOB | tc_string | tc_bool | tc_number | tc_array | tc_dict) // doesn't handle nested types
-
-struct obj_typechecking_type_to_obj_type {
-	enum obj_type type;
-	type_tag tc;
-};
-
 /* start of object structs */
 
 struct obj_typeinfo {
 	type_tag type, subtype;
 };
 
+struct obj_func {
+	const char *name;
+	struct ast *ast;
+	enum language_mode lang_mode;
+	uint32_t args_id, block_id, nargs, nkwargs;
+	obj kwarg_defaults, src, scope_stack;
+	type_tag return_type;
+
+	/*--------*/
+
+	uint32_t entry;
+	struct args_norm an[32];
+	struct args_kw akw[64];
+};
+
+struct obj_capture {
+	struct obj_func *func;
+	obj scope_stack, defargs;
+};
+
 enum tgt_type {
 	tgt_executable = 1 << 0,
 	tgt_static_library = 1 << 1,
 	tgt_dynamic_library = 1 << 2,
 	tgt_shared_module = 1 << 3,
 };
-enum tgt_type_count { tgt_type_count = 4, }; // keep in sync
+enum tgt_type_count {
+	tgt_type_count = 4,
+}; // keep in sync
 
 enum feature_opt_state {
 	feature_opt_auto,
@@ -151,40 +109,42 @@ enum feature_opt_state {
 	feature_opt_disabled,
 };
 
+#define FOREACH_BUILTIN_MODULE(_)                     \
+	_(fs, "public", true)                         \
+	_(keyval, "public", true)                     \
+	_(pkgconfig, "public", true)                  \
+	_(python, "public", true)                     \
+	_(python3, "public", true)                    \
+	_(sourceset, "public", true)                  \
+	_(toolchain, "private", true)                 \
+	_(cmake, "public", false)                     \
+	_(dlang, "public", false)                     \
+	_(gnome, "public", false)                     \
+	_(hotdoc, "public", false)                    \
+	_(i18n, "public", false)                      \
+	_(java, "public", false)                      \
+	_(modtest, "public", false)                   \
+	_(qt, "public", false)                        \
+	_(qt4, "public", false)                       \
+	_(qt5, "public", false)                       \
+	_(qt6, "public", false)                       \
+	_(unstable_cuda, "public", false)             \
+	_(unstable_external_project, "public", false) \
+	_(unstable_icestorm, "public", false)         \
+	_(unstable_rust, "public", false)             \
+	_(unstable_simd, "public", false)             \
+	_(unstable_wayland, "public", false)          \
+	_(windows, "public", false)
+
+#define MODULE_ENUM(mod, path, implemented) module_##mod,
 enum module {
-	module_fs,
-	module_keyval,
-	module_pkgconfig,
-	module_python,
-	module_python3,
-	module_sourceset,
-
-	// unimplemented
-	module_unimplemented_separator,
-	module_cmake = module_unimplemented_separator,
-	module_dlang,
-	module_gnome,
-	module_hotdoc,
-	module_i18n,
-	module_java,
-	module_modtest,
-	module_qt,
-	module_qt4,
-	module_qt5,
-	module_qt6,
-	module_unstable_cuda,
-	module_unstable_external_project,
-	module_unstable_icestorm,
-	module_unstable_rust,
-	module_unstable_simd,
-	module_unstable_wayland,
-	module_windows,
-
-	module_count,
+	FOREACH_BUILTIN_MODULE(MODULE_ENUM) module_count,
 };
+#undef MODULE_ENUM
 
 enum str_flags {
 	str_flag_big = 1 << 0,
+	str_flag_mutable = 1 << 1,
 };
 
 struct str {
@@ -206,6 +166,7 @@ struct obj_subproject {
 struct obj_module {
 	enum module module;
 	bool found, has_impl;
+	obj exports;
 };
 
 struct obj_array {
@@ -216,13 +177,21 @@ struct obj_array {
 	bool have_next;
 };
 
+enum obj_dict_flags {
+	obj_dict_flag_big = 1 << 0,
+	obj_dict_flag_int_key = 1 << 1,
+	obj_dict_flag_dont_expand = 1 << 2,
+};
+
+struct obj_dict_elem {
+	uint32_t next;
+	obj key, val;
+};
+
 struct obj_dict {
-	obj key; // obj_string
-	obj val; // any
-	obj next; // obj_array
-	obj tail; // obj_array
-	uint32_t len;
-	bool have_next;
+	uint32_t data, len;
+	obj tail;
+	enum obj_dict_flags flags;
 };
 
 enum build_tgt_flags {
@@ -249,6 +218,7 @@ struct build_dep {
 	obj include_directories; // obj_array
 
 	obj sources; // obj_array
+	obj objects; // obj_array
 
 	obj order_deps; // obj_array
 	obj rpath; // obj_array
@@ -309,7 +279,7 @@ struct obj_custom_target {
 	enum custom_target_flags flags;
 };
 
-struct  obj_alias_target {
+struct obj_alias_target {
 	obj name; // obj_string
 	obj depends; // obj_array
 };
@@ -320,6 +290,7 @@ enum dependency_type {
 	dependency_type_threads,
 	dependency_type_external_library,
 	dependency_type_appleframeworks,
+	dependency_type_not_found,
 };
 
 enum dep_flags {
@@ -346,10 +317,20 @@ struct obj_dependency {
 
 struct obj_external_program {
 	bool found, guessed_ver;
-	obj full_path;
+	obj cmd_array;
 	obj ver;
 };
 
+struct obj_python_installation {
+	obj prog;
+
+	bool pure;
+	obj language_version;
+	obj sysconfig_paths;
+	obj sysconfig_vars;
+	obj install_paths;
+};
+
 enum run_result_flags {
 	run_result_flag_from_compile = 1 << 0,
 	run_result_flag_compile_ok = 1 << 1,
@@ -395,10 +376,14 @@ struct obj_test {
 
 struct obj_compiler {
 	obj cmd_arr;
+	obj linker_cmd_arr;
+	obj static_linker_cmd_arr;
 	obj ver;
 	obj libdirs;
-	enum compiler_type type;
+	obj overrides[3];
+	uint32_t type[3];
 	enum compiler_language lang;
+	bool linker_passthrough;
 };
 
 enum install_target_type {
@@ -483,6 +468,7 @@ struct obj_generated_list {
 	obj input; // obj_array of obj_file
 	obj extra_arguments; // obj_array of obj_string
 	obj preserve_path_from; // obj_string
+	obj env;
 };
 
 struct obj_source_set {
@@ -494,23 +480,53 @@ struct obj_source_configuration {
 	obj sources, dependencies;
 };
 
+enum obj_iterator_type {
+	obj_iterator_type_array,
+	obj_iterator_type_dict_small,
+	obj_iterator_type_dict_big,
+	obj_iterator_type_range,
+	obj_iterator_type_typeinfo,
+};
+
+struct range_params {
+	uint32_t start, stop, step, i;
+};
+
+struct obj_iterator {
+	enum obj_iterator_type type;
+	union {
+		struct obj_array *array;
+		struct obj_dict_elem *dict_small;
+		struct {
+			struct hash *h;
+			uint32_t i;
+		} dict_big;
+		struct range_params range;
+		struct {
+			enum obj_type type;
+			uint32_t i;
+		} typeinfo;
+	} data;
+};
+
 /* end of object structs */
 
 struct obj_clear_mark {
 	uint32_t obji;
-	struct bucket_array_save objs, chrs;
-	struct bucket_array_save obj_aos[obj_type_count - _obj_aos_start];
+	struct bucket_arr_save objs, chrs;
+	struct bucket_arr_save obj_aos[obj_type_count - _obj_aos_start];
 };
 
 void make_obj(struct workspace *wk, obj *id, enum obj_type type);
 enum obj_type get_obj_type(struct workspace *wk, obj id);
-type_tag obj_type_to_tc_type(enum obj_type t);
 
 void obj_set_clear_mark(struct workspace *wk, struct obj_clear_mark *mk);
 void obj_clear(struct workspace *wk, const struct obj_clear_mark *mk);
 
+void set_obj_bool(struct workspace *wk, obj o, bool v);
 bool get_obj_bool(struct workspace *wk, obj o);
 void set_obj_bool(struct workspace *wk, obj o, bool v);
+obj make_number(struct workspace *wk, int64_t n);
 int64_t get_obj_number(struct workspace *wk, obj o);
 void set_obj_number(struct workspace *wk, obj o, int64_t v);
 obj *get_obj_file(struct workspace *wk, obj o);
@@ -518,8 +534,10 @@ const char *get_file_path(struct workspace *wk, obj o);
 const struct str *get_str(struct workspace *wk, obj s);
 enum feature_opt_state get_obj_feature_opt(struct workspace *wk, obj fo);
 void set_obj_feature_opt(struct workspace *wk, obj fo, enum feature_opt_state state);
+enum machine_kind get_obj_machine(struct workspace *wk, obj o);
+void set_obj_machine(struct workspace *wk, obj o, enum machine_kind kind);
 
-#define OBJ_GETTER(type) struct type *get_ ## type(struct workspace *wk, obj o)
+#define OBJ_GETTER(type) struct type *get_##type(struct workspace *wk, obj o)
 
 OBJ_GETTER(obj_array);
 OBJ_GETTER(obj_dict);
@@ -529,6 +547,7 @@ OBJ_GETTER(obj_custom_target);
 OBJ_GETTER(obj_subproject);
 OBJ_GETTER(obj_dependency);
 OBJ_GETTER(obj_external_program);
+OBJ_GETTER(obj_python_installation);
 OBJ_GETTER(obj_run_result);
 OBJ_GETTER(obj_configuration_data);
 OBJ_GETTER(obj_test);
@@ -542,27 +561,39 @@ OBJ_GETTER(obj_generated_list);
 OBJ_GETTER(obj_alias_target);
 OBJ_GETTER(obj_both_libs);
 OBJ_GETTER(obj_typeinfo);
+OBJ_GETTER(obj_func);
+OBJ_GETTER(obj_capture);
 OBJ_GETTER(obj_source_set);
 OBJ_GETTER(obj_source_configuration);
+OBJ_GETTER(obj_iterator);
 
 #undef OBJ_GETTER
 
 struct sbuf;
 
 const char *obj_type_to_s(enum obj_type t);
+bool s_to_type_tag(const char *s, type_tag *t);
 void obj_to_s(struct workspace *wk, obj o, struct sbuf *sb);
 bool obj_equal(struct workspace *wk, obj left, obj right);
 bool obj_clone(struct workspace *wk_src, struct workspace *wk_dest, obj val, obj *ret);
 
+#define LO(...)                                                   \
+	if (log_should_print(log_debug)) {                        \
+		char buf[4096];                                   \
+		log_print_prefix(log_debug, buf, ARRAY_LEN(buf)); \
+		log_plain("%s", buf);                             \
+		obj_fprintf(wk, log_file(), __VA_ARGS__);         \
+	}
+#define LOBJ(object_id) LO("%s: %o\n", #object_id, object_id)
+
 bool obj_vasprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, va_list ap);
-bool obj_asprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...)
-MUON_ATTR_FORMAT(printf, 3, 4);
-bool obj_vfprintf(struct workspace *wk, FILE *f, const char *fmt, va_list ap)
-MUON_ATTR_FORMAT(printf, 3, 0);
-bool obj_fprintf(struct workspace *wk, FILE *f, const char *fmt, ...)
-MUON_ATTR_FORMAT(printf, 3, 4);
-bool obj_printf(struct workspace *wk, const char *fmt, ...)
-MUON_ATTR_FORMAT(printf, 2, 3);
+bool obj_asprintf(struct workspace *wk, struct sbuf *sb, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
+bool obj_vfprintf(struct workspace *wk, FILE *f, const char *fmt, va_list ap) MUON_ATTR_FORMAT(printf, 3, 0);
+bool obj_fprintf(struct workspace *wk, FILE *f, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 3, 4);
+bool obj_printf(struct workspace *wk, const char *fmt, ...) MUON_ATTR_FORMAT(printf, 2, 3);
+uint32_t obj_vsnprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, va_list ap);
+uint32_t obj_snprintf(struct workspace *wk, char *buf, uint32_t len, const char *fmt, ...)
+	MUON_ATTR_FORMAT(printf, 4, 5);
 void obj_inspect(struct workspace *wk, FILE *out, obj val);
 
 typedef enum iteration_result (*obj_array_iterator)(struct workspace *wk, void *ctx, obj val);
@@ -581,22 +612,30 @@ void obj_array_tail(struct workspace *wk, obj arr, obj *res);
 void obj_array_set(struct workspace *wk, obj arr, int64_t i, obj v);
 void obj_array_del(struct workspace *wk, obj arr, int64_t i);
 void obj_array_dedup(struct workspace *wk, obj arr, obj *res);
+void obj_array_dedup_in_place(struct workspace *wk, obj *arr);
 bool obj_array_flatten_one(struct workspace *wk, obj val, obj *res);
 typedef int32_t (*obj_array_sort_func)(struct workspace *wk, void *_ctx, obj a, obj b);
 int32_t obj_array_sort_by_str(struct workspace *wk, void *_ctx, obj a, obj b);
 void obj_array_sort(struct workspace *wk, void *usr_ctx, obj arr, obj_array_sort_func func, obj *res);
+obj obj_array_slice(struct workspace *wk, obj arr, int64_t i0, int64_t i1);
+obj obj_array_get_tail(struct workspace *wk, obj arr);
+obj obj_array_pop(struct workspace *wk, obj arr);
 
 typedef enum iteration_result (*obj_dict_iterator)(struct workspace *wk, void *ctx, obj key, obj val);
 bool obj_dict_foreach(struct workspace *wk, obj dict, void *ctx, obj_dict_iterator cb);
 bool obj_dict_in(struct workspace *wk, obj dict, obj key);
 bool obj_dict_index(struct workspace *wk, obj dict, obj key, obj *res);
 bool obj_dict_index_strn(struct workspace *wk, obj dict, const char *str, uint32_t len, obj *res);
+bool obj_dict_index_str(struct workspace *wk, obj dict, const char *str, obj *res);
 void obj_dict_set(struct workspace *wk, obj dict, obj key, obj val);
 void obj_dict_dup(struct workspace *wk, obj dict, obj *res);
 void obj_dict_merge(struct workspace *wk, obj dict, obj dict2, obj *res);
 void obj_dict_merge_nodup(struct workspace *wk, obj dict, obj dict2);
 void obj_dict_seti(struct workspace *wk, obj dict, uint32_t key, obj val);
 bool obj_dict_geti(struct workspace *wk, obj dict, uint32_t key, obj *val);
+void obj_dict_del(struct workspace *wk, obj dict, obj key);
+void obj_dict_del_str(struct workspace *wk, obj dict, const char *str);
+void obj_dict_del_strn(struct workspace *wk, obj dict, const char *str, uint32_t len);
 
 bool obj_iterable_foreach(struct workspace *wk, obj dict_or_array, void *ctx, obj_dict_iterator cb);
 #endif
