@@ -22,15 +22,17 @@
 #include "cmd_test.c"
 #include "coerce.c"
 #include "compilers.c"
-#include "data/bucket_array.c"
-#include "data/darr.c"
-#include "data/hash.c"
+#include "datastructures/arr.c"
+#include "datastructures/bucket_arr.c"
+#include "datastructures/hash.c"
+#include "datastructures/stack.c"
 #include "embedded.c"
 #include "error.c"
-#include "external/bestline_null.c"
 #include "external/libarchive_null.c"
 #include "external/libcurl_null.c"
-#include "external/samurai_null.c"
+#include "external/readline_builtin.c"
+#include "external/tinyjson_null.c"
+#include "formats/editorconfig.c"
 #include "formats/ini.c"
 #include "formats/lines.c"
 #include "formats/tap.c"
@@ -38,7 +40,6 @@
 #include "functions/boolean.c"
 #include "functions/both_libs.c"
 #include "functions/build_target.c"
-#include "functions/common.c"
 #include "functions/compiler.c"
 #include "functions/configuration_data.c"
 #include "functions/custom_target.c"
@@ -66,6 +67,7 @@
 #include "functions/modules/pkgconfig.c"
 #include "functions/modules/python.c"
 #include "functions/modules/sourceset.c"
+#include "functions/modules/toolchain.c"
 #include "functions/number.c"
 #include "functions/run_result.c"
 #include "functions/source_configuration.c"
@@ -75,23 +77,31 @@
 #include "guess.c"
 #include "install.c"
 #include "lang/analyze.c"
+#include "lang/compiler.c"
 #include "lang/eval.c"
 #include "lang/fmt.c"
-#include "lang/interpreter.c"
+#include "lang/func_lookup.c"
 #include "lang/lexer.c"
 #include "lang/object.c"
+#include "lang/object_iterators.c"
 #include "lang/parser.c"
 #include "lang/serial.c"
 #include "lang/string.c"
+#include "lang/typecheck.c"
+#include "lang/vm.c"
 #include "lang/workspace.c"
 #include "log.c"
 #include "machine_file.c"
+#include "machines.c"
 #include "main.c"
+#include "memmem.c"
 #include "meson_opts.c"
 #include "options.c"
 #include "opts.c"
+#include "platform/assert.c"
 #include "platform/filesystem.c"
 #include "platform/mem.c"
+#include "platform/os.c"
 #include "platform/path.c"
 #include "platform/run_cmd.c"
 #include "platform/uname.c"
@@ -102,21 +112,26 @@
 
 #ifdef _WIN32
 #include "platform/windows/filesystem.c"
+#include "platform/windows/init.c"
 #include "platform/windows/log.c"
+#include "platform/windows/os.c"
 #include "platform/windows/path.c"
 #include "platform/windows/rpath_fixer.c"
 #include "platform/windows/run_cmd.c"
 #include "platform/windows/term.c"
+#include "platform/windows/timer.c"
 #include "platform/windows/uname.c"
 #include "platform/windows/win32_error.c"
-#include "platform/windows/win32_getopt.c"
 #else
 #include "platform/null/rpath_fixer.c"
 #include "platform/posix/filesystem.c"
+#include "platform/posix/init.c"
 #include "platform/posix/log.c"
+#include "platform/posix/os.c"
 #include "platform/posix/path.c"
 #include "platform/posix/run_cmd.c"
 #include "platform/posix/term.c"
+#include "platform/posix/timer.c"
 #include "platform/posix/uname.c"
 #endif
 
@@ -125,3 +140,21 @@
 #else
 #include "external/libpkgconf_null.c"
 #endif
+
+#ifdef BOOTSTRAP_NO_SAMU
+#include "external/samurai_null.c"
+#else
+#include "external/samurai.c"
+#include "external/samurai/build.c"
+#include "external/samurai/deps.c"
+#include "external/samurai/env.c"
+#include "external/samurai/graph.c"
+#include "external/samurai/htab.c"
+#include "external/samurai/log.c"
+#include "external/samurai/parse.c"
+#include "external/samurai/samu.c"
+#include "external/samurai/scan.c"
+#include "external/samurai/tool.c"
+#include "external/samurai/tree.c"
+#include "external/samurai/util.c"
+#endif
