@@ -6,19 +6,18 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <string.h>
 
 #include "args.h"
 #include "buf_size.h"
 #include "compilers.h"
+#include "error.h"
 #include "functions/build_target.h"
 #include "functions/environment.h"
-#include "lang/interpreter.h"
 #include "lang/workspace.h"
 #include "log.h"
+#include "platform/assert.h"
 #include "platform/path.h"
-#include "error.h"
 
 void
 push_args(struct workspace *wk, obj arr, const struct args *args)
@@ -77,7 +76,7 @@ pkgconf_escape(struct workspace *wk, struct sbuf *sb, const char *str)
 	simple_escape(wk, sb, str, " ", '\\');
 }
 
-typedef void ((*escape_func)(struct workspace *wk, struct sbuf *sb, const char *str));
+typedef void((*escape_func)(struct workspace *wk, struct sbuf *sb, const char *str));
 
 struct join_args_iter_ctx {
 	uint32_t i, len;
@@ -99,10 +98,10 @@ join_args_iter(struct workspace *wk, void *_ctx, obj val)
 		s = esc.buf;
 	}
 
-	str_app(wk, *ctx->obj, s);
+	str_app(wk, ctx->obj, s);
 
 	if (ctx->i < ctx->len - 1) {
-		str_app(wk, *ctx->obj, " ");
+		str_app(wk, ctx->obj, " ");
 	}
 
 	++ctx->i;
@@ -115,11 +114,7 @@ join_args(struct workspace *wk, obj arr, escape_func escape)
 {
 	obj o = make_str(wk, "");
 
-	struct join_args_iter_ctx ctx = {
-		.obj = &o,
-		.len = get_obj_array(wk, arr)->len,
-		.escape = escape
-	};
+	struct join_args_iter_ctx ctx = { .obj = &o, .len = get_obj_array(wk, arr)->len, .escape = escape };
 
 	obj_array_foreach(wk, arr, &ctx, join_args_iter);
 	return o;
@@ -169,9 +164,7 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 	enum obj_type t = get_obj_type(wk, src);
 
 	switch (t) {
-	case obj_string:
-		str = src;
-		break;
+	case obj_string: str = src; break;
 	case obj_file:
 		if (ctx->mode & arr_to_args_relativize_paths) {
 			SBUF(rel);
@@ -187,8 +180,7 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 		}
 		str = get_obj_alias_target(wk, src)->name;
 		break;
-	case obj_both_libs:
-		src = get_obj_both_libs(wk, src)->dynamic_lib;
+	case obj_both_libs: src = get_obj_both_libs(wk, src)->dynamic_lib;
 	/* fallthrough */
 	case obj_build_target: {
 		if (!(ctx->mode & arr_to_args_build_target)) {
@@ -212,7 +204,7 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 			goto type_err;
 		}
 
-		struct obj_custom_target* custom_tgt = get_obj_custom_target(wk, src);
+		struct obj_custom_target *custom_tgt = get_obj_custom_target(wk, src);
 		obj output_arr = custom_tgt->output;
 
 		// run_target is a custom_target without an output, so we yield
@@ -228,15 +220,14 @@ arr_to_args_iter(struct workspace *wk, void *_ctx, obj src)
 		return ir_cont;
 	}
 	case obj_external_program:
+	case obj_python_installation:
 		if (!(ctx->mode & arr_to_args_external_program)) {
 			goto type_err;
 		}
 
-		str = get_obj_external_program(wk, src)->full_path;
-		break;
-	case obj_compiler:
-		obj_array_extend(wk, ctx->res, get_obj_compiler(wk, src)->cmd_arr);
+		obj_array_extend(wk, ctx->res, get_obj_external_program(wk, src)->cmd_array);
 		return ir_cont;
+	case obj_compiler: obj_array_extend(wk, ctx->res, get_obj_compiler(wk, src)->cmd_arr); return ir_cont;
 	default:
 type_err:
 		LOG_E("cannot convert '%s' to argument", obj_type_to_s(t));
@@ -252,10 +243,7 @@ arr_to_args(struct workspace *wk, enum arr_to_args_flags mode, obj arr, obj *res
 {
 	make_obj(wk, res, obj_array);
 
-	struct arr_to_args_ctx ctx = {
-		.mode = mode,
-		.res = *res
-	};
+	struct arr_to_args_ctx ctx = { .mode = mode, .res = *res };
 
 	return obj_array_foreach_flat(wk, arr, &ctx, arr_to_args_iter);
 }
@@ -271,7 +259,7 @@ join_args_argstr_iter(struct workspace *wk, void *_ctx, obj v)
 
 	const struct str *s = get_str(wk, v);
 
-	str_appn(wk, ctx->str, s->s, s->len + 1);
+	str_appn(wk, &ctx->str, s->s, s->len + 1);
 
 	return ir_cont;
 }
@@ -298,11 +286,10 @@ env_to_envstr_dict_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 {
 	struct env_to_envstr_ctx *ctx = _ctx;
 
-	const struct str *k = get_str(wk, key),
-			 *v = get_str(wk, val);
+	const struct str *k = get_str(wk, key), *v = get_str(wk, val);
 
-	str_appn(wk, ctx->str, k->s, k->len + 1);
-	str_appn(wk, ctx->str, v->s, v->len + 1);
+	str_appn(wk, &ctx->str, k->s, k->len + 1);
+	str_appn(wk, &ctx->str, v->s, v->len + 1);
 
 	return ir_cont;
 }
