@@ -46,11 +46,8 @@ check_tgt_iter(struct workspace *wk, void *_ctx, obj tgt_id)
 	}
 	case obj_alias_target:
 	case obj_build_target:
-	case obj_both_libs:
-		break;
-	default:
-		LOG_E("invalid tgt type '%s'", obj_type_to_s(t));
-		return ir_err;
+	case obj_both_libs: break;
+	default: UNREACHABLE;
 	}
 
 	return ir_cont;
@@ -66,35 +63,35 @@ write_tgt_iter(struct workspace *wk, void *_ctx, obj tgt_id)
 	struct obj_clear_mark mk;
 	obj_set_clear_mark(wk, &mk);
 
+	enum obj_type t = get_obj_type(wk, tgt_id);
 	const char *name = NULL;
 
-	enum obj_type t = get_obj_type(wk, tgt_id);
 	switch (t) {
-	case obj_alias_target:
-		ret = ninja_write_alias_tgt(wk, tgt_id, ctx);
-		name = get_cstr(wk, get_obj_alias_target(wk, tgt_id)->name);
-		break;
-	case obj_both_libs:
-		tgt_id = get_obj_both_libs(wk, tgt_id)->dynamic_lib;
+	case obj_alias_target: name = get_cstr(wk, get_obj_alias_target(wk, tgt_id)->name); break;
+	case obj_both_libs: tgt_id = get_obj_both_libs(wk, tgt_id)->dynamic_lib;
 	/* fallthrough */
-	case obj_build_target:
-		ret = ninja_write_build_tgt(wk, tgt_id, ctx);
-		name = get_cstr(wk, get_obj_build_target(wk, tgt_id)->build_name);
-		break;
-	case obj_custom_target:
-		ret = ninja_write_custom_tgt(wk, tgt_id, ctx);
-		name = get_cstr(wk, get_obj_custom_target(wk, tgt_id)->name);
-		break;
-	default:
-		LOG_E("invalid tgt type '%s'", obj_type_to_s(t));
-		ret = ir_err;
-		break;
+	case obj_build_target: name = get_cstr(wk, get_obj_build_target(wk, tgt_id)->build_name); break;
+	case obj_custom_target: name = get_cstr(wk, get_obj_custom_target(wk, tgt_id)->name); break;
+	default: UNREACHABLE;
+	}
+
+	obj_array_push(wk, wk->backend_output_stack, make_strf(wk, "writing target %s", name));
+
+	switch (t) {
+	case obj_alias_target: ret = ninja_write_alias_tgt(wk, tgt_id, ctx); break;
+	case obj_both_libs: tgt_id = get_obj_both_libs(wk, tgt_id)->dynamic_lib;
+	/* fallthrough */
+	case obj_build_target: ret = ninja_write_build_tgt(wk, tgt_id, ctx); break;
+	case obj_custom_target: ret = ninja_write_custom_tgt(wk, tgt_id, ctx); break;
+	default: UNREACHABLE;
 	}
 
 	if (!ret) {
 		LOG_E("failed to write %s '%s'", obj_type_to_s(t), name);
 	}
 
+	obj_array_pop(wk, wk->backend_output_stack);
+
 	obj_clear(wk, &mk);
 
 	TracyCZoneAutoE;
@@ -113,7 +110,7 @@ ninja_write_build(struct workspace *wk, void *_ctx, FILE *out)
 
 	uint32_t i;
 	for (i = 0; i < wk->projects.len; ++i) {
-		struct project *proj = darr_get(&wk->projects, i);
+		struct project *proj = arr_get(&wk->projects, i);
 		if (proj->not_ok) {
 			continue;
 		}
@@ -121,14 +118,14 @@ ninja_write_build(struct workspace *wk, void *_ctx, FILE *out)
 		obj_array_foreach(wk, proj->targets, &check_ctx, check_tgt_iter);
 	}
 
-	if (!ninja_write_rules(out, wk, darr_get(&wk->projects, 0), check_ctx.need_phony, ctx->compiler_rule_arr)) {
+	if (!ninja_write_rules(out, wk, arr_get(&wk->projects, 0), check_ctx.need_phony, ctx->compiler_rule_arr)) {
 		return false;
 	}
 
 	bool wrote_default = false;
 
 	for (i = 0; i < wk->projects.len; ++i) {
-		struct project *proj = darr_get(&wk->projects, i);
+		struct project *proj = arr_get(&wk->projects, i);
 		if (proj->not_ok) {
 			continue;
 		}
@@ -146,8 +143,7 @@ ninja_write_build(struct workspace *wk, void *_ctx, FILE *out)
 	if (!wrote_default) {
 		fprintf(out,
 			"build muon_do_nothing: phony\n"
-			"default muon_do_nothing\n"
-			);
+			"default muon_do_nothing\n");
 	}
 
 	return true;
@@ -163,7 +159,7 @@ ninja_write_tests(struct workspace *wk, void *_ctx, FILE *out)
 
 	uint32_t i;
 	for (i = 0; i < wk->projects.len; ++i) {
-		struct project *proj = darr_get(&wk->projects, i);
+		struct project *proj = arr_get(&wk->projects, i);
 		if (proj->not_ok) {
 			continue;
 		}
@@ -202,7 +198,7 @@ ninja_write_install(struct workspace *wk, void *_ctx, FILE *out)
 	obj_array_push(wk, o, wk->install_scripts);
 	obj_array_push(wk, o, make_str(wk, wk->source_root));
 
-	struct project *proj = darr_get(&wk->projects, 0);
+	struct project *proj = arr_get(&wk->projects, 0);
 	obj prefix;
 	get_option_value(wk, proj, "prefix", &prefix);
 	obj_array_push(wk, o, prefix);
@@ -230,7 +226,7 @@ ninja_write_option_info(struct workspace *wk, void *_ctx, FILE *out)
 	make_obj(wk, &arr, obj_array);
 	obj_array_push(wk, arr, wk->global_opts);
 
-	struct project *main_proj = darr_get(&wk->projects, 0);
+	struct project *main_proj = arr_get(&wk->projects, 0);
 	obj_array_push(wk, arr, main_proj->opts);
 
 	return serial_dump(wk, arr, out);
@@ -242,17 +238,24 @@ ninja_write_all(struct workspace *wk)
 	struct write_build_ctx ctx = { 0 };
 	make_obj(wk, &ctx.compiler_rule_arr, obj_array);
 
+	obj_array_push(wk, wk->backend_output_stack, make_str(wk, "ninja_write_all"));
+
 	if (!(with_open(wk->build_root, "build.ninja", wk, &ctx, ninja_write_build)
-	      && with_open(wk->muon_private, output_path.tests, wk, NULL, ninja_write_tests)
-	      && with_open(wk->muon_private, output_path.install, wk, NULL, ninja_write_install)
-	      && with_open(wk->muon_private, output_path.compiler_check_cache, wk, NULL, ninja_write_compiler_check_cache)
-	      && with_open(wk->muon_private, output_path.summary, wk, NULL, ninja_write_summary_file)
-	      && with_open(wk->muon_private, output_path.option_info, wk, NULL, ninja_write_option_info)
-	      )) {
+		    && with_open(wk->muon_private, output_path.tests, wk, NULL, ninja_write_tests)
+		    && with_open(wk->muon_private, output_path.install, wk, NULL, ninja_write_install)
+		    && with_open(wk->muon_private,
+			    output_path.compiler_check_cache,
+			    wk,
+			    NULL,
+			    ninja_write_compiler_check_cache)
+		    && with_open(wk->muon_private, output_path.summary, wk, NULL, ninja_write_summary_file)
+		    && with_open(wk->muon_private, output_path.option_info, wk, NULL, ninja_write_option_info))) {
 		return false;
 	}
 
-	{/* compile_commands.json */
+	obj_array_pop(wk, wk->backend_output_stack);
+
+	{ /* compile_commands.json */
 		TracyCZoneN(tctx_compdb, "output compile_commands.json", true);
 
 		obj compdb_args;
@@ -262,7 +265,7 @@ ninja_write_all(struct workspace *wk)
 		obj_array_push(wk, compdb_args, make_str(wk, "-t"));
 		obj_array_push(wk, compdb_args, make_str(wk, "compdb"));
 		obj_array_extend_nodup(wk, compdb_args, ctx.compiler_rule_arr);
-		if (ninja_run(wk, compdb_args, wk->build_root, "compile_commands.json") != 0) {
+		if (!ninja_run(wk, compdb_args, wk->build_root, "compile_commands.json")) {
 			LOG_E("error writing compile_commands.json");
 		}
 
@@ -272,58 +275,43 @@ ninja_write_all(struct workspace *wk)
 	return true;
 }
 
-int
+bool
 ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture)
 {
-	// XXX since samu was designed to be an executable and not a library,
-	// lots of the resource management is left to the OS.  For instance,
-	// there are several important globals that are assumed to be
-	// zero-initialized.  Not to mention memory "leaks".  This is all fine
-	// since almost zero effort has been put in to making samu into a true
-	// libsamu, however it means that calling the internal samu more than
-	// once is riddled with UB.  Prevent that with this hacky static
-	// variable by  falling back to executing an external ninja-compatible
-	// tool if the internal samu has already been invoked.
-	static bool internal_samu_has_been_called = false;
-
 	const char *argstr;
 	uint32_t argstr_argc;
 
-	int ret = 1;
+	bool ret = false;
 	char *const *argv = NULL;
 	uint32_t argc;
 	SBUF_manual(cwd);
 
 	if (chdir) {
-		path_cwd(NULL, &cwd);
+		path_copy_cwd(NULL, &cwd);
 
 		if (!path_chdir(chdir)) {
 			goto ret;
 		}
 	}
 
-	if (have_samurai && !internal_samu_has_been_called) {
-		internal_samu_has_been_called = true;
-
+	if (have_samurai) {
 		join_args_argstr(wk, &argstr, &argstr_argc, args);
 		argc = argstr_to_argv(argstr, argstr_argc, "samu", &argv);
 
-		int old_stdout;
+		struct samu_opts samu_opts = { .out = stdout };
 		if (capture) {
-			if (!fs_redirect(capture, "wb", STDOUT_FILENO, &old_stdout)) {
+			if (!(samu_opts.out = fs_fopen(capture, "wb"))) {
 				goto ret;
 			}
 		}
 
-		bool res = muon_samu(argc, argv);
+		ret = samu_main(argc, (char **)argv, &samu_opts);
 
 		if (capture) {
-			if (!fs_redirect_restore(STDOUT_FILENO, old_stdout)) {
+			if (!fs_fclose(samu_opts.out)) {
 				goto ret;
 			}
 		}
-
-		ret = res ? 0 : 1;
 	} else {
 		struct run_cmd_ctx cmd_ctx = { 0 };
 
@@ -343,8 +331,8 @@ ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture
 			obj_array_push(wk, cmd_arr, make_str(wk, "samu"));
 			obj_array_extend_nodup(wk, cmd_arr, args);
 			args = cmd_arr;
-		} else if (ninja_opt->source == option_value_source_default &&
-			   (fs_find_cmd(NULL, &cmd, "samu") || fs_find_cmd(NULL, &cmd, "ninja"))) {
+		} else if (ninja_opt->source == option_value_source_default
+			   && (fs_find_cmd(NULL, &cmd, "samu") || fs_find_cmd(NULL, &cmd, "ninja"))) {
 			prepend = cmd.buf;
 		} else {
 			obj cmd_arr;
@@ -371,7 +359,7 @@ ninja_run(struct workspace *wk, obj args, const char *chdir, const char *capture
 			}
 		}
 
-		ret = cmd_ctx.status;
+		ret = cmd_ctx.status == 0;
 run_cmd_done:
 		sbuf_destroy(&cmd);
 		run_cmd_ctx_destroy(&cmd_ctx);
