@@ -13,7 +13,6 @@
 #include "backend/ninja/build_target.h"
 #include "error.h"
 #include "functions/build_target.h"
-#include "functions/kernel/dependency.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "platform/filesystem.h"
@@ -104,11 +103,11 @@ write_tgt_sources_iter(struct workspace *wk, void *_ctx, obj val)
 	if (!specialized_rule) {
 		obj args;
 		if (!obj_dict_geti(wk, ctx->joined_args, lang, &args)) {
-			UNREACHABLE;
+			LOG_E("No compiler defined for language %s", compiler_language_to_s(lang));
+			return ir_err;
 		}
 
-		fprintf(ctx->out,
-			" ARGS = %s\n", get_cstr(wk, args));
+		fprintf(ctx->out, " ARGS = %s\n", get_cstr(wk, args));
 	}
 
 	return ir_cont;
@@ -121,8 +120,8 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 	L("writing rules for target '%s'", get_cstr(wk, tgt->build_name));
 
 	SBUF(esc_path);
+	SBUF(rel_build_path);
 	{
-		SBUF(rel_build_path);
 		path_relative_to(wk, &rel_build_path, wk->build_root, get_cstr(wk, tgt->build_path));
 		ninja_escape(wk, &esc_path, rel_build_path.buf);
 	}
@@ -133,15 +132,16 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		.out = wctx->out,
 	};
 
-	enum linker_type linker;
+	struct obj_compiler *compiler;
 	{ /* determine linker */
 		obj comp_id;
 		if (!obj_dict_geti(wk, ctx.proj->compilers, tgt->dep_internal.link_language, &comp_id)) {
-			LOG_E("no compiler defined for language %s", compiler_language_to_s(tgt->dep_internal.link_language));
+			LOG_E("no compiler defined for language %s",
+				compiler_language_to_s(tgt->dep_internal.link_language));
 			return false;
 		}
 
-		linker = compilers[get_obj_compiler(wk, comp_id)->type].linker;
+		compiler = get_obj_compiler(wk, comp_id);
 	}
 
 	make_obj(wk, &ctx.object_names, obj_array);
@@ -151,6 +151,7 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 	relativize_paths(wk, ctx.args.link_with, true, &ctx.args.link_with);
 	relativize_paths(wk, ctx.args.link_whole, true, &ctx.args.link_whole);
 
+	bool have_custom_order_deps = false;
 	{ /* order deps */
 		if ((ctx.have_order_deps = (get_obj_array(wk, ctx.args.order_deps)->len > 0))) {
 			obj deduped;
@@ -158,9 +159,13 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 			obj order_deps = join_args_ninja(wk, deduped);
 
 			if (get_obj_array(wk, deduped)->len > 1) {
-				fprintf(wctx->out, "build %s-order_deps: phony || %s\n", esc_path.buf, get_cstr(wk, order_deps));
+				fprintf(wctx->out,
+					"build %s-order_deps: phony || %s\n",
+					esc_path.buf,
+					get_cstr(wk, order_deps));
 				ctx.have_order_deps = false;
 				ctx.implicit_deps = make_strf(wk, "%s-order_deps", esc_path.buf);
+				have_custom_order_deps = true;
 			} else {
 				ctx.order_deps = order_deps;
 			}
@@ -177,15 +182,14 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 
 	obj implicit_link_deps;
 	make_obj(wk, &implicit_link_deps, obj_array);
-	if (ctx.implicit_deps) {
-		obj_array_push(wk, implicit_link_deps, ctx.implicit_deps);
+	if (have_custom_order_deps) {
+		obj_array_push(wk, implicit_link_deps, make_strf(wk, "%s-order_deps", rel_build_path.buf));
 	}
 
-	if (!(tgt->type & (tgt_static_library))) {
+	if (!(tgt->type & tgt_static_library)) {
 		struct setup_linker_args_ctx sctx = {
-			.linker = linker,
-			.link_lang = tgt->dep_internal.link_language,
-			.args = &ctx.args
+			.compiler = compiler,
+			.args = &ctx.args,
 		};
 
 		setup_linker_args(wk, ctx.proj, tgt, &sctx);
@@ -208,37 +212,22 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		obj_array_extend_nodup(wk, implicit_link_deps, arr);
 	}
 
-	if (tgt->type & (tgt_dynamic_library | tgt_shared_module)) {
-		push_args(wk, ctx.args.link_args, linkers[linker].args.shared());
-		push_args(wk, ctx.args.link_args, linkers[linker].args.soname(get_cstr(wk, tgt->soname)));
-		if (tgt->type == tgt_shared_module) {
-			push_args(wk, ctx.args.link_args, linkers[linker].args.allow_shlib_undefined());
-		}
-	}
-
 	const char *linker_type, *link_args;
-	bool linker_rule_prefix = false;
 	switch (tgt->type) {
 	case tgt_shared_module:
 	case tgt_dynamic_library:
 	case tgt_executable:
 		linker_type = compiler_language_to_s(tgt->dep_internal.link_language);
-		linker_rule_prefix = true;
 		link_args = get_cstr(wk, join_args_shell_ninja(wk, ctx.args.link_args));
 		break;
 	case tgt_static_library:
 		linker_type = "static";
-		link_args = ar_arguments();
+		link_args = 0;
 		break;
-	default:
-		assert(false);
-		return false;
+	default: assert(false); return false;
 	}
 
-	fprintf(wctx->out, "build %s: %s%s%s_linker ", esc_path.buf,
-		linker_rule_prefix ? get_cstr(wk, ctx.proj->rule_prefix) : "",
-		linker_rule_prefix ? "_" : "",
-		linker_type);
+	fprintf(wctx->out, "build %s: %s_%s_linker ", esc_path.buf, get_cstr(wk, ctx.proj->rule_prefix), linker_type);
 	fputs(get_cstr(wk, join_args_ninja(wk, ctx.object_names)), wctx->out);
 	if (get_obj_array(wk, implicit_link_deps)->len) {
 		implicit_link_deps = join_args_ninja(wk, implicit_link_deps);
@@ -249,11 +238,13 @@ ninja_write_build_tgt(struct workspace *wk, obj tgt_id, struct write_tgt_ctx *wc
 		fputs(" || ", wctx->out);
 		fputs(get_cstr(wk, ctx.order_deps), wctx->out);
 	}
-	fprintf(wctx->out, "\n LINK_ARGS = %s\n", link_args);
+	if (link_args) {
+		fprintf(wctx->out, "\n LINK_ARGS = %s", link_args);
+	}
 
 	if (tgt->flags & build_tgt_flag_build_by_default) {
 		wctx->wrote_default = true;
-		fprintf(wctx->out, "default %s\n", esc_path.buf);
+		fprintf(wctx->out, "\ndefault %s\n", esc_path.buf);
 	}
 	fprintf(wctx->out, "\n");
 	return true;
