@@ -12,6 +12,7 @@
 #include "backend/ninja/rules.h"
 #include "backend/output.h"
 #include "error.h"
+#include "functions/machine.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "platform/path.h"
@@ -37,25 +38,48 @@ uniqify_name(struct workspace *wk, obj arr, obj name, obj *res)
 	*res = name;
 }
 
-static enum iteration_result
-write_linker_rule_iter(struct workspace *wk, void *_ctx, enum compiler_language l, obj comp_id)
+static void
+escape_rule(struct sbuf *buf)
 {
+	uint32_t i;
+	for (i = 0; i < buf->len; ++i) {
+		if (buf->buf[i] == '_' || ('a' <= buf->buf[i] && buf->buf[i] <= 'z')
+			|| ('A' <= buf->buf[i] && buf->buf[i] <= 'Z') || ('0' <= buf->buf[i] && buf->buf[i] <= '9')) {
+			continue;
+		}
+
+		buf->buf[i] = '_';
+	}
+}
+
+static enum iteration_result
+write_linker_rule_iter(struct workspace *wk, void *_ctx, obj k, obj comp_id)
+{
+	enum compiler_language l = k;
 	struct write_compiler_rule_ctx *ctx = _ctx;
 	struct obj_compiler *comp = get_obj_compiler(wk, comp_id);
-	enum compiler_type t = comp->type;
 
 	obj args;
 	make_obj(wk, &args, obj_array);
-	obj_array_extend(wk, args, comp->cmd_arr);
-	obj_array_push(wk, args, make_str(wk, "$ARGS"));
 
-	push_args(wk, args, compilers[t].args.output("$out"));
-	obj_array_push(wk, args, make_str(wk, "$in"));
-	obj_array_push(wk, args, make_str(wk, "$LINK_ARGS"));
+	if (comp->linker_passthrough) {
+		obj_array_extend(wk, args, comp->cmd_arr);
+		obj_array_push(wk, args, make_str(wk, "$ARGS"));
+
+		push_args(wk, args, toolchain_compiler_output(wk, comp, "$out"));
+		obj_array_push(wk, args, make_str(wk, "$in"));
+		obj_array_push(wk, args, make_str(wk, "$LINK_ARGS"));
+	} else {
+		obj_array_extend(wk, args, comp->linker_cmd_arr);
+		obj_array_push(wk, args, make_str(wk, "$ARGS"));
+		push_args(wk, args, toolchain_linker_input_output(wk, comp, "$in", "$out"));
+		obj_array_push(wk, args, make_str(wk, "$LINK_ARGS"));
+	}
 
 	obj link_command = join_args_plain(wk, args);
 
-	fprintf(ctx->out, "rule %s_%s_linker\n"
+	fprintf(ctx->out,
+		"rule %s_%s_linker\n"
 		" command = %s\n"
 		" description = linking $out\n\n",
 		get_cstr(wk, ctx->proj->rule_prefix),
@@ -69,18 +93,13 @@ static void
 write_compiler_rule(struct workspace *wk, FILE *out, obj rule_args, obj rule_name, enum compiler_language l, obj comp_id)
 {
 	struct obj_compiler *comp = get_obj_compiler(wk, comp_id);
-	enum compiler_type t = comp->type;
 
-	const char *deps = NULL;
-	switch (compilers[t].deps) {
-	case compiler_deps_none:
-		break;
-	case compiler_deps_gcc:
-		deps = "gcc";
-		break;
-	case compiler_deps_msvc:
-		deps = "msvc";
-		break;
+	const char *deps = 0;
+	{
+		const struct args *deps_args = toolchain_compiler_deps_type(wk, comp);
+		if (deps_args->len) {
+			deps = deps_args->args[0];
+		}
 	}
 
 	obj args;
@@ -88,33 +107,36 @@ write_compiler_rule(struct workspace *wk, FILE *out, obj rule_args, obj rule_nam
 	obj_array_extend(wk, args, comp->cmd_arr);
 	obj_array_push(wk, args, rule_args);
 
-	if (compilers[t].deps) {
-		push_args(wk, args, compilers[t].args.deps("$out", "${out}.d"));
+	if (deps) {
+		push_args(wk, args, toolchain_compiler_deps(wk, comp, "$out", "${out}.d"));
 	}
-	push_args(wk, args, compilers[t].args.output("$out"));
-	push_args(wk, args, compilers[t].args.compile_only());
+
+	push_args(wk, args, toolchain_compiler_debugfile(wk, comp, "$out"));
+
+	push_args(wk, args, toolchain_compiler_output(wk, comp, "$out"));
+	push_args(wk, args, toolchain_compiler_compile_only(wk, comp));
 	obj_array_push(wk, args, make_str(wk, "$in"));
 
 	obj compile_command = join_args_plain(wk, args);
 
-	fprintf(out, "rule %s\n"
+	fprintf(out,
+		"rule %s\n"
 		" command = %s\n",
 		get_cstr(wk, rule_name),
 		get_cstr(wk, compile_command));
-	if (compilers[t].deps) {
+	if (deps) {
 		fprintf(out,
 			" deps = %s\n"
 			" depfile = ${out}.d\n",
 			deps);
 	}
-	fprintf(out,
-		" description = compiling %s $out\n\n",
-		compiler_language_to_s(l));
+	fprintf(out, " description = compiling %s $out\n\n", compiler_language_to_s(l));
 }
 
 static enum iteration_result
-write_compiler_rule_iter(struct workspace *wk, void *_ctx, enum compiler_language l, obj comp_id)
+write_compiler_rule_iter(struct workspace *wk, void *_ctx, obj k, obj comp_id)
 {
+	enum compiler_language l = k;
 	struct write_compiler_rule_ctx *ctx = _ctx;
 
 	obj rule_name;
@@ -172,8 +194,9 @@ ret:
 }
 
 static enum iteration_result
-write_generic_compiler_rule_iter(struct workspace *wk, void *_ctx, enum compiler_language l, obj comp_id)
+write_generic_compiler_rule_iter(struct workspace *wk, void *_ctx, obj k, obj comp_id)
 {
+	enum compiler_language l = k;
 	struct write_compiler_rule_ctx *ctx = _ctx;
 	obj rule_name;
 
@@ -194,26 +217,35 @@ struct name_compiler_rule_ctx {
 };
 
 static enum iteration_result
-name_compiler_rule_iter(struct workspace *wk, void *_ctx, enum compiler_language l, uint32_t count)
+name_compiler_rule_iter(struct workspace *wk, void *_ctx, obj k, uint32_t count)
 {
+	enum compiler_language l = k;
 	struct name_compiler_rule_ctx *ctx = _ctx;
 	bool specialized_rule = count > 2;
 
 	obj rule_name;
+	SBUF(rule_name_buf);
 	if (specialized_rule) {
-		obj name = make_strf(wk, "%s_%s_compiler_for_%s",
+		sbuf_pushf(wk,
+			&rule_name_buf,
+			"%s_%s_compiler_for_%s",
 			get_cstr(wk, ctx->proj->rule_prefix),
 			compiler_language_to_s(l),
-			get_cstr(wk, ctx->tgt->build_name)
-			);
+			get_cstr(wk, ctx->tgt->build_name));
 
+		escape_rule(&rule_name_buf);
+		obj name = sbuf_into_str(wk, &rule_name_buf);
 		uniqify_name(wk, ctx->compiler_rule_arr, name, &rule_name);
 	} else {
 		if (!obj_dict_geti(wk, ctx->generic_rules, l, &rule_name)) {
-			obj name = make_strf(wk, "%s_%s_compiler",
+			sbuf_pushf(wk,
+				&rule_name_buf,
+				"%s_%s_compiler",
 				get_cstr(wk, ctx->proj->rule_prefix),
 				compiler_language_to_s(l));
 
+			escape_rule(&rule_name_buf);
+			obj name = sbuf_into_str(wk, &rule_name_buf);
 			uniqify_name(wk, ctx->compiler_rule_arr, name, &rule_name);
 			obj_dict_seti(wk, ctx->generic_rules, l, rule_name);
 		}
@@ -246,109 +278,33 @@ name_compiler_rule_tgt_iter(struct workspace *wk, void *_ctx, obj tgt_id)
 	return ir_cont;
 }
 
-static enum iteration_result
-add_global_opts_set_from_env_iter(struct workspace *wk, void *_ctx, obj key, obj val)
-{
-	obj regen_args = *(obj *)_ctx;
-
-	struct obj_option *o = get_obj_option(wk, val);
-	if (o->source != option_value_source_environment) {
-		return ir_cont;
-	}
-
-
-	// NOTE: This only handles options of type str or [str], which is okay since
-	// the only options that can be set from the environment are of this
-	// type.
-	// TODO: The current implementation of array stringification would
-	// choke on spaces, etc.
-
-	const char *sval;
-	switch (get_obj_type(wk, o->val)) {
-	case obj_string:
-		sval = get_cstr(wk, o->val);
-		break;
-	case obj_array: {
-		obj joined;
-		obj_array_join(wk, true, o->val, make_str(wk, ","), &joined);
-		sval = get_cstr(wk, joined);
-		break;
-	}
-	default:
-		UNREACHABLE;
-	}
-
-	obj_array_push(wk, regen_args, make_strf(wk, "-D%s=%s", get_cstr(wk, o->name), sval));
-	return ir_cont;
-}
-
 bool
-ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj,
-	bool need_phony,
-	obj compiler_rule_arr)
+ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj, bool need_phony, obj compiler_rule_arr)
 {
 	TracyCZoneAutoS;
+	obj_array_push(wk, wk->backend_output_stack, make_str(wk, "ninja_write_rules"));
+
 	bool res = false;
 
-	fprintf(
-		out,
+	fprintf(out,
 		"# This is the build file for project \"%s\"\n"
 		"# It is autogenerated by the muon build system.\n"
-		"ninja_required_version = 1.7.1\n\n",
-		get_cstr(wk, main_proj->cfg.name)
-		);
+		"ninja_required_version = 1.7.1\n"
+		"builddir = %s\n\n",
+		get_cstr(wk, main_proj->cfg.name),
+		output_path.private_dir);
 
-	fprintf(out,
-		"rule static_linker\n"
-		" command = rm -f $out && ar $LINK_ARGS $out $in\n"
-		" description = linking static $out\n"
-		"\n"
-		"rule CUSTOM_COMMAND\n"
-		" command = $COMMAND\n"
-		" description = $COMMAND\n"
-		" restat = 1\n"
-		"\n"
-		"rule CUSTOM_COMMAND_DEP\n"
-		" command = $COMMAND\n"
-		" description = $COMMAND\n"
-		" deps = gcc\n"
-		" depfile = $DEPFILE\n"
-		" restat = 1\n"
-		"\n"
-		);
-
-	obj regen_args;
-	make_obj(wk, &regen_args, obj_array);
-
-	obj_array_push(wk, regen_args, make_str(wk, wk->argv0));
-	obj_array_push(wk, regen_args, make_str(wk, "-C"));
-	obj_array_push(wk, regen_args, make_str(wk, wk->source_root));
-	obj_array_push(wk, regen_args, make_str(wk, "setup"));
-
-	SBUF(compiler_check_cache_path);
-	path_join(wk, &compiler_check_cache_path,
-		wk->muon_private, output_path.compiler_check_cache);
-
-	obj_array_push(wk, regen_args, make_str(wk, "-c"));
-	obj_array_push(wk, regen_args, make_str(wk, compiler_check_cache_path.buf));
-
-	obj_dict_foreach(wk, wk->global_opts, &regen_args, add_global_opts_set_from_env_iter);
-
-	uint32_t i;
-	for (i = 0; i < wk->original_commandline.argc; ++i) {
-		obj_array_push(wk, regen_args,
-			make_str(wk, wk->original_commandline.argv[i]));
-	}
-
-	obj regen_cmd = join_args_shell(wk, regen_args);
+	obj regen_cmd = join_args_shell(wk, regenerate_build_command(wk, false));
 
 	fprintf(out,
 		"rule REGENERATE_BUILD\n"
-		" command = %s", get_cstr(wk, regen_cmd));
+		" command = %s",
+		get_cstr(wk, regen_cmd));
 
 	fputs("\n description = Regenerating build files.\n"
-		" generator = 1\n"
-		"\n", out);
+	      " generator = 1\n"
+	      "\n",
+		out);
 
 	obj regenerate_deps_rel;
 	{
@@ -360,8 +316,21 @@ ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj,
 	fprintf(out,
 		"build build.ninja: REGENERATE_BUILD %s\n"
 		" pool = console\n\n",
-		get_cstr(wk, join_args_ninja(wk, regenerate_deps_rel))
-		);
+		get_cstr(wk, join_args_ninja(wk, regenerate_deps_rel)));
+
+	fprintf(out,
+		"rule CUSTOM_COMMAND\n"
+		" command = $COMMAND\n"
+		" description = $COMMAND\n"
+		" restat = 1\n"
+		"\n"
+		"rule CUSTOM_COMMAND_DEP\n"
+		" command = $COMMAND\n"
+		" description = $COMMAND\n"
+		" deps = gcc\n"
+		" depfile = $DEPFILE\n"
+		" restat = 1\n"
+		"\n");
 
 	if (need_phony) {
 		fprintf(out, "build build_always_stale: phony\n\n");
@@ -369,8 +338,9 @@ ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj,
 
 	obj rule_prefix_arr;
 	make_obj(wk, &rule_prefix_arr, obj_array);
+	uint32_t i;
 	for (i = 0; i < wk->projects.len; ++i) {
-		struct project *proj = darr_get(&wk->projects, i);
+		struct project *proj = arr_get(&wk->projects, i);
 		if (proj->not_ok) {
 			continue;
 		}
@@ -378,23 +348,10 @@ ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj,
 		TracyCZoneN(tctx_name, "name rules", true);
 
 		{ // determine project rule prefix
-			const char *proj_name = get_cstr(wk, proj->cfg.name);
-			char buf[BUF_SIZE_1k] = { 0 }, *p;
-			strncpy(buf, proj_name, BUF_SIZE_1k - 1);
-			for (p = buf; *p; ++p) {
-				if (*p == '_'
-				    || ('a' <= *p && *p <= 'z')
-				    || ('A' <= *p && *p <= 'Z')
-				    || ('0' <= *p && *p <= '9')) {
-					continue;
-				}
-
-				*p = '_';
-			}
-
-			obj pre = make_str(wk, buf);
-
-			uniqify_name(wk, rule_prefix_arr, pre, &proj->rule_prefix);
+			SBUF(pre);
+			sbuf_pushs(wk, &pre, get_cstr(wk, proj->cfg.name));
+			escape_rule(&pre);
+			uniqify_name(wk, rule_prefix_arr, sbuf_into_str(wk, &pre), &proj->rule_prefix);
 		}
 
 		obj generic_rules;
@@ -443,12 +400,64 @@ ninja_write_rules(FILE *out, struct workspace *wk, struct project *main_proj,
 
 			TracyCZoneEnd(tctx_rules);
 		}
+
+		{ // static linker
+			enum compiler_language static_linker_precedence[] = {
+				compiler_language_c,
+				compiler_language_cpp,
+				compiler_language_objc,
+				compiler_language_objcpp,
+				compiler_language_nasm,
+			};
+
+			obj comp_id = 0;
+			uint32_t j;
+			for (j = 0; j < ARRAY_LEN(static_linker_precedence); ++j) {
+				if (obj_dict_geti(wk, proj->compilers, static_linker_precedence[j], &comp_id)) {
+					break;
+				}
+			}
+
+			if (comp_id) {
+				struct obj_compiler *comp = get_obj_compiler(wk, comp_id);
+
+				obj static_link_args;
+				make_obj(wk, &static_link_args, obj_array);
+
+				// TODO: make this overrideable
+				const enum static_linker_type type = comp->type[toolchain_component_static_linker];
+				if (type == static_linker_ar_posix || type == static_linker_ar_gcc) {
+					obj_array_push(wk, static_link_args, make_str(wk, wk->argv0));
+					obj_array_push(wk, static_link_args, make_str(wk, "internal"));
+					obj_array_push(wk, static_link_args, make_str(wk, "exe"));
+					obj_array_push(wk, static_link_args, make_str(wk, "-R"));
+					obj_array_push(wk, static_link_args, make_str(wk, "$out"));
+					obj_array_push(wk, static_link_args, make_str(wk, "--"));
+				}
+
+				obj_array_extend(wk, static_link_args, comp->static_linker_cmd_arr);
+				push_args(wk, static_link_args, toolchain_static_linker_always(wk, comp));
+				push_args(wk, static_link_args, toolchain_static_linker_base(wk, comp));
+				push_args(wk,
+					static_link_args,
+					toolchain_static_linker_input_output(wk, comp, "$in", "$out"));
+
+				fprintf(out,
+					"rule %s_static_linker\n"
+					" command = %s\n"
+					" description = linking static $out\n"
+					"\n",
+					get_cstr(wk, proj->rule_prefix),
+					get_cstr(wk, join_args_plain(wk, static_link_args)));
+			}
+		}
 	}
 
 	fprintf(out, "# targets\n\n");
 
 	res = true;
 ret:
+	obj_array_pop(wk, wk->backend_output_stack);
 	TracyCZoneAutoE;
 	return res;
 }
