@@ -13,7 +13,7 @@
 #include "tracy.h"
 
 const struct output_path output_path = {
-	.private_dir = "muon-private",
+	.private_dir = ".muon",
 	.summary = "summary.txt",
 	.tests = "tests.dat",
 	.install = "install.dat",
@@ -33,8 +33,7 @@ output_open(const char *dir, const char *name)
 }
 
 bool
-with_open(const char *dir, const char *name, struct workspace *wk,
-	void *ctx, with_open_callback cb)
+with_open(const char *dir, const char *name, struct workspace *wk, void *ctx, with_open_callback cb)
 {
 	TracyCZone(tctx_func, true);
 #ifdef TRACY_ENABLE
@@ -43,6 +42,8 @@ with_open(const char *dir, const char *name, struct workspace *wk,
 	TracyCZoneName(tctx_func, buf, strlen(buf));
 #endif
 
+	obj_array_push(wk, wk->backend_output_stack, make_strf(wk, "writing %s", name));
+
 	bool ret = false;
 	FILE *out;
 	if (!(out = output_open(dir, name))) {
@@ -55,6 +56,8 @@ with_open(const char *dir, const char *name, struct workspace *wk,
 
 	ret = true;
 ret:
+	obj_array_pop(wk, wk->backend_output_stack);
+
 	TracyCZoneEnd(tctx_func);
 	return ret;
 }
