@@ -46,18 +46,15 @@ copy_subdir_iter(void *_ctx, const char *path)
 	obj rel_str = sbuf_into_str(ctx->wk, &rel);
 
 	if (fs_dir_exists(src.buf)) {
-		if (ctx->exclude_directories &&
-		    obj_array_in(ctx->wk, ctx->exclude_directories, rel_str)) {
+		if (ctx->exclude_directories && obj_array_in(ctx->wk, ctx->exclude_directories, rel_str)) {
 			LOG_I("skipping dir '%s'", src.buf);
 			return ir_cont;
 		}
 
 		LOG_I("make dir '%s'", dest.buf);
 
-		if (!fs_dir_exists(dest.buf)) {
-			if (!fs_mkdir(dest.buf)) {
-				return ir_err;
-			}
+		if (!fs_mkdir(dest.buf, true)) {
+			return ir_err;
 		}
 
 		struct copy_subdir_ctx new_ctx = {
@@ -75,8 +72,7 @@ copy_subdir_iter(void *_ctx, const char *path)
 			return ir_err;
 		}
 	} else if (fs_symlink_exists(src.buf) || fs_file_exists(src.buf)) {
-		if (ctx->exclude_files &&
-		    obj_array_in(ctx->wk, ctx->exclude_files, rel_str)) {
+		if (ctx->exclude_files && obj_array_in(ctx->wk, ctx->exclude_files, rel_str)) {
 			LOG_I("skipping file '%s'", src.buf);
 			return ir_cont;
 		}
@@ -112,8 +108,7 @@ install_iter(struct workspace *wk, void *_ctx, obj v_id)
 	struct obj_install_target *in = get_obj_install_target(wk, v_id);
 
 	SBUF(dest_dirname);
-	const char *dest = get_cstr(wk, in->dest),
-	*src = get_cstr(wk, in->src);
+	const char *dest = get_cstr(wk, in->dest), *src = get_cstr(wk, in->src);
 
 	assert(in->type == install_target_symlink || in->type == install_target_emptydir || path_is_absolute(src));
 
@@ -124,20 +119,11 @@ install_iter(struct workspace *wk, void *_ctx, obj v_id)
 	}
 
 	switch (in->type) {
-	case install_target_default:
-		LOG_I("install '%s' -> '%s'", src, dest);
-		break;
-	case install_target_subdir:
-		LOG_I("install subdir '%s' -> '%s'", src, dest);
-		break;
-	case install_target_symlink:
-		LOG_I("install symlink '%s' -> '%s'", dest, src);
-		break;
-	case install_target_emptydir:
-		LOG_I("install emptydir '%s'", dest);
-		break;
-	default:
-		abort();
+	case install_target_default: LOG_I("install '%s' -> '%s'", src, dest); break;
+	case install_target_subdir: LOG_I("install subdir '%s' -> '%s'", src, dest); break;
+	case install_target_symlink: LOG_I("install symlink '%s' -> '%s'", dest, src); break;
+	case install_target_emptydir: LOG_I("install emptydir '%s'", dest); break;
+	default: abort();
 	}
 
 	if (ctx->opts->dry_run) {
@@ -205,8 +191,7 @@ install_iter(struct workspace *wk, void *_ctx, obj v_id)
 			return ir_err;
 		}
 		break;
-	default:
-		abort();
+	default: abort();
 	}
 
 	if (in->has_perm && !fs_chmod(dest, in->perm)) {
@@ -217,10 +202,16 @@ install_iter(struct workspace *wk, void *_ctx, obj v_id)
 }
 
 static enum iteration_result
-install_scripts_iter(struct workspace *wk, void *_ctx, obj v)
+install_scripts_iter(struct workspace *wk, void *_ctx, obj install_script)
 {
 	struct install_ctx *ctx = _ctx;
 
+	obj install_script_dry_run, install_script_cmdline;
+	obj_array_index(wk, install_script, 0, &install_script_dry_run);
+	obj_array_index(wk, install_script, 1, &install_script_cmdline);
+
+	bool script_can_dry_run = get_obj_bool(wk, install_script_dry_run);
+
 	obj env;
 	make_obj(wk, &env, obj_dict);
 	if (ctx->destdir) {
@@ -228,16 +219,19 @@ install_scripts_iter(struct workspace *wk, void *_ctx, obj v)
 	}
 	obj_dict_set(wk, env, make_str(wk, "MESON_INSTALL_PREFIX"), ctx->prefix);
 	obj_dict_set(wk, env, make_str(wk, "MESON_INSTALL_DESTDIR_PREFIX"), ctx->full_prefix);
+	if (ctx->opts->dry_run && script_can_dry_run) {
+		obj_dict_set(wk, env, make_str(wk, "MESON_INSTALL_DRY_RUN"), make_str(wk, "1"));
+	}
 	set_default_environment_vars(wk, env, false);
 
 	const char *argstr, *envstr;
 	uint32_t argc, envc;
 	env_to_envstr(wk, &envstr, &envc, env);
-	join_args_argstr(wk, &argstr, &argc, v);
+	join_args_argstr(wk, &argstr, &argc, install_script_cmdline);
 
 	LOG_I("running install script '%s'", argstr);
 
-	if (ctx->opts->dry_run) {
+	if (ctx->opts->dry_run && !script_can_dry_run) {
 		return ir_cont;
 	}
 
@@ -298,15 +292,14 @@ install_run(struct install_options *opts)
 	obj_array_index(&wk, install, 3, &ctx.prefix);
 
 	SBUF(build_root);
-	path_cwd(&wk, &build_root);
+	path_copy_cwd(&wk, &build_root);
 	wk.build_root = get_cstr(&wk, sbuf_into_str(&wk, &build_root));
 	wk.source_root = get_cstr(&wk, source_root);
 
-	const char *destdir;
-	if ((destdir = getenv("DESTDIR"))) {
+	if ((opts->destdir)) {
 		SBUF(full_prefix);
 		SBUF(abs_destdir);
-		path_make_absolute(&wk, &abs_destdir, destdir);
+		path_make_absolute(&wk, &abs_destdir, opts->destdir);
 		path_join_absolute(&wk, &full_prefix, abs_destdir.buf, get_cstr(&wk, ctx.prefix));
 
 		ctx.full_prefix = sbuf_into_str(&wk, &full_prefix);
