@@ -5,14 +5,13 @@
 
 #include "compat.h"
 
-#include <errno.h>
 #include <string.h>
-#include <time.h>
 
 #include "args.h"
 #include "backend/ninja.h"
 #include "backend/output.h"
 #include "cmd_test.h"
+#include "embedded.h"
 #include "error.h"
 #include "formats/tap.h"
 #include "functions/environment.h"
@@ -20,21 +19,27 @@
 #include "log.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
+#include "platform/os.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
 #include "platform/term.h"
+#include "platform/timer.h"
+#include "util.h"
+
+#define SLEEP_TIME 10000000 // 10ms
 
 enum test_result_status {
 	test_result_status_running,
 	test_result_status_ok,
 	test_result_status_failed,
 	test_result_status_timedout,
+	test_result_status_skipped,
 };
 
 struct test_result {
 	struct run_cmd_ctx cmd_ctx;
 	struct obj_test *test;
-	struct timespec start;
+	struct timer t;
 	float dur, timeout;
 	enum test_result_status status;
 	bool busy;
@@ -54,7 +59,8 @@ struct run_test_ctx {
 		uint32_t test_i, test_len, error_count;
 		uint32_t total_count, total_error_count, total_expect_fail_count;
 		uint32_t total_skipped;
-		uint32_t term_width;
+		uint32_t term_width, term_height;
+		uint32_t prev_jobs_displayed;
 		bool term;
 		bool ran_tests;
 	} stats;
@@ -66,7 +72,8 @@ struct run_test_ctx {
 		float timeout_multiplier;
 	} setup;
 
-	struct darr test_results;
+	struct arr test_results;
+	struct arr jobs_sorted;
 
 	struct test_result *jobs;
 	uint32_t busy_jobs;
@@ -81,45 +88,50 @@ static const char *
 test_category_label(enum test_category cat)
 {
 	switch (cat) {
-	case test_category_test:
-		return "test";
-	case test_category_benchmark:
-		return "benchmark";
-	default:
-		UNREACHABLE_RETURN;
+	case test_category_test: return "test";
+	case test_category_benchmark: return "benchmark";
+	default: UNREACHABLE_RETURN;
 	}
 }
 
-static void
-print_test_result(struct workspace *wk, const struct test_result *res)
+static const char *
+test_suites_label(struct workspace *wk, const struct test_result *res)
 {
-	const char *name = get_cstr(wk, res->test->name);
+	const char *suite_str = 0;
+	uint32_t suites_len = 0;
+	if (res->test->suites) {
+		suites_len = get_obj_array(wk, res->test->suites)->len;
+		if (suites_len == 1) {
+			obj s;
+			obj_array_index(wk, res->test->suites, 0, &s);
+			suite_str = get_cstr(wk, s);
+		} else if (suites_len > 1) {
+			obj s;
+			obj_array_join(wk, true, res->test->suites, make_str(wk, "+"), &s);
+			suite_str = get_cstr(wk, s);
+		}
+	}
+	return suite_str;
+}
 
-	enum {
-		status_failed,
-		status_should_have_failed,
-		status_ok,
-		status_failed_ok,
-		status_running,
-		status_timedout,
-	} status = status_ok;
+enum test_detailed_status {
+	status_ok,
+	status_failed,
+	status_failed_ok,
+	status_running,
+	status_timedout,
+	status_skipped,
+	status_should_have_failed,
+};
 
-	const char *status_msg[] = {
-		[status_failed]             = "fail ",
-		[status_should_have_failed] = "ok*  ",
-		[status_ok]                 = "ok   ",
-		[status_failed_ok]          = "fail*",
-		[status_running]            = "start",
-		[status_timedout]           = "timeout",
-	};
+static enum test_detailed_status
+test_detailed_status(const struct test_result *res)
+{
+	enum test_detailed_status status = status_ok;
 
 	switch (res->status) {
-	case test_result_status_running:
-		status = status_running;
-		break;
-	case test_result_status_timedout:
-		status = status_timedout;
-		break;
+	case test_result_status_running: status = status_running; break;
+	case test_result_status_timedout: status = status_timedout; break;
 	case test_result_status_failed:
 		if (res->test->should_fail) {
 			status = status_should_have_failed;
@@ -134,22 +146,30 @@ print_test_result(struct workspace *wk, const struct test_result *res)
 			status = status_ok;
 		}
 		break;
+	case test_result_status_skipped: status = status_skipped; break;
 	}
 
-	const char *suite_str = NULL;
-	uint32_t suites_len = 0;
-	if (res->test->suites) {
-		suites_len = get_obj_array(wk, res->test->suites)->len;
-		if (suites_len == 1) {
-			obj s;
-			obj_array_index(wk, res->test->suites, 0, &s);
-			suite_str = get_cstr(wk, s);
-		} else if (suites_len > 1) {
-			obj s;
-			obj_array_join(wk, true, res->test->suites, make_str(wk, ", "), &s);
-			suite_str = get_cstr(wk, s);
-		}
-	}
+	return status;
+}
+
+static void
+print_test_result(struct workspace *wk, const struct test_result *res)
+{
+	const char *name = get_cstr(wk, res->test->name);
+
+	const char *status_msg[] = {
+		[status_failed] = "fail ",
+		[status_should_have_failed] = "ok*  ",
+		[status_ok] = "ok   ",
+		[status_failed_ok] = "fail*",
+		[status_running] = "start",
+		[status_timedout] = "timeout",
+		[status_skipped] = "skip ",
+	};
+
+	enum test_detailed_status status = test_detailed_status(res);
+
+	const char *suite_str = test_suites_label(wk, res);
 
 	if (log_clr()) {
 		uint32_t clr[] = {
@@ -159,16 +179,17 @@ print_test_result(struct workspace *wk, const struct test_result *res)
 			[status_failed_ok] = 33,
 			[status_running] = 0,
 			[status_timedout] = 31,
+			[status_skipped] = 33,
 		};
-		log_plain("[\033[%dm%s\033[0m]", clr[status], status_msg[status]);
+		log_plain("\033[%dm%s\033[0m", clr[status], status_msg[status]);
 	} else {
-		log_plain("[%s]", status_msg[status]);
+		log_plain("%s", status_msg[status]);
 	}
 
 	if (res->status == test_result_status_running) {
 		log_plain("          ");
 	} else {
-		log_plain(" %6.2fs, ", res->dur);
+		log_plain(" %6.2fs ", res->dur);
 	}
 
 	if (res->subtests.have) {
@@ -176,11 +197,7 @@ print_test_result(struct workspace *wk, const struct test_result *res)
 	}
 
 	if (suite_str) {
-		if (suites_len > 1) {
-			log_plain("[%s]:", suite_str);
-		} else {
-			log_plain("%s:", suite_str);
-		}
+		log_plain("%s:", suite_str);
 	}
 
 	log_plain("%s", name);
@@ -190,31 +207,48 @@ print_test_result(struct workspace *wk, const struct test_result *res)
 	}
 }
 
+static int32_t
+sort_jobs(const void *_a, const void *_b, void *_ctx)
+{
+	const uint32_t *a = _a, *b = _b;
+	struct run_test_ctx *ctx = _ctx;
+
+	const struct test_result *res_a = &ctx->jobs[*a], *res_b = &ctx->jobs[*b];
+
+	if (res_a->dur > res_b->dur) {
+		return -1;
+	} else if (res_a->dur < res_b->dur) {
+		return 1;
+	} else {
+		return 0;
+	}
+}
+
 static void
 print_test_progress(struct workspace *wk, struct run_test_ctx *ctx, const struct test_result *res, bool write_line)
 {
-	if (res->status != test_result_status_running) {
+	if (res && res->status != test_result_status_running) {
 		++ctx->stats.total_count;
 		++ctx->stats.test_i;
-		if (res->status != test_result_status_ok) {
+		switch (res->status) {
+		case test_result_status_running: UNREACHABLE; break;
+		case test_result_status_ok:
+		case test_result_status_skipped: break;
+		case test_result_status_failed:
+		case test_result_status_timedout:
 			++ctx->stats.total_error_count;
 			++ctx->stats.error_count;
+			break;
 		}
 	}
 
 	if (!ctx->stats.term && !ctx->opts->verbosity) {
-		if (res->status != test_result_status_running) {
+		if (res && res->status != test_result_status_running) {
 			char c;
 			switch (res->status) {
-			case test_result_status_failed:
-				c = 'E';
-				break;
-			case test_result_status_timedout:
-				c = 'T';
-				break;
-			default:
-				c = '.';
-				break;
+			case test_result_status_failed: c = 'E'; break;
+			case test_result_status_timedout: c = 'T'; break;
+			default: c = '.'; break;
 			}
 
 			log_plain("%c", c);
@@ -241,29 +275,114 @@ print_test_progress(struct workspace *wk, struct run_test_ctx *ctx, const struct
 	uint32_t i, pad = 2;
 
 	char info[BUF_SIZE_4k];
-	pad += snprintf(info, BUF_SIZE_4k, "%d/%d f: %d (%d) ", ctx->stats.test_i, ctx->stats.test_len, ctx->stats.error_count, ctx->busy_jobs);
+	pad += snprintf(info,
+		BUF_SIZE_4k,
+		"%d/%d f:%d s:%d j:%d ",
+		ctx->stats.test_i,
+		ctx->stats.test_len,
+		ctx->stats.error_count,
+		ctx->stats.total_skipped,
+		ctx->busy_jobs);
 
 	log_plain("%s[", info);
-	uint32_t pct = (float)(ctx->stats.test_i) * (float)(ctx->stats.term_width - pad) / (float)ctx->stats.test_len;
+	const float pct_scale = (float)(ctx->stats.term_width - pad) / (float)ctx->stats.test_len;
+	uint32_t pct_done = (float)(ctx->stats.test_i) * pct_scale;
+	uint32_t pct_working = (float)(ctx->stats.test_i + ctx->busy_jobs) * pct_scale;
+
 	for (i = 0; i < ctx->stats.term_width - pad; ++i) {
-		if (i < pct) {
+		if (i <= pct_done) {
 			log_plain("=");
-		} else if (i == pct) {
+		} else if (i < pct_working) {
+			log_plain("-");
+		} else if (i == pct_working) {
 			log_plain(">");
 		} else {
 			log_plain(" ");
 		}
 	}
 	log_plain("]");
+
+	log_plain("\n");
+
+	arr_sort(&ctx->jobs_sorted, ctx, sort_jobs);
+
+	const uint32_t term_height = ctx->stats.term_height - 2;
+	const uint32_t max_jobs_to_display = MIN(term_height, ctx->opts->jobs);
+	uint32_t jobs_displayed = 0;
+
+	for (i = 0; i < ctx->opts->jobs; ++i) {
+		res = &ctx->jobs[*(uint32_t *)arr_get(&ctx->jobs_sorted, i)];
+
+		if (!res->busy) {
+			continue;
+		}
+
+		struct {
+			bool have;
+			struct tap_parse_result result;
+		} tap = { 0 };
+
+		if (res->test->protocol == test_protocol_tap) {
+			if (res->cmd_ctx.out.buf) {
+				tap_parse(res->cmd_ctx.out.buf, res->cmd_ctx.out.len, &tap.result);
+				tap.have = true;
+			}
+		}
+
+		log_plain("%6.2fs %s", res->dur, get_cstr(wk, res->test->name));
+
+		if (tap.have) {
+			log_plain(" (%d/", tap.result.pass + tap.result.fail + tap.result.skip);
+			if (tap.result.have_plan) {
+				log_plain("%d", tap.result.total);
+			} else {
+				log_plain("?");
+			}
+
+			if (tap.result.fail) {
+				log_plain(" f:%d", tap.result.fail);
+			}
+
+			if (tap.result.skip) {
+				log_plain(" s:%d", tap.result.skip);
+			}
+
+			log_plain(")");
+		}
+
+		log_plain("\033[K\n");
+
+		++jobs_displayed;
+		if (jobs_displayed >= max_jobs_to_display) {
+			break;
+		}
+	}
+	for (i = jobs_displayed; i < ctx->stats.prev_jobs_displayed; ++i) {
+		log_plain("\033[K\n");
+	}
+
+	log_plain("\033[%dA", MAX(jobs_displayed, ctx->stats.prev_jobs_displayed) + 1);
+
+	ctx->stats.prev_jobs_displayed = jobs_displayed;
 }
 
 /*
  * test setup / suites
  */
 
+/*
+ * Matching rules used for -e (setup) and -s (suite).  name1 is user-provided.
+ *
+ * If name1 is fully qualified with a project name, it will match the main
+ * project only.
+ *
+ * e.g.
+ * name == main_proj:name
+ * name != sub_proj:name
+ * sub_proj:name == sub_proj:name
+ */
 static bool
-project_namespaced_name_matches(const char *name1, bool proj2_is_main,
-	const struct str *proj2, const struct str *name2)
+project_namespaced_name_matches(const char *name1, bool proj2_is_main, const struct str *proj2, const struct str *name2)
 {
 	struct str proj1 = { 0 };
 	const char *sep;
@@ -302,8 +421,9 @@ test_in_suite_iter(struct workspace *wk, void *_ctx, obj s)
 
 	for (i = 0; i < opts->suites_len; ++i) {
 		if (!project_namespaced_name_matches(opts->suites[i],
-			ctx->run_test_ctx->proj_i == 0,
-			get_str(wk, ctx->run_test_ctx->proj_name), get_str(wk, s))) {
+			    ctx->run_test_ctx->proj_i == 0,
+			    get_str(wk, ctx->run_test_ctx->proj_name),
+			    get_str(wk, s))) {
 			continue;
 		}
 
@@ -320,9 +440,9 @@ test_in_exclude_suites_exclude_suites_iter(struct workspace *wk, void *_ctx, obj
 	struct test_in_suite_ctx *ctx = _ctx;
 
 	if (project_namespaced_name_matches(get_cstr(wk, exclude),
-		ctx->run_test_ctx->proj_i == 0,
-		get_str(wk, ctx->run_test_ctx->proj_name), get_str(wk, ctx->suite))) {
-
+		    ctx->run_test_ctx->proj_i == 0,
+		    get_str(wk, ctx->run_test_ctx->proj_name),
+		    get_str(wk, ctx->suite))) {
 		ctx->found = true;
 		return ir_done;
 	}
@@ -336,8 +456,7 @@ test_in_exclude_suites_iter(struct workspace *wk, void *_ctx, obj suite)
 	struct test_in_suite_ctx *ctx = _ctx;
 	ctx->suite = suite;
 
-	obj_array_foreach(wk, ctx->run_test_ctx->setup.exclude_suites,
-		ctx, test_in_exclude_suites_exclude_suites_iter);
+	obj_array_foreach(wk, ctx->run_test_ctx->setup.exclude_suites, ctx, test_in_exclude_suites_exclude_suites_iter);
 
 	if (ctx->found) {
 		return ir_done;
@@ -354,7 +473,6 @@ test_in_suite(struct workspace *wk, obj suites, struct run_test_ctx *run_test_ct
 
 	if (!run_test_ctx->opts->suites_len) {
 		// no suites given on command line
-
 		if (run_test_ctx->setup.exclude_suites) {
 			obj_array_foreach(wk, suites, &ctx, test_in_exclude_suites_iter);
 			return !ctx.found;
@@ -392,8 +510,9 @@ find_test_setup_iter(struct workspace *wk, void *_ctx, obj arr)
 
 	if (ctx->rtctx->opts->setup) {
 		if (!project_namespaced_name_matches(ctx->rtctx->opts->setup,
-			ctx->rtctx->proj_i == 0,
-			get_str(wk, ctx->rtctx->proj_name), get_str(wk, name))) {
+			    ctx->rtctx->proj_i == 0,
+			    get_str(wk, ctx->rtctx->proj_name),
+			    get_str(wk, name))) {
 			return ir_cont;
 		}
 	} else if (!is_default || !get_obj_bool(wk, is_default)) {
@@ -409,9 +528,7 @@ find_test_setup_iter(struct workspace *wk, void *_ctx, obj arr)
 	ctx->rtctx->setup.env = env;
 	ctx->rtctx->setup.exclude_suites = exclude_suites;
 	ctx->rtctx->setup.wrapper = exe_wrapper;
-	ctx->rtctx->setup.timeout_multiplier =
-		timeout_multiplier ? get_obj_number(wk, timeout_multiplier)
-			: 1.0f;
+	ctx->rtctx->setup.timeout_multiplier = timeout_multiplier ? get_obj_number(wk, timeout_multiplier) : 1.0f;
 	ctx->found = true;
 	return ir_done;
 }
@@ -466,33 +583,28 @@ ret:
  * Test runner
  */
 
-static void
-calculate_test_duration(struct test_result *res)
+static enum test_result_status
+check_test_result_exitcode(struct workspace *wk, const struct run_test_ctx *ctx, const struct test_result *res)
 {
-	struct timespec end;
+	if (res->cmd_ctx.status == 0) {
+		return test_result_status_ok;
+	} else if (res->cmd_ctx.status == 77) {
+		return test_result_status_skipped;
+	} else if (res->cmd_ctx.status == 99) {
+		return test_result_status_failed;
+	} else {
+		return test_result_status_failed;
+	}
+}
 
-	if (clock_gettime(CLOCK_MONOTONIC, &end)) {
-		LOG_E("error getting test end time: %s", strerror(errno));
-		return;
+static enum test_result_status
+check_test_result_tap(struct workspace *wk, const struct run_test_ctx *ctx, struct test_result *res)
+{
+	enum test_result_status exitcode_status = check_test_result_exitcode(wk, ctx, res);
+	if (exitcode_status != test_result_status_ok) {
+		return exitcode_status;
 	}
 
-	double secs = (double)end.tv_sec - (double)res->start.tv_sec;
-	double ns = ((secs * 1000000000.0) + end.tv_nsec) - res->start.tv_nsec;
-	res->dur = ns / 1000000000.0;
-}
-
-static void
-test_delay(void)
-{
-	struct timespec req = {
-		.tv_nsec = 10000000,
-	};
-	nanosleep(&req, NULL);
-}
-
-static bool
-check_test_result_tap(struct workspace *wk, struct run_test_ctx *ctx, struct test_result *res)
-{
 	struct tap_parse_result tap_result = { 0 };
 	tap_parse(res->cmd_ctx.out.buf, res->cmd_ctx.out.len, &tap_result);
 
@@ -500,22 +612,7 @@ check_test_result_tap(struct workspace *wk, struct run_test_ctx *ctx, struct tes
 	res->subtests.pass = tap_result.pass + tap_result.skip;
 	res->subtests.total = tap_result.total;
 
-	return tap_result.all_ok && res->status == 0;
-}
-
-static bool
-check_test_result_exitcode(struct workspace *wk, struct run_test_ctx *ctx, struct test_result *res)
-{
-	if (res->cmd_ctx.status == 0) {
-		return true;
-	} else if (res->cmd_ctx.status == 77) {
-		++ctx->stats.total_skipped;
-		return true;
-	} else if (res->cmd_ctx.status == 99) {
-		return false;
-	} else {
-		return false;
-	}
+	return tap_result.all_ok && res->status == 0 ? test_result_status_ok : test_result_status_failed;
 }
 
 static void
@@ -523,21 +620,24 @@ collect_tests(struct workspace *wk, struct run_test_ctx *ctx)
 {
 	uint32_t i;
 
+	if (ctx->stats.term) {
+		print_test_progress(wk, ctx, 0, false);
+	}
+
 	for (i = 0; i < ctx->opts->jobs; ++i) {
-		if (!ctx->jobs[i].busy) {
+		struct test_result *res = &ctx->jobs[i];
+
+		if (!res->busy) {
 			continue;
 		}
 
-		struct test_result *res = &ctx->jobs[i];
-		calculate_test_duration(res);
+		res->dur = timer_read(&res->t);
 
 		enum run_cmd_state state = run_cmd_collect(&res->cmd_ctx);
 
-		if (state != run_cmd_running
-		    && res->status == test_result_status_timedout) {
-			run_cmd_ctx_destroy(&res->cmd_ctx);
+		if (state != run_cmd_running && res->status == test_result_status_timedout) {
 			print_test_progress(wk, ctx, res, true);
-			darr_push(&ctx->test_results, res);
+			arr_push(&ctx->test_results, res);
 			goto free_slot;
 		}
 
@@ -559,37 +659,36 @@ collect_tests(struct workspace *wk, struct run_test_ctx *ctx)
 		case run_cmd_error:
 			res->status = test_result_status_failed;
 			print_test_progress(wk, ctx, res, true);
-			darr_push(&ctx->test_results, res);
+			arr_push(&ctx->test_results, res);
 			break;
 		case run_cmd_finished: {
-			bool ok;
+			enum test_result_status status;
 
 			switch (res->test->protocol) {
-			case test_protocol_tap:
-				ok = check_test_result_tap(wk, ctx, res);
-				break;
-			default:
-				ok = check_test_result_exitcode(wk, ctx, res);
-				break;
+			case test_protocol_tap: status = check_test_result_tap(wk, ctx, res); break;
+			default: status = check_test_result_exitcode(wk, ctx, res); break;
 			}
 
-			if (!ok && res->test->should_fail) {
-				ok = true;
+			if (status == test_result_status_skipped) {
+				++ctx->stats.total_skipped;
 			}
 
-			if (ok) {
+			if (status == test_result_status_failed && res->test->should_fail) {
+				status = test_result_status_ok;
+			}
+
+			if (status != test_result_status_failed) {
 				if (res->test->should_fail) {
 					++ctx->stats.total_expect_fail_count;
 				}
 
-				res->status = test_result_status_ok;
-				run_cmd_ctx_destroy(&res->cmd_ctx);
+				res->status = status;
 			} else {
 				res->status = test_result_status_failed;
 			}
 
 			print_test_progress(wk, ctx, res, true);
-			darr_push(&ctx->test_results, res);
+			arr_push(&ctx->test_results, res);
 			break;
 		}
 		}
@@ -602,6 +701,8 @@ free_slot:
 			ctx->serial = false;
 		}
 
+		*res = (struct test_result){ 0 };
+
 		if (ctx->opts->fail_fast && ctx->stats.total_error_count) {
 			break;
 		}
@@ -609,8 +710,13 @@ free_slot:
 }
 
 static void
-push_test(struct workspace *wk, struct run_test_ctx *ctx, struct obj_test *test,
-	const char *argstr, uint32_t argc, const char *envstr, uint32_t envc)
+push_test(struct workspace *wk,
+	struct run_test_ctx *ctx,
+	struct obj_test *test,
+	const char *argstr,
+	uint32_t argc,
+	const char *envstr,
+	uint32_t envc)
 {
 	uint32_t i;
 	while (true) {
@@ -633,7 +739,7 @@ push_test(struct workspace *wk, struct run_test_ctx *ctx, struct obj_test *test,
 		}
 
 cont:
-		test_delay();
+		timer_sleep(SLEEP_TIME);
 		collect_tests(wk, ctx);
 	}
 found_slot:
@@ -654,16 +760,18 @@ found_slot:
 	};
 
 	if (ctx->opts->verbosity > 1) {
-		cmd_ctx->flags |= run_cmd_ctx_flag_dont_capture;
+		if (res->test->protocol == test_protocol_tap) {
+			cmd_ctx->flags |= run_cmd_ctx_flag_tee;
+		} else {
+			cmd_ctx->flags |= run_cmd_ctx_flag_dont_capture;
+		}
 	}
 
 	if (test->workdir) {
 		cmd_ctx->chdir = get_cstr(wk, test->workdir);
 	}
 
-	if (clock_gettime(CLOCK_MONOTONIC, &res->start)) {
-		LOG_E("error getting test start time: %s", strerror(errno));
-	}
+	timer_start(&res->t);
 
 	print_test_progress(wk, ctx, res, ctx->serial);
 
@@ -671,10 +779,10 @@ found_slot:
 		res->busy = false;
 		--ctx->busy_jobs;
 
-		calculate_test_duration(res);
+		res->dur = timer_read(&res->t);
 		res->status = test_result_status_failed;
 		print_test_progress(wk, ctx, res, true);
-		darr_push(&ctx->test_results, res);
+		arr_push(&ctx->test_results, res);
 	}
 }
 
@@ -731,14 +839,55 @@ run_test(struct workspace *wk, void *_ctx, obj t)
  * Test filtering and dispatch
  */
 
+static bool
+test_matches_cmdline_tests(struct workspace *wk, struct run_test_ctx *ctx, struct obj_test *test)
+{
+	if (!ctx->opts->tests_len) {
+		return true;
+	}
+
+	uint32_t i;
+	for (i = 0; i < ctx->opts->tests_len; ++i) {
+		const char *testspec = ctx->opts->tests[i];
+		const char *sep = strchr(testspec, ':');
+
+		struct str proj = { 0 }, name = { 0 };
+
+		if (sep) {
+			proj.s = testspec;
+			proj.len = sep - testspec;
+			name.s = sep + 1;
+			name.len = strlen(name.s);
+		} else {
+			name.s = testspec;
+			name.len = strlen(name.s);
+		}
+
+		if (proj.len && !str_eql(&proj, get_str(wk, ctx->proj_name))) {
+			continue;
+		}
+
+		if (name.len && !str_eql_glob(&name, get_str(wk, test->name))) {
+			continue;
+		}
+
+		return true;
+	}
+
+	return false;
+}
+
 static enum iteration_result
 gather_project_tests_iter(struct workspace *wk, void *_ctx, obj val)
 {
 	struct run_test_ctx *ctx = _ctx;
 	struct obj_test *t = get_obj_test(wk, val);
 
-	if (!(t->category == ctx->opts->cat
-	      && test_in_suite(wk, t->suites, ctx))) {
+	if (!(t->category == ctx->opts->cat && test_in_suite(wk, t->suites, ctx))) {
+		return ir_cont;
+	}
+
+	if (!test_matches_cmdline_tests(wk, ctx, t)) {
 		return ir_cont;
 	}
 
@@ -754,8 +903,7 @@ gather_project_tests_iter(struct workspace *wk, void *_ctx, obj val)
 static int32_t
 test_compare(struct workspace *wk, void *_ctx, obj t1_id, obj t2_id)
 {
-	struct obj_test *t1 = get_obj_test(wk, t1_id),
-			*t2 = get_obj_test(wk, t2_id);
+	struct obj_test *t1 = get_obj_test(wk, t1_id), *t2 = get_obj_test(wk, t2_id);
 
 	int64_t p1 = t1->priority ? get_obj_number(wk, t1->priority) : 0,
 		p2 = t2->priority ? get_obj_number(wk, t2->priority) : 0;
@@ -813,7 +961,7 @@ run_project_tests(struct workspace *wk, void *_ctx, obj proj_name, obj arr)
 	if (get_obj_array(wk, ctx->deps)->len && !ctx->opts->no_rebuild) {
 		obj ninja_cmd;
 		obj_array_dedup(wk, ctx->deps, &ninja_cmd);
-		if (ninja_run(wk, ninja_cmd, NULL, NULL) != 0) {
+		if (!ninja_run(wk, ninja_cmd, NULL, NULL)) {
 			LOG_W("failed to run ninja");
 		}
 	}
@@ -833,7 +981,7 @@ run_project_tests(struct workspace *wk, void *_ctx, obj proj_name, obj arr)
 	}
 
 	while (ctx->busy_jobs) {
-		test_delay();
+		timer_sleep(SLEEP_TIME);
 		collect_tests(wk, ctx);
 	}
 
@@ -844,6 +992,150 @@ run_project_tests(struct workspace *wk, void *_ctx, obj proj_name, obj arr)
 	return ir_cont;
 }
 
+static bool
+tests_output_term(struct workspace *wk, struct run_test_ctx *ctx)
+{
+	bool ret = true;
+
+	uint32_t i;
+	for (i = 0; i < ctx->test_results.len; ++i) {
+		struct test_result *res = arr_get(&ctx->test_results, i);
+
+		if (ctx->opts->print_summary
+			|| (res->status == test_result_status_failed || res->status == test_result_status_timedout)) {
+			print_test_result(wk, res);
+			if (res->status == test_result_status_failed && res->cmd_ctx.err_msg) {
+				log_plain(": %s", res->cmd_ctx.err_msg);
+			}
+			log_plain("\n");
+		}
+
+		if (res->status == test_result_status_failed) {
+			if (res->test->should_fail) {
+				ret = false;
+			} else {
+				ret = false;
+				if (res->cmd_ctx.out.len) {
+					log_plain("stdout: '%s'\n", res->cmd_ctx.out.buf);
+				}
+				if (res->cmd_ctx.err.len) {
+					log_plain("stderr: '%s'\n", res->cmd_ctx.err.buf);
+				}
+			}
+		} else if (res->status == test_result_status_timedout) {
+			ret = false;
+		}
+	}
+
+	return ret;
+}
+
+static void
+tests_as_json(struct workspace *wk, struct run_test_ctx *ctx, struct sbuf *data)
+{
+	const char *status_str[] = {
+		[status_failed] = "failed",
+		[status_should_have_failed] = "should_have_failed",
+		[status_ok] = "ok",
+		[status_failed_ok] = "failed_ok",
+		[status_running] = "running",
+		[status_timedout] = "timedout",
+		[status_skipped] = "skipped",
+	};
+
+	sbuf_push(wk, data, '{');
+	sbuf_pushf(wk, data, "\"project\":{\"name\":\"%s\"},", get_cstr(wk, ctx->proj_name));
+	sbuf_pushf(wk, data, "\"tests\":[");
+
+	uint32_t i;
+	for (i = 0; i < ctx->test_results.len; ++i) {
+		struct test_result *res = arr_get(&ctx->test_results, i);
+		enum test_detailed_status status = test_detailed_status(res);
+		const char *suite_str = test_suites_label(wk, res);
+
+		sbuf_pushf(wk,
+			data,
+			"{"
+			"\"status\":\"%s\","
+			"\"name\":\"%s\","
+			"\"suite\":\"%s\","
+			"\"duration\":%f,",
+			status_str[status],
+			get_cstr(wk, res->test->name),
+			suite_str,
+			res->dur);
+
+		if (res->subtests.have) {
+			sbuf_pushf(wk,
+				data,
+				"\"subtests\":{\"pass\":%d,\"total\":%d},",
+				res->subtests.pass,
+				res->subtests.total);
+		}
+
+		sbuf_pushs(wk, data, "\"stdout\":\"");
+		sbuf_push_json_escaped(wk, data, res->cmd_ctx.out.buf, res->cmd_ctx.out.len);
+		sbuf_pushs(wk, data, "\",\"stderr\":\"");
+		sbuf_push_json_escaped(wk, data, res->cmd_ctx.err.buf, res->cmd_ctx.err.len);
+		sbuf_pushs(wk, data, "\"");
+
+		sbuf_pushs(wk, data, "}");
+
+		if (i + 1 != ctx->test_results.len) {
+			sbuf_push(wk, data, ',');
+		}
+	}
+	sbuf_pushf(wk, data, "]}");
+}
+
+static bool
+tests_output_html(struct workspace *wk, struct run_test_ctx *ctx)
+{
+	SBUF(data);
+
+	tests_as_json(wk, ctx, &data);
+
+	SBUF(html_path);
+	path_join(wk, &html_path, output_path.private_dir, "tests.html");
+	SBUF(abs);
+	path_make_absolute(wk, &abs, html_path.buf);
+
+	FILE *f = 0;
+	if (!(f = fs_fopen(abs.buf, "wb"))) {
+		return false;
+	}
+
+	fprintf(f, embedded_get("html/test_out.html"), data.buf);
+	LOG_I("wrote html output to %s", abs.buf);
+	fclose(f);
+
+	return true;
+}
+
+static bool
+tests_output_json(struct workspace *wk, struct run_test_ctx *ctx)
+{
+	SBUF(data);
+
+	tests_as_json(wk, ctx, &data);
+
+	SBUF(json_path);
+	path_join(wk, &json_path, output_path.private_dir, "tests.json");
+	SBUF(abs);
+	path_make_absolute(wk, &abs, json_path.buf);
+
+	FILE *f = 0;
+	if (!(f = fs_fopen(abs.buf, "wb"))) {
+		return false;
+	}
+
+	fprintf(f, "%s", data.buf);
+	LOG_I("wrote json output to %s", abs.buf);
+	fclose(f);
+
+	return true;
+}
+
 bool
 tests_run(struct test_options *opts, const char *argv0)
 {
@@ -855,7 +1147,7 @@ tests_run(struct test_options *opts, const char *argv0)
 	wk.argv0 = argv0;
 
 	if (!opts->jobs) {
-		opts->jobs = 4;
+		opts->jobs = os_parallel_job_count();
 	}
 
 	struct run_test_ctx ctx = {
@@ -863,7 +1155,11 @@ tests_run(struct test_options *opts, const char *argv0)
 		.setup = { .timeout_multiplier = 1.0f, },
 	};
 
-	darr_init(&ctx.test_results, 32, sizeof(struct test_result));
+	arr_init(&ctx.test_results, 32, sizeof(struct test_result));
+	arr_init(&ctx.jobs_sorted, ctx.opts->jobs, sizeof(uint32_t));
+	for (uint32_t i = 0; i < ctx.opts->jobs; ++i) {
+		arr_push(&ctx.jobs_sorted, &i);
+	}
 	ctx.jobs = z_calloc(ctx.opts->jobs, sizeof(struct test_result));
 
 	{ // load global opts
@@ -882,22 +1178,21 @@ tests_run(struct test_options *opts, const char *argv0)
 	}
 
 	{
-		int fd;
-		if (!fs_fileno(log_file(), &fd)) {
+		int term_fd;
+		if (!fs_fileno(log_file(), &term_fd)) {
 			return false;
 		}
 
 		if (opts->display == test_display_auto) {
 			opts->display = test_display_dots;
-			if (fs_is_a_tty_from_fd(fd)) {
+			if (fs_is_a_tty_from_fd(term_fd)) {
 				opts->display = test_display_bar;
 			}
 		}
 
 		if (opts->display == test_display_bar) {
-			uint32_t h;
 			ctx.stats.term = true;
-			term_winsize(fd, &h, &ctx.stats.term_width);
+			term_winsize(term_fd, &ctx.stats.term_height, &ctx.stats.term_width);
 		} else if (opts->display == test_display_dots) {
 			ctx.stats.term = false;
 		} else {
@@ -931,48 +1226,36 @@ tests_run(struct test_options *opts, const char *argv0)
 			test_category_label(opts->cat),
 			ctx.stats.total_expect_fail_count,
 			ctx.stats.total_error_count,
-			ctx.stats.total_skipped
-			);
+			ctx.stats.total_skipped);
 	}
 
-	ret = true;
-	uint32_t i;
-	for (i = 0; i < ctx.test_results.len; ++i) {
-		struct test_result *res = darr_get(&ctx.test_results, i);
+	switch (opts->output) {
+	case test_output_term: ret = tests_output_term(&wk, &ctx); break;
+	case test_output_html: ret = tests_output_html(&wk, &ctx); break;
+	case test_output_json: ret = tests_output_json(&wk, &ctx); break;
+	}
 
-		if (opts->print_summary
-		    || (res->status == test_result_status_failed
-			|| res->status == test_result_status_timedout)) {
-			print_test_result(&wk, res);
-			if (res->status == test_result_status_failed && res->cmd_ctx.err_msg) {
-				log_plain(": %s", res->cmd_ctx.err_msg);
-			}
-			log_plain("\n");
-		}
-
-		if (res->status == test_result_status_failed) {
-			if (res->test->should_fail) {
-				ret = false;
-			} else {
-				ret = false;
-				if (res->cmd_ctx.out.len) {
-					log_plain("stdout: '%s'\n", res->cmd_ctx.out.buf);
+	{
+		uint32_t i;
+		for (i = 0; i < ctx.test_results.len; ++i) {
+			struct test_result *res = arr_get(&ctx.test_results, i);
+			if (res->status == test_result_status_failed) {
+				if (res->test->should_fail) {
+					ret = false;
+				} else {
+					ret = false;
 				}
-				if (res->cmd_ctx.err.len) {
-					log_plain("stderr: '%s'\n", res->cmd_ctx.err.buf);
-				}
-
+			} else if (res->status == test_result_status_timedout) {
+				ret = false;
 			}
-
 			run_cmd_ctx_destroy(&res->cmd_ctx);
-		} else if (res->status == test_result_status_timedout) {
-			ret = false;
 		}
 	}
 
 ret:
 	workspace_destroy_bare(&wk);
-	darr_destroy(&ctx.test_results);
+	arr_destroy(&ctx.test_results);
+	arr_destroy(&ctx.jobs_sorted);
 	z_free(ctx.jobs);
 	return ret;
 }
