@@ -5,43 +5,42 @@
 
 #include "compat.h"
 
-#include "functions/common.h"
 #include "functions/array.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/typecheck.h"
 #include "log.h"
 
 static bool
-func_array_length(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_array_length(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
 	make_obj(wk, res, obj_number);
-	set_obj_number(wk, *res, get_obj_array(wk, rcvr)->len);
+	set_obj_number(wk, *res, get_obj_array(wk, self)->len);
 	return true;
 }
 
 static bool
-func_array_get(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_array_get(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { obj_number }, ARG_TYPE_NULL };
-	struct args_norm ao[] = { { tc_any }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, ao, NULL)) {
+	struct args_norm an[] = { { obj_number }, { tc_any, .optional = true }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	int64_t i = get_obj_number(wk, an[0].val);
 
-	if (!bounds_adjust(wk, get_obj_array(wk, rcvr)->len, &i)) {
-		if (ao[0].set) {
-			*res = ao[0].val;
+	if (!bounds_adjust(get_obj_array(wk, self)->len, &i)) {
+		if (an[1].set) {
+			*res = an[1].val;
 		} else {
-			interp_error(wk, an[0].node, "index out of bounds");
+			vm_error_at(wk, an[0].node, "index out of bounds");
 			return false;
 		}
 	} else {
-		obj_array_index(wk, rcvr, i, res);
+		obj_array_index(wk, self, i, res);
 	}
 
 	return true;
@@ -73,15 +72,15 @@ array_contains_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 static bool
-func_array_contains(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_array_contains(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { tc_any }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	struct array_contains_ctx ctx = { .item = an[0].val };
-	obj_array_foreach(wk, rcvr, &ctx, array_contains_iter);
+	obj_array_foreach(wk, self, &ctx, array_contains_iter);
 
 	make_obj(wk, res, obj_bool);
 	set_obj_bool(wk, *res, ctx.found);
@@ -89,33 +88,36 @@ func_array_contains(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res
 }
 
 static bool
-func_array_delete(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_array_delete(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { tc_number }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	int64_t idx = get_obj_number(wk, an[0].val);
-	if (!boundscheck(wk, an[0].node, get_obj_array(wk, rcvr)->len, &idx)) {
+	if (!boundscheck(wk, an[0].node, get_obj_array(wk, self)->len, &idx)) {
 		return false;
 	}
 
-	obj_array_del(wk, rcvr, idx);
+	obj_array_del(wk, self, idx);
 	return true;
 }
 
-const struct func_impl_name impl_tbl_array[] = {
+const struct func_impl impl_tbl_array[] = {
 	{ "length", func_array_length, tc_number, true },
 	{ "get", func_array_get, tc_any, true },
 	{ "contains", func_array_contains, tc_bool, true },
 	{ NULL, NULL },
 };
 
-const struct func_impl_name impl_tbl_array_internal[] = {
+const struct func_impl impl_tbl_array_internal[] = {
 	{ "length", func_array_length, tc_number, true },
 	{ "get", func_array_get, tc_any, true },
 	{ "contains", func_array_contains, tc_bool, true },
-	{ "delete", func_array_delete, },
+	{
+		"delete",
+		func_array_delete,
+	},
 	{ NULL, NULL },
 };
