@@ -8,17 +8,22 @@
 
 #include <string.h>
 
+#include "args.h"
 #include "coerce.h"
 #include "error.h"
 #include "functions/build_target.h"
-#include "functions/common.h"
 #include "functions/generator.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "platform/path.h"
 
 bool
-tgt_src_to_object_path(struct workspace *wk, const struct obj_build_target *tgt, obj src_file, bool relative, struct sbuf *res)
+tgt_src_to_object_path(struct workspace *wk,
+	const struct obj_build_target *tgt,
+	obj src_file,
+	bool relative,
+	struct sbuf *res)
 {
 	obj src = *get_obj_file(wk, src_file);
 
@@ -55,36 +60,48 @@ tgt_src_to_object_path(struct workspace *wk, const struct obj_build_target *tgt,
 
 		uint32_t i;
 		for (i = 0; i < rel.len; ++i) {
-			if (rel.buf[i] == PATH_SEP) {
+			if (rel.buf[i] == PATH_SEP || rel.buf[i] == ':') {
 				rel.buf[i] = '_';
 			}
 		}
 	}
 
 	path_join(wk, res, private_path, rel.buf);
-	sbuf_pushs(wk, res, ".o");
+
+	const char *ext = ".o";
+
+	{
+		enum compiler_language lang;
+		obj comp_id;
+		if (filename_to_compiler_language(res->buf, &lang)
+			&& obj_dict_geti(wk, current_project(wk)->compilers, lang, &comp_id)) {
+			ext = toolchain_compiler_object_ext(wk, get_obj_compiler(wk, comp_id))->args[0];
+		}
+	}
+
+	sbuf_pushs(wk, res, ext);
 	return true;
 }
 
 static bool
-func_build_target_name(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_name(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = get_obj_build_target(wk, rcvr)->name;
+	*res = get_obj_build_target(wk, self)->name;
 	return true;
 }
 
 static bool
-func_build_target_full_path(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_full_path(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	struct obj_build_target *tgt = get_obj_build_target(wk, rcvr);
+	struct obj_build_target *tgt = get_obj_build_target(wk, self);
 
 	*res = tgt->build_path;
 	return true;
@@ -104,8 +121,7 @@ build_target_extract_objects_iter(struct workspace *wk, void *_ctx, obj val)
 	obj file;
 	enum obj_type t = get_obj_type(wk, val);
 
-	if (!typecheck(wk, ctx->err_node, val,
-		tc_file | tc_string | tc_custom_target | tc_generated_list)) {
+	if (!typecheck(wk, ctx->err_node, val, tc_file | tc_string | tc_custom_target | tc_generated_list)) {
 		return false;
 	}
 
@@ -116,13 +132,11 @@ build_target_extract_objects_iter(struct workspace *wk, void *_ctx, obj val)
 		}
 		break;
 	}
-	case obj_file:
-		file = val;
-		break;
+	case obj_file: file = val; break;
 	case obj_custom_target: {
 		struct obj_custom_target *tgt = get_obj_custom_target(wk, val);
 		if (!obj_array_flatten_one(wk, tgt->output, &file)) {
-			interp_error(wk, ctx->err_node, "cannot coerce custom_target with multiple outputs to file");
+			vm_error_at(wk, ctx->err_node, "cannot coerce custom_target with multiple outputs to file");
 			return ir_err;
 		}
 		break;
@@ -138,11 +152,9 @@ build_target_extract_objects_iter(struct workspace *wk, void *_ctx, obj val)
 		}
 		return ir_cont;
 	}
-	default:
-		UNREACHABLE_RETURN;
+	default: UNREACHABLE_RETURN;
 	}
 
-
 	enum compiler_language l;
 	if (!filename_to_compiler_language(get_file_path(wk, file), &l)) {
 		return ir_cont;
@@ -159,15 +171,14 @@ build_target_extract_objects_iter(struct workspace *wk, void *_ctx, obj val)
 	case compiler_language_c:
 	case compiler_language_cpp:
 	case compiler_language_llvm_ir:
-	case compiler_language_objc:
-		break;
+	case compiler_language_objcpp:
+	case compiler_language_objc: break;
 	case compiler_language_null:
-	case compiler_language_count:
-		UNREACHABLE;
+	case compiler_language_count: UNREACHABLE;
 	}
 
 	if (!obj_array_in(wk, ctx->tgt->src, file)) {
-		interp_error(wk, ctx->err_node, "%o is not in target sources (%o)", file, ctx->tgt->src);
+		vm_error_at(wk, ctx->err_node, "%o is not in target sources (%o)", file, ctx->tgt->src);
 		return ir_err;
 	}
 
@@ -184,29 +195,30 @@ build_target_extract_objects_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 static bool
-build_target_extract_objects(struct workspace *wk, obj rcvr, uint32_t err_node, obj *res, obj arr)
+build_target_extract_objects(struct workspace *wk, obj self, uint32_t err_node, obj *res, obj arr)
 {
 	make_obj(wk, res, obj_array);
 
 	struct build_target_extract_objects_ctx ctx = {
 		.err_node = err_node,
 		.res = res,
-		.tgt = get_obj_build_target(wk, rcvr),
-		.tgt_id = rcvr,
+		.tgt = get_obj_build_target(wk, self),
+		.tgt_id = self,
 	};
 
 	return obj_array_foreach_flat(wk, arr, &ctx, build_target_extract_objects_iter);
 }
 
 static bool
-func_build_target_extract_objects(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_extract_objects(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_string | tc_file | tc_custom_target | tc_generated_list }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	struct args_norm an[]
+		= { { TYPE_TAG_GLOB | tc_string | tc_file | tc_custom_target | tc_generated_list }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	return build_target_extract_objects(wk, rcvr, an[0].node, res, an[0].val);
+	return build_target_extract_objects(wk, self, an[0].node, res, an[0].val);
 }
 
 static enum iteration_result
@@ -218,15 +230,14 @@ build_target_extract_all_objects_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 bool
-build_target_extract_all_objects(struct workspace *wk, uint32_t err_node, obj rcvr, obj *res, bool recursive)
+build_target_extract_all_objects(struct workspace *wk, uint32_t ip, obj self, obj *res, bool recursive)
 {
 	make_obj(wk, res, obj_array);
 
 	struct build_target_extract_objects_ctx ctx = {
-		.err_node = err_node,
 		.res = res,
-		.tgt = get_obj_build_target(wk, rcvr),
-		.tgt_id = rcvr,
+		.tgt = get_obj_build_target(wk, self),
+		.tgt_id = self,
 	};
 
 	if (!obj_array_foreach_flat(wk, ctx.tgt->src, &ctx, build_target_extract_all_objects_iter)) {
@@ -241,44 +252,39 @@ build_target_extract_all_objects(struct workspace *wk, uint32_t err_node, obj rc
 }
 
 static bool
-func_build_target_extract_all_objects(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_extract_all_objects(struct workspace *wk, obj self, obj *res)
 {
 	enum kwargs {
 		kw_recursive,
 	};
-	struct args_kw akw[] = {
-		[kw_recursive] = { "recursive", obj_bool },
-		0
-	};
-	if (!interp_args(wk, args_node, NULL, NULL, akw)) {
+	struct args_kw akw[] = { [kw_recursive] = { "recursive", obj_bool }, 0 };
+	if (!pop_args(wk, NULL, akw)) {
 		return false;
 	}
 
-	bool recursive = akw[kw_recursive].set
-		? get_obj_bool(wk, akw[kw_recursive].val)
-		: false;
+	bool recursive = akw[kw_recursive].set ? get_obj_bool(wk, akw[kw_recursive].val) : false;
 
-	return build_target_extract_all_objects(wk, args_node, rcvr, res, recursive);
+	return build_target_extract_all_objects(wk, 0, self, res, recursive);
 }
 
 static bool
-func_build_target_private_dir_include(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_private_dir_include(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
 	make_obj(wk, res, obj_include_directory);
 	struct obj_include_directory *inc = get_obj_include_directory(wk, *res);
 
-	inc->path = get_obj_build_target(wk, rcvr)->private_path;
+	inc->path = get_obj_build_target(wk, self)->private_path;
 	return true;
 }
 
 static bool
-func_build_target_found(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_build_target_found(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -287,7 +293,7 @@ func_build_target_found(struct workspace *wk, obj rcvr, uint32_t args_node, obj
 	return true;
 }
 
-const struct func_impl_name impl_tbl_build_target[] = {
+const struct func_impl impl_tbl_build_target[] = {
 	{ "extract_all_objects", func_build_target_extract_all_objects, tc_array },
 	{ "extract_objects", func_build_target_extract_objects, tc_array },
 	{ "found", func_build_target_found, tc_bool },
