@@ -7,13 +7,13 @@
 
 #include <string.h>
 
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/configuration_data.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 
 static bool
-func_configuration_data_set_quoted(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_set_quoted(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, { obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -24,31 +24,31 @@ func_configuration_data_set_quoted(struct workspace *wk, obj rcvr, uint32_t args
 		0
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
-	obj dict = get_obj_configuration_data(wk, rcvr)->dict;
+	obj dict = get_obj_configuration_data(wk, self)->dict;
 
 	const char *s = get_cstr(wk, an[1].val);
 	obj str = make_str(wk, "\"");
 
 	for (; *s; ++s) {
 		if (*s == '"') {
-			str_app(wk, str, "\\");
+			str_app(wk, &str, "\\");
 		}
 
-		str_appn(wk, str, s, 1);
+		str_appn(wk, &str, s, 1);
 	}
 
-	str_app(wk, str, "\"");
+	str_app(wk, &str, "\"");
 
 	obj_dict_set(wk, dict, an[0].val, str);
 	return true;
 }
 
 static bool
-func_configuration_data_set(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_set(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, { tc_string | tc_number | tc_bool }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -58,11 +58,11 @@ func_configuration_data_set(struct workspace *wk, obj rcvr, uint32_t args_node,
 		[kw_description] = { "description", obj_string, },
 		0
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
-	obj dict = get_obj_configuration_data(wk, rcvr)->dict;
+	obj dict = get_obj_configuration_data(wk, self)->dict;
 
 	obj_dict_set(wk, dict, an[0].val, an[1].val);
 
@@ -70,7 +70,7 @@ func_configuration_data_set(struct workspace *wk, obj rcvr, uint32_t args_node,
 }
 
 static bool
-func_configuration_data_set10(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_set10(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, { obj_bool }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -81,11 +81,11 @@ func_configuration_data_set10(struct workspace *wk, obj rcvr, uint32_t args_node
 		0
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
-	obj dict = get_obj_configuration_data(wk, rcvr)->dict;
+	obj dict = get_obj_configuration_data(wk, self)->dict;
 
 	obj n;
 	make_obj(wk, &n, obj_number);
@@ -96,8 +96,7 @@ func_configuration_data_set10(struct workspace *wk, obj rcvr, uint32_t args_node
 }
 
 static bool
-configuration_data_get(struct workspace *wk, uint32_t err_node, obj conf,
-	obj key, obj def, obj *res)
+configuration_data_get(struct workspace *wk, uint32_t err_node, obj conf, obj key, obj def, obj *res)
 {
 	obj dict = get_obj_configuration_data(wk, conf)->dict;
 
@@ -105,7 +104,7 @@ configuration_data_get(struct workspace *wk, uint32_t err_node, obj conf,
 		if (def) {
 			*res = def;
 		} else {
-			interp_error(wk, err_node, "key '%s' not found", get_cstr(wk, key));
+			vm_error_at(wk, err_node, "key '%s' not found", get_cstr(wk, key));
 			return false;
 		}
 	}
@@ -114,30 +113,28 @@ configuration_data_get(struct workspace *wk, uint32_t err_node, obj conf,
 }
 
 static bool
-func_configuration_data_get(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_get(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	struct args_norm ao[] = { { tc_any }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string }, { tc_any, .optional = true }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, ao, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	return configuration_data_get(wk, an[0].node, rcvr, an[0].val, ao[0].val, res);
+	return configuration_data_get(wk, an[0].node, self, an[0].val, an[1].val, res);
 }
 
 static bool
-func_configuration_data_get_unquoted(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_get_unquoted(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	struct args_norm ao[] = { { tc_any }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string }, { tc_any, .optional = true }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, ao, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	obj v;
-	if (!configuration_data_get(wk, an[0].node, rcvr, an[0].val, ao[0].val, &v)) {
+	if (!configuration_data_get(wk, an[0].node, self, an[0].val, an[1].val, &v)) {
 		return false;
 	}
 
@@ -164,13 +161,13 @@ obj_dict_keys_iter(struct workspace *wk, void *_ctx, obj k, obj _v)
 }
 
 static bool
-func_configuration_data_keys(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_keys(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	obj dict = get_obj_configuration_data(wk, rcvr)->dict;
+	obj dict = get_obj_configuration_data(wk, self)->dict;
 
 	make_obj(wk, res, obj_array);
 	obj_dict_foreach(wk, dict, res, obj_dict_keys_iter);
@@ -178,36 +175,35 @@ func_configuration_data_keys(struct workspace *wk, obj rcvr, uint32_t args_node,
 }
 
 static bool
-func_configuration_data_has(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_has(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	obj _, dict = get_obj_configuration_data(wk, rcvr)->dict;
+	obj _, dict = get_obj_configuration_data(wk, self)->dict;
 	make_obj(wk, res, obj_bool);
 	set_obj_bool(wk, *res, obj_dict_index(wk, dict, an[0].val, &_));
 	return true;
 }
 
 static bool
-func_configuration_data_merge_from(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_configuration_data_merge_from(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_configuration_data }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	obj_dict_merge_nodup(wk, get_obj_configuration_data(wk, rcvr)->dict,
-		get_obj_configuration_data(wk, an[0].val)->dict
-		);
+	obj_dict_merge_nodup(
+		wk, get_obj_configuration_data(wk, self)->dict, get_obj_configuration_data(wk, an[0].val)->dict);
 	return true;
 }
 
-const struct func_impl_name impl_tbl_configuration_data[] = {
+const struct func_impl impl_tbl_configuration_data[] = {
 	{ "get", func_configuration_data_get, tc_any },
 	{ "get_unquoted", func_configuration_data_get_unquoted, tc_any },
 	{ "has", func_configuration_data_has, tc_bool },
