@@ -5,10 +5,10 @@
 
 #include "compat.h"
 
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/custom_target.h"
 #include "functions/file.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 
 bool
@@ -27,26 +27,26 @@ custom_target_is_linkable(struct workspace *wk, obj ct)
 }
 
 static bool
-func_custom_target_to_list(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_custom_target_to_list(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = get_obj_custom_target(wk, rcvr)->output;
+	*res = get_obj_custom_target(wk, self)->output;
 	return true;
 }
 
 static bool
-func_custom_target_full_path(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_custom_target_full_path(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
 	obj elem;
-	if (!obj_array_flatten_one(wk, get_obj_custom_target(wk, rcvr)->output, &elem)) {
-		interp_error(wk, args_node, "this custom_target has multiple outputs");
+	if (!obj_array_flatten_one(wk, get_obj_custom_target(wk, self)->output, &elem)) {
+		vm_error(wk, "this custom_target has multiple outputs");
 		return false;
 	}
 
@@ -54,7 +54,7 @@ func_custom_target_full_path(struct workspace *wk, obj rcvr, uint32_t args_node,
 	return true;
 }
 
-const struct func_impl_name impl_tbl_custom_target[] = {
+const struct func_impl impl_tbl_custom_target[] = {
 	{ "full_path", func_custom_target_full_path, tc_string },
 	{ "to_list", func_custom_target_to_list, tc_array },
 	{ NULL, NULL },
