@@ -5,15 +5,15 @@
 
 #include "compat.h"
 
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/disabler.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 
 static bool
-func_disabler_found(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_disabler_found(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -22,7 +22,7 @@ func_disabler_found(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res
 	return true;
 }
 
-const struct func_impl_name impl_tbl_disabler[] = {
+const struct func_impl impl_tbl_disabler[] = {
 	{ "found", func_disabler_found, tc_bool },
 	{ NULL, NULL },
 };
