@@ -8,9 +8,10 @@
 #include <stdlib.h>
 
 #include "error.h"
-#include "functions/common.h"
 #include "functions/environment.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/object_iterators.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "platform/path.h"
 
@@ -51,8 +52,7 @@ evironment_to_dict_iter(struct workspace *wk, void *_ctx, obj action)
 	case environment_set_mode_prepend:
 		str = make_strf(wk, "%s%s%s", get_cstr(wk, val), get_cstr(wk, sep), oval);
 		break;
-	default:
-		UNREACHABLE;
+	default: UNREACHABLE;
 	}
 
 	obj_dict_set(wk, env, key, str);
@@ -76,20 +76,21 @@ static void
 environment_or_dict_set(struct workspace *wk, obj env, const char *key, const char *val)
 {
 	switch (get_obj_type(wk, env)) {
-	case obj_dict:
-		obj_dict_set(wk, env, make_str(wk, key), make_str(wk, val));
-		break;
+	case obj_dict: obj_dict_set(wk, env, make_str(wk, key), make_str(wk, val)); break;
 	case obj_environment:
 		environment_set(wk, env, environment_set_mode_set, make_str(wk, key), make_str(wk, val), 0);
 		break;
-	default:
-		UNREACHABLE;
+	default: UNREACHABLE;
 	}
 }
 
 void
 set_default_environment_vars(struct workspace *wk, obj env, bool set_subdir)
 {
+	if (wk->vm.lang_mode == language_internal) {
+		return;
+	}
+
 	if (wk->argv0) {
 		// argv0 may not be set, e.g. during `muon install`
 		environment_or_dict_set(wk, env, "MUON_PATH", wk->argv0);
@@ -134,49 +135,82 @@ environment_set(struct workspace *wk, obj env, enum environment_set_mode mode, o
 }
 
 static bool
-func_environment_set_common(struct workspace *wk, obj rcvr, uint32_t args_node, enum environment_set_mode mode)
+func_environment_set_common(struct workspace *wk, obj self, enum environment_set_mode mode)
 {
-	struct args_norm an[] = { { obj_string }, { ARG_TYPE_GLOB | obj_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string }, { TYPE_TAG_GLOB | obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_separator,
 	};
 	struct args_kw akw[] = {
 		[kw_separator] = { "separator", obj_string },
-		0
+		0,
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	if (!get_obj_array(wk, an[1].val)->len) {
-		interp_error(wk, an[1].node, "you must pass at least one value");
+		vm_error_at(wk, an[1].node, "you must pass at least one value");
 		return false;
 	}
 
-	return environment_set(wk, rcvr, mode, an[0].val, an[1].val, akw[kw_separator].val);
+	return environment_set(wk, self, mode, an[0].val, an[1].val, akw[kw_separator].val);
 }
 
 static bool
-func_environment_set(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_environment_set(struct workspace *wk, obj self, obj *res)
 {
-	return func_environment_set_common(wk, rcvr, args_node, environment_set_mode_set);
+	return func_environment_set_common(wk, self, environment_set_mode_set);
 }
 
 static bool
-func_environment_append(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_environment_append(struct workspace *wk, obj self, obj *res)
 {
-	return func_environment_set_common(wk, rcvr, args_node, environment_set_mode_append);
+	return func_environment_set_common(wk, self, environment_set_mode_append);
 }
 
 static bool
-func_environment_prepend(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_environment_prepend(struct workspace *wk, obj self, obj *res)
 {
-	return func_environment_set_common(wk, rcvr, args_node, environment_set_mode_prepend);
+	return func_environment_set_common(wk, self, environment_set_mode_prepend);
 }
 
-const struct func_impl_name impl_tbl_environment[] = {
+static bool
+func_environment_unset(struct workspace *wk, obj self, obj *res)
+{
+	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, 0)) {
+		return false;
+	}
+
+	obj to_delete, action, key, actions = get_obj_environment(wk, self)->actions;
+
+	make_obj(wk, &to_delete, obj_array);
+
+	uint32_t i = 0;
+	obj_array_for(wk, actions, action) {
+		/* obj_array_index(wk, action, 0, &mode_num); */
+		obj_array_index(wk, action, 1, &key);
+		/* obj_array_index(wk, action, 2, &val); */
+		/* obj_array_index(wk, action, 3, &sep); */
+
+		if (obj_equal(wk, action, an[0].val)) {
+			obj_array_push(wk, to_delete, i + 1);
+		}
+
+		++i;
+	}
+
+	obj_array_for(wk, to_delete, i) {
+		obj_array_del(wk, actions, i - 1);
+	}
+	return true;
+}
+
+const struct func_impl impl_tbl_environment[] = {
 	{ "set", func_environment_set },
 	{ "append", func_environment_append },
 	{ "prepend", func_environment_prepend },
+	{ "unset", func_environment_unset },
 	{ NULL, NULL },
 };
