@@ -7,9 +7,9 @@
 
 #include <string.h>
 
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/file.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 
 bool
@@ -29,17 +29,17 @@ file_is_linkable(struct workspace *wk, obj file)
 }
 
 static bool
-func_file_full_path(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_file_full_path(struct workspace *wk, obj self, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
-	*res = *get_obj_file(wk, rcvr);
+	*res = *get_obj_file(wk, self);
 	return true;
 }
 
-const struct func_impl_name impl_tbl_file[] = {
+const struct func_impl impl_tbl_file[] = {
 	{ "full_path", func_file_full_path, tc_string },
 	{ NULL, NULL },
 };
