@@ -13,12 +13,13 @@
 #include "buf_size.h"
 #include "coerce.h"
 #include "error.h"
-#include "functions/common.h"
 #include "functions/environment.h"
 #include "functions/kernel/configure_file.h"
 #include "functions/kernel/custom_target.h"
 #include "install.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/object_iterators.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
@@ -28,6 +29,7 @@
 enum configure_file_output_format {
 	configure_file_output_format_c,
 	configure_file_output_format_nasm,
+	configure_file_output_format_json,
 };
 
 static bool
@@ -61,13 +63,9 @@ configure_var_len(const char *p)
 	uint32_t i = 0;
 
 	// Only allow (a-z, A-Z, 0-9, _, -) as valid characters for a define
-	while (p[i] &&
-	       (('a' <= p[i] && p[i] <= 'z')
-		|| ('A' <= p[i] && p[i] <= 'Z')
-		|| ('0' <= p[i] && p[i] <= '9')
-		|| '_' == p[i]
-		|| '-' == p[i]
-	       )) {
+	while (p[i]
+		&& (('a' <= p[i] && p[i] <= 'z') || ('A' <= p[i] && p[i] <= 'Z') || ('0' <= p[i] && p[i] <= '9')
+			|| '_' == p[i] || '-' == p[i])) {
 		++i;
 	}
 
@@ -82,7 +80,12 @@ enum configure_file_syntax {
 };
 
 static bool
-substitute_config(struct workspace *wk, uint32_t dict, uint32_t in_node, const char *in, obj out, enum configure_file_syntax syntax)
+substitute_config(struct workspace *wk,
+	uint32_t dict,
+	uint32_t in_node,
+	const char *in,
+	obj out,
+	enum configure_file_syntax syntax)
 {
 	const char *define;
 	if (syntax & configure_file_syntax_cmakedefine) {
@@ -100,8 +103,7 @@ substitute_config(struct workspace *wk, uint32_t dict, uint32_t in_node, const c
 		varstart = "@";
 		varend = '@';
 	}
-	const uint32_t define_len = strlen(define),
-		       varstart_len = strlen(varstart);
+	const uint32_t define_len = strlen(define), varstart_len = strlen(varstart);
 
 	bool ret = true;
 	struct source src;
@@ -113,25 +115,24 @@ substitute_config(struct workspace *wk, uint32_t dict, uint32_t in_node, const c
 
 	SBUF_manual(out_buf);
 
-	uint32_t i, id_start, id_len,
-		 line = 1, start_of_line = 0, id_start_col = 0, id_start_line = 0;
+	struct source_location location = { 0, 1 }, id_location;
+	uint32_t i, id_start, id_len, col = 0, orig_i;
 	obj elem;
 	char tmp_buf[BUF_SIZE_1k] = { 0 };
 
 	for (i = 0; i < src.len; ++i) {
 		if (src.src[i] == '\n') {
-			start_of_line = i + 1;
-			++line;
+			col = i + 1;
 		}
 
-		if (i == start_of_line && strncmp(&src.src[i], define, define_len) == 0) {
+		if (i == col && strncmp(&src.src[i], define, define_len) == 0) {
 			i += define_len;
 
 			configure_file_skip_whitespace(&src, &i);
 
 			id_start = i;
-			id_start_line = line;
-			id_start_col = i - start_of_line + 1;
+			id_location = location;
+			++id_location.off;
 			id_len = configure_var_len(&src.src[id_start]);
 			i += id_len;
 
@@ -142,8 +143,8 @@ substitute_config(struct workspace *wk, uint32_t dict, uint32_t in_node, const c
 			if (!(src.src[i] == '\n' || src.src[i] == 0)) {
 				if (syntax & configure_file_syntax_cmakedefine) {
 					if (src.src[i] == '@'
-					    && strncmp(&src.src[i + 1], &src.src[id_start], id_len) == 0
-					    && src.src[i + 1 + id_len] == '@') {
+						&& strncmp(&src.src[i + 1], &src.src[id_start], id_len) == 0
+						&& src.src[i + 1 + id_len] == '@') {
 						i += 2 + id_len;
 						configure_file_skip_whitespace(&src, &i);
 
@@ -152,27 +153,32 @@ substitute_config(struct workspace *wk, uint32_t dict, uint32_t in_node, const c
 						}
 					} else {
 extraneous_cmake_chars:
-						error_messagef(&src, id_start_line, i - start_of_line + 1, log_error,
-							"cmakedefine only supports two modes:\n"
-							"\t--> #cmakedefine %.*s\n "
-							"\t--> #cmakedefine %.*s @%.*s@\n"
-							"In particular, no value other than \"%.*s\" will be substituted and\n"
-							"no extra characters (comments, etc.) are allowed on the line.",
-							id_len, &src.src[id_start],
-							id_len, &src.src[id_start],
-							id_len, &src.src[id_start],
-							id_len, &src.src[id_start]
-							);
-						return false;
+						orig_i = i;
+
+						while (src.src[i] && src.src[i] != '\n') {
+							++i;
+						}
+
+						/* id_location.col = orig_i - location.col + 1; */
+						error_messagef(&src,
+							id_location,
+							log_warn,
+							"ignoring trailing characters (%.*s) in cmakedefine",
+							i - orig_i,
+							&src.src[orig_i]);
 					}
 				} else {
-					error_messagef(&src, id_start_line, i - start_of_line + 1, log_error, "expected exactly one token on mesondefine line");
+					/* id_location.col = i - location.col + 1; */
+					error_messagef(&src,
+						id_location,
+						log_error,
+						"expected exactly one token on mesondefine line");
 					return false;
 				}
 			}
 
 			if (i == id_start) {
-				error_messagef(&src, id_start_line, id_start_col, log_error, "key of zero length not supported");
+				error_messagef(&src, id_location, log_error, "key of zero length not supported");
 				return false;
 			} else if (!obj_dict_index_strn(wk, dict, &src.src[id_start], id_len, &elem)) {
 				deftype = "/* undef";
@@ -196,7 +202,9 @@ extraneous_cmake_chars:
 				sub = tmp_buf;
 				break;
 			default:
-				error_messagef(&src, id_start_line, id_start_col, log_error,
+				error_messagef(&src,
+					id_location,
+					log_error,
 					"invalid type for %s: '%s'",
 					define,
 					obj_type_to_s(get_obj_type(wk, elem)));
@@ -216,10 +224,13 @@ write_mesondefine:
 		} else if (src.src[i] == '\\') {
 			/* cope with weird config file escaping rules :(
 			 *
-			 * - Backslashes not directly preceeding a format character are not modified.
-			 * - The number of backslashes preceding varstart in the
-			 *   output is equal to the number of backslashes in
-			 *   the input divided by two, rounding down.
+			 * - Backslashes not directly preceeding a format character are not
+			 *   modified.
+			 * - The number of backslashes preceding varstart in the output is
+			 *   equal to the number of backslashes in the input divided by
+			 *   two, rounding down.
+			 * - If mode is cmake and the number of backslashes is even, don't
+			 *   escape the variable, otherwise always escape the variable.
 			 */
 
 			uint32_t j, output_backslashes;
@@ -230,11 +241,17 @@ write_mesondefine:
 
 			if (strncmp(&src.src[i + j], varstart, varstart_len) == 0) {
 				output_backslashes = j / 2;
-				if ((j & 1) != 0) {
+
+				if (*varstart == '@') {
 					output_format_char = true;
 					i += j;
 				} else {
-					i += j - 1;
+					if ((j & 1) != 0) {
+						output_format_char = true;
+						i += j;
+					} else {
+						i += j - 1;
+					}
 				}
 			} else {
 				i += j - 1;
@@ -252,9 +269,9 @@ write_mesondefine:
 			}
 		} else if (strncmp(&src.src[i], varstart, varstart_len) == 0) {
 			i += varstart_len;
-			id_start_line = line;
 			id_start = i;
-			id_start_col = id_start - start_of_line + 1;
+			id_location = location;
+			/* id_location.col = id_start - location.col + 1; */
 			i += configure_var_len(&src.src[id_start]);
 
 			if (src.src[i] != varend) {
@@ -264,16 +281,17 @@ write_mesondefine:
 			}
 
 			if (i <= id_start) {
-				error_messagef(&src, id_start_line, id_start_col, log_error, "key of zero length not supported");
-				return false;
+				// This means we got a key of length zero
+				sbuf_pushs(wk, &out_buf, "@@");
+				continue;
 			} else if (!obj_dict_index_strn(wk, dict, &src.src[id_start], i - id_start, &elem)) {
-				error_messagef(&src, id_start_line, id_start_col, log_error, "key not found in configuration data");
+				error_messagef(&src, id_location, log_error, "key not found in configuration data");
 				return false;
 			}
 
 			obj sub;
 			if (!coerce_string(wk, in_node, elem, &sub)) {
-				error_messagef(&src, id_start_line, id_start_col, log_error, "unable to substitute value");
+				error_messagef(&src, id_location, log_error, "unable to substitute value");
 				return false;
 			}
 
@@ -304,79 +322,132 @@ cleanup:
 	return ret;
 }
 
-struct generate_config_ctx {
-	struct sbuf *out_buf;
-	uint32_t node;
-	enum configure_file_output_format output_format;
-};
-
-static enum iteration_result
-generate_config_iter(struct workspace *wk, void *_ctx, obj key, obj val)
-{
-	struct generate_config_ctx *ctx = _ctx;
-	enum obj_type t = get_obj_type(wk, val);
-
-	char define_prefix = (char[]){
-		[configure_file_output_format_c] = '#',
-		[configure_file_output_format_nasm] = '%',
-	}[ctx->output_format];
-
-	switch (t) {
-	case obj_string:
-		/* conf_data.set('FOO', '"string"') => #define FOO "string" */
-		/* conf_data.set('FOO', 'a_token')  => #define FOO a_token */
-		sbuf_pushf(wk, ctx->out_buf, "%cdefine %s %s\n", define_prefix, get_cstr(wk, key), get_cstr(wk, val));
-		break;
-	case obj_bool:
-		/* conf_data.set('FOO', true)       => #define FOO */
-		/* conf_data.set('FOO', false)      => #undef FOO */
-		if (get_obj_bool(wk, val)) {
-			sbuf_pushf(wk, ctx->out_buf, "%cdefine %s\n", define_prefix, get_cstr(wk, key));
-		} else {
-			sbuf_pushf(wk, ctx->out_buf, "%cundef %s\n", define_prefix, get_cstr(wk, key));
-		}
-		break;
-	case obj_number:
-		/* conf_data.set('FOO', 1)          => #define FOO 1 */
-		/* conf_data.set('FOO', 0)          => #define FOO 0 */
-		sbuf_pushf(wk, ctx->out_buf, "%cdefine %s %" PRId64 "\n", define_prefix, get_cstr(wk, key), get_obj_number(wk, val));
-		break;
-	default:
-		interp_error(wk, ctx->node, "invalid type for config data value: '%s'", obj_type_to_s(t));
-		return ir_err;
-	}
-
-	return ir_cont;
-}
-
 static bool
-generate_config(struct workspace *wk, enum configure_file_output_format format,
-	obj dict, uint32_t node, obj out_path)
+generate_config(struct workspace *wk,
+	enum configure_file_output_format format,
+	obj macro_name,
+	obj dict,
+	uint32_t node,
+	obj out_path)
 {
 	SBUF_manual(out_buf);
 
-	struct generate_config_ctx ctx = {
-		.out_buf = &out_buf,
-		.node = node,
-		.output_format = format,
-	};
+	if (macro_name) {
+		sbuf_pushf(
+			wk, &out_buf, "#ifndef %s\n#define %s\n", get_cstr(wk, macro_name), get_cstr(wk, macro_name));
+	} else if (format == configure_file_output_format_json) {
+		sbuf_push(wk, &out_buf, '{');
+	}
 
-	bool ret;
-	ret = obj_dict_foreach(wk, dict, &ctx, generate_config_iter);
+	bool ret = false, first = true;
 
-	if (!file_exists_with_content(wk, get_cstr(wk, out_path), ctx.out_buf->buf, ctx.out_buf->len)) {
-		if (!fs_write(get_cstr(wk, out_path), (uint8_t *)ctx.out_buf->buf, ctx.out_buf->len)) {
-			ret = false;
+	obj key, val;
+	obj_dict_for(wk, dict, key, val) {
+		enum obj_type t = get_obj_type(wk, val);
+
+		char define_prefix = (char[]){
+			[configure_file_output_format_c] = '#',
+			[configure_file_output_format_nasm] = '%',
+			[configure_file_output_format_json] = 0,
+		}[format];
+
+		if (format == configure_file_output_format_json) {
+			if (!first) {
+				sbuf_push(wk, &out_buf, ',');
+			}
+			first = false;
+
+			sbuf_push_json_escaped_quoted(wk, &out_buf, get_str(wk, key));
+			sbuf_push(wk, &out_buf, ':');
+		}
+
+		switch (t) {
+		case obj_string:
+			/* conf_data.set('FOO', '"string"') => #define FOO "string" */
+			/* conf_data.set('FOO', 'a_token')  => #define FOO a_token */
+			switch (format) {
+			case configure_file_output_format_c:
+			case configure_file_output_format_nasm:
+				sbuf_pushf(wk,
+					&out_buf,
+					"%cdefine %s %s\n",
+					define_prefix,
+					get_cstr(wk, key),
+					get_cstr(wk, val));
+				break;
+			case configure_file_output_format_json: {
+				sbuf_push_json_escaped_quoted(wk, &out_buf, get_str(wk, val));
+				break;
+			}
+			}
+			break;
+		case obj_bool:
+			/* conf_data.set('FOO', true)       => #define FOO */
+			/* conf_data.set('FOO', false)      => #undef FOO */
+			switch (format) {
+			case configure_file_output_format_c:
+			case configure_file_output_format_nasm:
+				if (get_obj_bool(wk, val)) {
+					sbuf_pushf(wk, &out_buf, "%cdefine %s\n", define_prefix, get_cstr(wk, key));
+				} else {
+					sbuf_pushf(wk, &out_buf, "%cundef %s\n", define_prefix, get_cstr(wk, key));
+				}
+				break;
+			case configure_file_output_format_json:
+				sbuf_pushs(wk, &out_buf, get_obj_bool(wk, val) ? "true" : "false");
+				break;
+			}
+			break;
+		case obj_number:
+			/* conf_data.set('FOO', 1)          => #define FOO 1 */
+			/* conf_data.set('FOO', 0)          => #define FOO 0 */
+			switch (format) {
+			case configure_file_output_format_c:
+			case configure_file_output_format_nasm:
+				sbuf_pushf(wk,
+					&out_buf,
+					"%cdefine %s %" PRId64 "\n",
+					define_prefix,
+					get_cstr(wk, key),
+					get_obj_number(wk, val));
+				break;
+			case configure_file_output_format_json: {
+				char buf[32] = { 0 };
+				snprintf(buf, sizeof(buf), "%" PRId64, get_obj_number(wk, val));
+				sbuf_pushs(wk, &out_buf, buf);
+				break;
+			}
+			}
+			break;
+		default: vm_error_at(wk, node, "invalid type for config data value: '%s'", obj_type_to_s(t)); goto ret;
 		}
 	}
 
+	if (macro_name) {
+		sbuf_pushf(wk, &out_buf, "#endif\n");
+	} else if (format == configure_file_output_format_json) {
+		sbuf_pushs(wk, &out_buf, "}\n");
+	}
+
+	if (!file_exists_with_content(wk, get_cstr(wk, out_path), out_buf.buf, out_buf.len)) {
+		if (!fs_write(get_cstr(wk, out_path), (uint8_t *)out_buf.buf, out_buf.len)) {
+			goto ret;
+		}
+	}
+
+	ret = true;
+ret:
 	sbuf_destroy(&out_buf);
 	return ret;
 }
 
 static bool
-configure_file_with_command(struct workspace *wk, uint32_t node,
-	obj command, obj input, obj out_path, obj depfile,
+configure_file_with_command(struct workspace *wk,
+	uint32_t node,
+	obj command,
+	obj input,
+	obj out_path,
+	obj depfile,
 	bool capture)
 {
 	obj args, output_arr;
@@ -399,10 +470,10 @@ configure_file_with_command(struct workspace *wk, uint32_t node,
 	}
 
 	struct process_custom_target_commandline_opts opts = {
-		.err_node   = node,
-		.input      = input,
-		.output     = output_arr,
-		.depfile    = depfile,
+		.err_node = node,
+		.input = input,
+		.output = output_arr,
+		.depfile = depfile,
 	};
 	make_obj(wk, &opts.depends, obj_array);
 
@@ -426,12 +497,12 @@ configure_file_with_command(struct workspace *wk, uint32_t node,
 	join_args_argstr(wk, &argstr, &argc, args);
 	env_to_envstr(wk, &envstr, &envc, env);
 	if (!run_cmd(&cmd_ctx, argstr, argc, envstr, envc)) {
-		interp_error(wk, node, "error running command: %s", cmd_ctx.err_msg);
+		vm_error_at(wk, node, "error running command: %s", cmd_ctx.err_msg);
 		goto ret;
 	}
 
 	if (cmd_ctx.status != 0) {
-		interp_error(wk, node, "error running command: %s", cmd_ctx.err.buf);
+		vm_error_at(wk, node, "error running command: %s", cmd_ctx.err.buf);
 		goto ret;
 	}
 
@@ -461,7 +532,7 @@ array_to_elem_or_err(struct workspace *wk, uint32_t node, uint32_t arr, uint32_t
 	}
 
 	if (get_obj_array(wk, arr)->len != 1) {
-		interp_error(wk, node, "expected an array of length 1");
+		vm_error_at(wk, node, "expected an array of length 1");
 		return false;
 	}
 
@@ -499,7 +570,7 @@ perform_output_string_substitutions(struct workspace *wk, uint32_t node, uint32_
 				buf.len = strlen(buf.buf);
 			}
 
-			str_app(wk, str, buf.buf);
+			str_app(wk, &str, buf.buf);
 			s += len - 1;
 		} else if (is_substr(s, "@PLAINNAME@", &len)) {
 			if (!array_to_elem_or_err(wk, node, input_arr, &e)) {
@@ -508,10 +579,10 @@ perform_output_string_substitutions(struct workspace *wk, uint32_t node, uint32_
 
 			SBUF(buf);
 			path_basename(wk, &buf, get_file_path(wk, e));
-			str_app(wk, str, buf.buf);
+			str_app(wk, &str, buf.buf);
 			s += len - 1;
 		} else {
-			str_appn(wk, str, s, 1);
+			str_appn(wk, &str, s, 1);
 		}
 	}
 
@@ -538,7 +609,7 @@ exclusive_or(bool *vals, uint32_t len)
 }
 
 bool
-func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_configure_file(struct workspace *wk, obj _, obj *res)
 {
 	obj input_arr = 0, output_str;
 	enum kwargs {
@@ -556,10 +627,11 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		kw_output_format,
 		kw_encoding, // TODO: ignored
 		kw_depfile, // TODO: ignored
+		kw_macro_name,
 	};
 	struct args_kw akw[] = {
 		[kw_configuration] = { "configuration", tc_configuration_data | tc_dict },
-		[kw_input] = { "input", ARG_TYPE_ARRAY_OF | tc_coercible_files, },
+		[kw_input] = { "input", TYPE_TAG_LISTIFY | tc_coercible_files, },
 		[kw_output] = { "output", obj_string, .required = true },
 		[kw_command] = { "command", obj_array },
 		[kw_capture] = { "capture", obj_bool },
@@ -572,10 +644,11 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		[kw_output_format] = { "output_format", obj_string },
 		[kw_encoding] = { "encoding", obj_string },
 		[kw_depfile] = { "depfile", obj_string },
-		0
+		[kw_macro_name] = { "macro_name", obj_string },
+		0,
 	};
 
-	if (!interp_args(wk, args_node, NULL, NULL, akw)) {
+	if (!pop_args(wk, NULL, akw)) {
 		return false;
 	}
 
@@ -586,8 +659,11 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 			output_format = configure_file_output_format_c;
 		} else if (str_eql(output_format_str, &WKSTR("nasm"))) {
 			output_format = configure_file_output_format_nasm;
+		} else if (str_eql(output_format_str, &WKSTR("json"))) {
+			output_format = configure_file_output_format_json;
 		} else {
-			interp_error(wk, akw[kw_output_format].node, "invalid output format %o", akw[kw_output_format].val);
+			vm_error_at(
+				wk, akw[kw_output_format].node, "invalid output format %o", akw[kw_output_format].val);
 			return false;
 		}
 	}
@@ -596,15 +672,12 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		if (!coerce_files(wk, akw[kw_input].node, akw[kw_input].val, &input_arr)) {
 			return false;
 		}
-	} else {
-		// set this so we can use it in error handling later
-		akw[kw_input].node = args_node;
 	}
 
-	{       /* setup out file */
+	{ /* setup out file */
 		obj subd;
-		if (!perform_output_string_substitutions(wk, akw[kw_output].node,
-			akw[kw_output].val, input_arr, &subd)) {
+		if (!perform_output_string_substitutions(
+			    wk, akw[kw_output].node, akw[kw_output].val, input_arr, &subd)) {
 			return false;
 		}
 
@@ -612,7 +685,7 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		SBUF(out_path);
 
 		if (!path_is_basename(out)) {
-			interp_error(wk, akw[kw_output].node, "config file output '%s' contains path separator", out);
+			vm_error_at(wk, akw[kw_output].node, "config file output '%s' contains path separator", out);
 			return false;
 		}
 
@@ -628,17 +701,21 @@ func_configure_file(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		*get_obj_file(wk, *res) = output_str;
 	}
 
-	if (!exclusive_or((bool []) { akw[kw_command].set, akw[kw_configuration].set, akw[kw_copy].set }, 3)) {
-		interp_error(wk, args_node, "you must pass either command:, configuration:, or copy:");
+	if (!exclusive_or((bool[]){ akw[kw_command].set, akw[kw_configuration].set, akw[kw_copy].set }, 3)) {
+		vm_error(wk, "you must pass either command:, configuration:, or copy:");
 		return false;
 	}
 
 	if (akw[kw_command].set) {
 		bool capture = akw[kw_capture].set && get_obj_bool(wk, akw[kw_capture].val);
 
-		if (!configure_file_with_command(wk, akw[kw_command].node,
-			akw[kw_command].val, input_arr, output_str,
-			akw[kw_depfile].val, capture)) {
+		if (!configure_file_with_command(wk,
+			    akw[kw_command].node,
+			    akw[kw_command].val,
+			    input_arr,
+			    output_str,
+			    akw[kw_depfile].val,
+			    capture)) {
 			return false;
 		}
 	} else if (akw[kw_copy].set) {
@@ -677,14 +754,12 @@ copy_err:
 
 		enum obj_type t = get_obj_type(wk, conf);
 		switch (t) {
-		case obj_dict:
-			dict = conf;
-			break;
-		case obj_configuration_data:
-			dict = get_obj_configuration_data(wk, conf)->dict;
-			break;
+		case obj_dict: dict = conf; break;
+		case obj_configuration_data: dict = get_obj_configuration_data(wk, conf)->dict; break;
 		default:
-			interp_error(wk, akw[kw_configuration].node, "invalid type for configuration data '%s'",
+			vm_error_at(wk,
+				akw[kw_configuration].node,
+				"invalid type for configuration data '%s'",
 				obj_type_to_s(t));
 			return false;
 		}
@@ -703,43 +778,49 @@ copy_err:
 			workspace_add_regenerate_deps(wk, *get_obj_file(wk, input));
 			const char *path = get_file_path(wk, input);
 
-			enum configure_file_syntax syntax =
-				configure_file_syntax_mesondefine
-				| configure_file_syntax_mesonvar;
+			enum configure_file_syntax syntax = configure_file_syntax_mesondefine
+							    | configure_file_syntax_mesonvar;
 
 			if (akw[kw_format].set) {
 				const struct str *fmt = get_str(wk, akw[kw_format].val);
 				if (str_eql(fmt, &WKSTR("meson"))) {
-					syntax = configure_file_syntax_mesondefine
-						 | configure_file_syntax_mesonvar;
+					syntax = configure_file_syntax_mesondefine | configure_file_syntax_mesonvar;
 				} else if (str_eql(fmt, &WKSTR("cmake"))) {
-					syntax = configure_file_syntax_cmakedefine
-						 | configure_file_syntax_cmakevar;
+					syntax = configure_file_syntax_cmakedefine | configure_file_syntax_cmakevar;
 				} else if (str_eql(fmt, &WKSTR("cmake@"))) {
-					syntax = configure_file_syntax_cmakedefine
-						 | configure_file_syntax_mesonvar;
+					syntax = configure_file_syntax_cmakedefine | configure_file_syntax_mesonvar;
 				} else {
-					interp_error(wk, akw[kw_format].node, "invalid format type %o", akw[kw_format].val);
+					vm_error_at(
+						wk, akw[kw_format].node, "invalid format type %o", akw[kw_format].val);
 					return false;
 				}
 			}
 
-			if (!substitute_config(wk, dict, akw[kw_input].node,
-				path, output_str, syntax)) {
+			if (!substitute_config(wk, dict, akw[kw_input].node, path, output_str, syntax)) {
 				return false;
 			}
 		} else {
-			if (!generate_config(wk, output_format, dict,
-				akw[kw_configuration].node, output_str)) {
+			if (akw[kw_macro_name].set && output_format != configure_file_output_format_c) {
+				vm_error_at(
+					wk, akw[kw_macro_name].node, "macro_name specified with a non c output format");
+				return false;
+			}
+
+			if (!generate_config(wk,
+				    output_format,
+				    akw[kw_macro_name].val,
+				    dict,
+				    akw[kw_configuration].node,
+				    output_str)) {
 				return false;
 			}
 		}
 	}
 
 	if ((akw[kw_install].set && get_obj_bool(wk, akw[kw_install].val))
-	    || (!akw[kw_install].set && akw[kw_install_dir].set)) {
+		|| (!akw[kw_install].set && akw[kw_install_dir].set)) {
 		if (!akw[kw_install_dir].set) {
-			interp_error(wk, akw[kw_install].node, "configure_file installation requires install_dir");
+			vm_error_at(wk, akw[kw_install].node, "configure_file installation requires install_dir");
 			return false;
 		}
 
