@@ -18,7 +18,7 @@
 #include "functions/kernel/custom_target.h"
 #include "functions/string.h"
 #include "install.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
@@ -35,10 +35,7 @@ prefix_plus_index(const struct str *ss, const char *prefix, int64_t *index)
 {
 	uint32_t len = strlen(prefix);
 	if (str_startswith(ss, &WKSTR(prefix))) {
-		return str_to_i(&(struct str) {
-			.s = &ss->s[len],
-			.len = ss->len - len
-		}, index);
+		return str_to_i(&(struct str){ .s = &ss->s[len], .len = ss->len - len }, index, false);
 	}
 
 	return false;
@@ -121,8 +118,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 			continue;
 		}
 
-		if (!key_names[key].valid
-		    || (key_names[key].needs_name && !ctx->opts->name)) {
+		if (!key_names[key].valid || (key_names[key].needs_name && !ctx->opts->name)) {
 			return format_cb_not_found;
 		}
 
@@ -155,7 +151,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		 * currently processed meson.build is located in. Depending on
 		 * the backend, this may be an absolute or a relative to
 		 * current workdir path. */
-		str_relative_to_build_root(wk, ctx, get_cstr(wk, current_project(wk)->cwd), elem);
+		str_relative_to_build_root(wk, ctx, workspace_cwd(wk), elem);
 		return format_cb_found;
 	case key_private_dir: {
 		/* @PRIVATE_DIR@ (since 0.50.1): path to a directory where the
@@ -164,6 +160,10 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		path_join(wk, &path, get_cstr(wk, current_project(wk)->build_dir), get_cstr(wk, ctx->opts->name));
 		sbuf_pushs(wk, &path, ".p");
 
+		if (!fs_mkdir_p(path.buf)) {
+			return format_cb_error;
+		}
+
 		str_relative_to_build_root(wk, ctx, path.buf, elem);
 		return format_cb_found;
 	}
@@ -194,7 +194,8 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		/* @BASENAME@: the input filename, with extension removed */
 		struct obj_array *in = get_obj_array(wk, ctx->opts->input);
 		if (in->len != 1) {
-			interp_error(wk, ctx->opts->err_node,
+			vm_error_at(wk,
+				ctx->opts->err_node,
 				"to use @PLAINNAME@ and @BASENAME@ in a custom "
 				"target command, there must be exactly one input");
 			return format_cb_error;
@@ -217,8 +218,7 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		}
 		return format_cb_found;
 	}
-	default:
-		break;
+	default: break;
 	}
 
 	int64_t index;
@@ -230,7 +230,11 @@ format_cmd_arg_cb(struct workspace *wk, uint32_t node, void *_ctx, const struct
 		arr = ctx->opts->output;
 	} else {
 		if (ctx->opts->err_node) {
-			interp_warning(wk, ctx->opts->err_node, "not substituting unknown key '%.*s' in commandline", strkey->len, strkey->s);
+			vm_warning_at(wk,
+				ctx->opts->err_node,
+				"not substituting unknown key '%.*s' in commandline",
+				strkey->len,
+				strkey->s);
 		}
 		return format_cb_skip;
 	}
@@ -257,9 +261,10 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 	case obj_both_libs:
 	case obj_build_target:
 	case obj_external_program:
+	case obj_python_installation:
 	case obj_file: {
-		obj str;
-		if (!coerce_executable(wk, ctx->opts->err_node, val, &str)) {
+		obj str, args;
+		if (!coerce_executable(wk, ctx->opts->err_node, val, &str, &args)) {
 			return ir_err;
 		}
 
@@ -268,6 +273,13 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 		if (!ctx->skip_depends) {
 			obj_array_push(wk, ctx->opts->depends, ss);
 		}
+
+		if (args) {
+			obj_array_push(wk, *ctx->res, ss);
+			obj_array_extend_nodup(wk, *ctx->res, args);
+			return ir_cont;
+		}
+
 		break;
 	}
 	case obj_string: {
@@ -316,9 +328,7 @@ custom_target_cmd_fmt_iter(struct workspace *wk, void *_ctx, obj val)
 		}
 		goto cont;
 	}
-	default:
-		interp_error(wk, ctx->opts->err_node, "unable to coerce %o to string", val);
-		return ir_err;
+	default: vm_error_at(wk, ctx->opts->err_node, "unable to coerce %o to string", val); return ir_err;
 	}
 
 	assert(get_obj_type(wk, ss) == obj_string);
@@ -332,7 +342,8 @@ cont:
 bool
 process_custom_target_commandline(struct workspace *wk,
 	struct process_custom_target_commandline_opts *opts,
-	obj arr, obj *res)
+	obj arr,
+	obj *res)
 {
 	make_obj(wk, res, obj_array);
 
@@ -346,7 +357,7 @@ process_custom_target_commandline(struct workspace *wk,
 	}
 
 	if (!get_obj_array(wk, *res)->len) {
-		interp_error(wk, opts->err_node, "cmd cannot be empty");
+		vm_error_at(wk, opts->err_node, "cmd cannot be empty");
 		return false;
 	}
 	return true;
@@ -357,11 +368,7 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 {
 	struct custom_target_cmd_fmt_ctx *ctx = _ctx;
 
-	enum cmd_output_fmt_key {
-		key_plainname,
-		key_basename,
-		cmd_output_fmt_key_count
-	};
+	enum cmd_output_fmt_key { key_plainname, key_basename, cmd_output_fmt_key_count };
 
 	const char *key_names[cmd_output_fmt_key_count] = {
 		[key_plainname] = "PLAINNAME",
@@ -381,7 +388,8 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 
 	struct obj_array *in = get_obj_array(wk, ctx->opts->input);
 	if (in->len != 1) {
-		interp_error(wk, ctx->opts->err_node,
+		vm_error_at(wk,
+			ctx->opts->err_node,
 			"to use @PLAINNAME@ and @BASENAME@ in a custom "
 			"target output, there must be exactly one input");
 		return format_cb_error;
@@ -393,18 +401,14 @@ format_cmd_output_cb(struct workspace *wk, uint32_t node, void *_ctx, const stru
 	SBUF(buf);
 
 	switch (key) {
-	case key_plainname:
-		path_basename(wk, &buf, ss->s);
-		break;
+	case key_plainname: path_basename(wk, &buf, ss->s); break;
 	case key_basename: {
 		SBUF(basename);
 		path_basename(wk, &basename, ss->s);
 		path_without_ext(wk, &buf, basename.buf);
 		break;
 	}
-	default:
-		assert(false && "unreachable");
-		return format_cb_error;
+	default: assert(false && "unreachable"); return format_cb_error;
 	}
 
 	*elem = sbuf_into_str(wk, &buf);
@@ -463,8 +467,7 @@ process_custom_tgt_sources_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 bool
-make_custom_target(struct workspace *wk,
-	struct make_custom_target_opts *opts, obj *res)
+make_custom_target(struct workspace *wk, struct make_custom_target_opts *opts, obj *res)
 {
 	obj input, raw_output, output, args;
 
@@ -507,7 +510,7 @@ make_custom_target(struct workspace *wk,
 		if (!coerce_output_files(wk, opts->output_node, opts->output_orig, opts->output_dir, &raw_output)) {
 			return false;
 		} else if (!get_obj_array(wk, raw_output)->len) {
-			interp_error(wk, opts->output_node, "output cannot be empty");
+			vm_error_at(wk, opts->output_node, "output cannot be empty");
 			return false;
 		}
 
@@ -553,13 +556,13 @@ make_custom_target(struct workspace *wk,
 	}
 
 	struct process_custom_target_commandline_opts cmdline_opts = {
-		.err_node   = opts->command_node,
+		.err_node = opts->command_node,
 		.relativize = true,
-		.name       = opts->name,
-		.input      = input,
-		.output     = output,
-		.depfile    = depfile,
-		.build_dir  = opts->build_dir,
+		.name = opts->name,
+		.input = input,
+		.output = output,
+		.depfile = depfile,
+		.build_dir = opts->build_dir,
 		.extra_args = opts->extra_args,
 		.extra_args_valid = opts->extra_args_valid,
 	};
@@ -569,7 +572,7 @@ make_custom_target(struct workspace *wk,
 	}
 
 	if (opts->extra_args && !cmdline_opts.extra_args_used) {
-		interp_warning(wk, opts->command_node, "extra args passed, but no @EXTRA_ARGS@ key found to substitute");
+		vm_warning_at(wk, opts->command_node, "extra args passed, but no @EXTRA_ARGS@ key found to substitute");
 	}
 
 	if (opts->capture) {
@@ -589,9 +592,9 @@ make_custom_target(struct workspace *wk,
 }
 
 bool
-func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_custom_target(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm ao[] = { { obj_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string, .optional = true }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_input,
 		kw_output,
@@ -612,36 +615,36 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		kw_console,
 	};
 	struct args_kw akw[] = {
-		[kw_input] = { "input", ARG_TYPE_ARRAY_OF | tc_coercible_files | tc_generated_list, },
-		[kw_output] = { "output", ARG_TYPE_ARRAY_OF | tc_string, .required = true },
+		[kw_input] = { "input", TYPE_TAG_LISTIFY | tc_coercible_files | tc_generated_list, },
+		[kw_output] = { "output", TYPE_TAG_LISTIFY | tc_string, .required = true },
 		[kw_command] = { "command", tc_command_array | tc_both_libs, .required = true },
 		[kw_capture] = { "capture", obj_bool },
 		[kw_install] = { "install", obj_bool },
-		[kw_install_dir] = { "install_dir", ARG_TYPE_ARRAY_OF | tc_string | tc_bool },
+		[kw_install_dir] = { "install_dir", TYPE_TAG_LISTIFY | tc_string | tc_bool },
 		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_install_tag] = { "install_tag", tc_string }, // TODO
 		[kw_build_by_default] = { "build_by_default", obj_bool },
 		[kw_depfile] = { "depfile", obj_string },
-		[kw_depend_files] = { "depend_files", ARG_TYPE_ARRAY_OF | tc_string | tc_file },
+		[kw_depend_files] = { "depend_files", TYPE_TAG_LISTIFY | tc_string | tc_file },
 		[kw_depends] = { "depends", tc_depends_kw },
 		[kw_build_always_stale] = { "build_always_stale", obj_bool },
 		[kw_build_always] = { "build_always", obj_bool },
 		[kw_env] = { "env", tc_coercible_env },
 		[kw_feed] = { "feed", obj_bool },
 		[kw_console] = { "console", obj_bool },
-		0
+		0,
 	};
 
-	if (!interp_args(wk, args_node, NULL, ao, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	obj name;
-	if (ao[0].set) {
-		name = ao[0].val;
+	if (an[0].set) {
+		name = an[0].val;
 	} else {
 		if (!get_obj_array(wk, akw[kw_output].val)->len) {
-			interp_error(wk, akw[kw_output].node, "output cannot be empty");
+			vm_error_at(wk, akw[kw_output].node, "output cannot be empty");
 			return false;
 		}
 
@@ -651,17 +654,17 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 	}
 
 	struct make_custom_target_opts opts = {
-		.name         = name,
-		.input_node   = akw[kw_input].node,
-		.output_node  = akw[kw_output].node,
+		.name = name,
+		.input_node = akw[kw_input].node,
+		.output_node = akw[kw_output].node,
 		.command_node = akw[kw_command].node,
-		.input_orig   = akw[kw_input].val,
-		.output_orig  = akw[kw_output].val,
-		.output_dir   = get_cstr(wk, current_project(wk)->build_dir),
+		.input_orig = akw[kw_input].val,
+		.output_orig = akw[kw_output].val,
+		.output_dir = get_cstr(wk, current_project(wk)->build_dir),
 		.command_orig = akw[kw_command].val,
 		.depfile_orig = akw[kw_depfile].val,
-		.capture      = akw[kw_capture].set && get_obj_bool(wk, akw[kw_capture].val),
-		.feed         = akw[kw_feed].set && get_obj_bool(wk, akw[kw_feed].val),
+		.capture = akw[kw_capture].set && get_obj_bool(wk, akw[kw_capture].val),
+		.feed = akw[kw_feed].set && get_obj_bool(wk, akw[kw_feed].val),
 	};
 
 	if (!make_custom_target(wk, &opts, res)) {
@@ -702,7 +705,7 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 
 	if (akw[kw_console].set && get_obj_bool(wk, akw[kw_console].val)) {
 		if (opts.capture) {
-			interp_error(wk, akw[kw_console].node, "console and capture cannot both be set to true");
+			vm_error_at(wk, akw[kw_console].node, "console and capture cannot both be set to true");
 			return false;
 		}
 
@@ -710,9 +713,9 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 	}
 
 	if ((akw[kw_install].set && get_obj_bool(wk, akw[kw_install].val))
-	    || (!akw[kw_install].set && akw[kw_install_dir].set)) {
+		|| (!akw[kw_install].set && akw[kw_install_dir].set)) {
 		if (!akw[kw_install_dir].set || !get_obj_array(wk, akw[kw_install_dir].val)->len) {
-			interp_error(wk, akw[kw_install].node, "custom target installation requires install_dir");
+			vm_error_at(wk, akw[kw_install].node, "custom target installation requires install_dir");
 			return false;
 		}
 
@@ -732,8 +735,8 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 			install_dir = i0;
 		}
 
-		if (!push_install_targets(wk, akw[kw_install_dir].node, tgt->output,
-			install_dir, install_mode_id, false)) {
+		if (!push_install_targets(
+			    wk, akw[kw_install_dir].node, tgt->output, install_dir, install_mode_id, false)) {
 			return false;
 		}
 	}
@@ -749,7 +752,7 @@ func_custom_target(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 }
 
 bool
-func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_vcs_tag(struct workspace *wk, obj _, obj *res)
 {
 	enum kwargs {
 		kw_input,
@@ -758,22 +761,18 @@ func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		kw_fallback,
 		kw_replace_string,
 	};
-	struct args_kw akw[] = {
-		[kw_input] = { "input", ARG_TYPE_ARRAY_OF | tc_coercible_files, .required = true },
+	struct args_kw akw[] = { [kw_input] = { "input", TYPE_TAG_LISTIFY | tc_coercible_files, .required = true },
 		[kw_output] = { "output", obj_string, .required = true },
 		[kw_command] = { "command", tc_command_array | tc_both_libs },
 		[kw_fallback] = { "fallback", obj_string },
 		[kw_replace_string] = { "replace_string", obj_string },
-		0
-	};
+		0 };
 
-	if (!interp_args(wk, args_node, NULL, NULL, akw)) {
+	if (!pop_args(wk, NULL, akw)) {
 		return false;
 	}
 
-	obj replace_string = akw[kw_replace_string].set
-		? akw[kw_replace_string].val
-		: make_str(wk, "@VCS_TAG@");
+	obj replace_string = akw[kw_replace_string].set ? akw[kw_replace_string].val : make_str(wk, "\\@VCS_TAG\\@");
 
 	obj fallback;
 	if (akw[kw_fallback].set) {
@@ -785,14 +784,16 @@ func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 	obj command;
 	make_obj(wk, &command, obj_array);
 
-	push_args_null_terminated(wk, command, (char *const []){
-		(char *)wk->argv0,
-		"internal",
-		"eval",
-		"-e",
-		"vcs_tagger.meson",
-		NULL,
-	});
+	push_args_null_terminated(wk,
+		command,
+		(char *const[]){
+			(char *)wk->argv0,
+			"internal",
+			"eval",
+			"-e",
+			"vcs_tagger.meson",
+			NULL,
+		});
 
 	obj input;
 	{
@@ -802,7 +803,7 @@ func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		}
 
 		if (!obj_array_flatten_one(wk, input_arr, &input)) {
-			interp_error(wk, akw[kw_input].node, "expected exactly one input");
+			vm_error_at(wk, akw[kw_input].node, "expected exactly one input");
 			return false;
 		}
 	}
@@ -818,12 +819,12 @@ func_vcs_tag(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 	}
 
 	struct make_custom_target_opts opts = {
-		.name         = make_str(wk, "vcs_tag"),
-		.input_node   = akw[kw_input].node,
-		.output_node  = akw[kw_output].node,
-		.input_orig   = akw[kw_input].val,
-		.output_orig  = akw[kw_output].val,
-		.output_dir   = get_cstr(wk, current_project(wk)->build_dir),
+		.name = make_str(wk, "vcs_tag"),
+		.input_node = akw[kw_input].node,
+		.output_node = akw[kw_output].node,
+		.input_orig = akw[kw_input].val,
+		.output_orig = akw[kw_output].val,
+		.output_dir = get_cstr(wk, current_project(wk)->build_dir),
 		.command_orig = command,
 	};
 
