@@ -12,14 +12,14 @@
 #include "coerce.h"
 #include "error.h"
 #include "external/libpkgconf.h"
-#include "functions/common.h"
 #include "functions/kernel/dependency.h"
 #include "functions/kernel/subproject.h"
-#include "functions/machine.h"
 #include "functions/string.h"
 #include "functions/subproject.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/typecheck.h"
 #include "log.h"
+#include "machines.h"
 #include "options.h"
 #include "platform/filesystem.h"
 #include "platform/path.h"
@@ -52,7 +52,7 @@ enum dep_lib_mode {
 
 struct dep_lookup_ctx {
 	obj *res;
-	struct args_kw *default_options, *versions;
+	struct args_kw *default_options, *versions, *handler_kwargs;
 	enum requirement_type requirement;
 	uint32_t err_node;
 	uint32_t fallback_node;
@@ -146,15 +146,10 @@ handle_dependency_fallback(struct workspace *wk, struct dep_lookup_ctx *ctx, boo
 	obj subproj_name, subproj_dep = 0, subproj;
 
 	switch (get_obj_array(wk, ctx->fallback)->len) {
-	case 2:
-		obj_array_index(wk, ctx->fallback, 1, &subproj_dep);
+	case 2: obj_array_index(wk, ctx->fallback, 1, &subproj_dep);
 	/* FALLTHROUGH */
-	case 1:
-		obj_array_index(wk, ctx->fallback, 0, &subproj_name);
-		break;
-	default:
-		interp_error(wk, ctx->err_node, "expected array of length 1-2 for fallback");
-		return false;
+	case 1: obj_array_index(wk, ctx->fallback, 0, &subproj_name); break;
+	default: vm_error_at(wk, ctx->err_node, "expected array of length 1-2 for fallback"); return false;
 	}
 
 	if (ctx->lib_mode != dep_lib_mode_default) {
@@ -190,18 +185,21 @@ handle_dependency_fallback(struct workspace *wk, struct dep_lookup_ctx *ctx, boo
 
 	if (subproj_dep) {
 		if (!subproject_get_variable(wk, ctx->fallback_node, subproj_dep, 0, subproj, ctx->res)) {
-			interp_warning(wk, ctx->fallback_node, "subproject dependency variable %o is not defined", subproj_dep);
+			vm_warning_at(wk,
+				ctx->fallback_node,
+				"subproject dependency variable %o is not defined",
+				subproj_dep);
 			goto not_found;
 		}
 	} else {
 		if (!check_dependency_override(wk, ctx)) {
-			interp_warning(wk, ctx->fallback_node, "subproject does not override dependency %o", ctx->name);
+			vm_warning_at(wk, ctx->fallback_node, "subproject does not override dependency %o", ctx->name);
 			goto not_found;
 		}
 	}
 
 	if (get_obj_type(wk, *ctx->res) != obj_dependency) {
-		interp_warning(wk, ctx->fallback_node, "overridden dependency is not a dependency object");
+		vm_warning_at(wk, ctx->fallback_node, "overridden dependency is not a dependency object");
 		goto not_found;
 	}
 
@@ -229,8 +227,12 @@ get_dependency_pkgconfig(struct workspace *wk, struct dep_lookup_ctx *ctx, bool
 	if (!check_dependency_version(wk, ver_str, ctx->err_node, ctx->versions->val, &ver_match)) {
 		return false;
 	} else if (!ver_match) {
-		obj_fprintf(wk, log_file(), "pkgconf found dependency %o, but the version %o does not match the requested version %o\n",
-			ctx->name, ver_str, ctx->versions->val);
+		obj_fprintf(wk,
+			log_file(),
+			"pkgconf found dependency %o, but the version %o does not match the requested version %o\n",
+			ctx->name,
+			ver_str,
+			ctx->versions->val);
 		return true;
 	}
 
@@ -258,8 +260,8 @@ get_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx)
 		if (check_dependency_cache(wk, ctx, &cached_dep)) {
 			bool ver_match;
 			struct obj_dependency *dep = get_obj_dependency(wk, cached_dep);
-			if (!check_dependency_version(wk, dep->version, ctx->versions->node,
-				ctx->versions->val, &ver_match)) {
+			if (!check_dependency_version(
+				    wk, dep->version, ctx->versions->node, ctx->versions->val, &ver_match)) {
 				return false;
 			}
 
@@ -283,8 +285,7 @@ get_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx)
 
 	if (!ctx->fallback) {
 		obj provided_fallback;
-		if (obj_dict_index(wk, current_project(wk)->wrap_provides_deps,
-			ctx->name, &provided_fallback)) {
+		if (obj_dict_index(wk, current_project(wk)->wrap_provides_deps, ctx->name, &provided_fallback)) {
 			ctx->fallback = provided_fallback;
 		}
 	}
@@ -301,10 +302,8 @@ get_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx)
 		get_option_value(wk, current_project(wk), "force_fallback_for", &force_fallback_for);
 		obj_array_index(wk, ctx->fallback, 0, &subproj_name);
 
-		force_fallback =
-			wrap_mode == wrap_mode_forcefallback
-			|| obj_array_in(wk, force_fallback_for, ctx->name)
-			|| obj_dict_in(wk, wk->subprojects, subproj_name);
+		force_fallback = wrap_mode == wrap_mode_forcefallback || obj_array_in(wk, force_fallback_for, ctx->name)
+				 || obj_dict_in(wk, wk->subprojects, subproj_name);
 	}
 
 	if (!ctx->found) {
@@ -342,7 +341,14 @@ handle_appleframeworks_modules_iter(struct workspace *wk, void *_ctx, obj val)
 static bool
 handle_special_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx, bool *handled)
 {
-	if (strcmp(get_cstr(wk, ctx->name), "threads") == 0) {
+	obj handler;
+	if (obj_dict_index(wk, wk->dependency_handlers, ctx->name, &handler)) {
+		*handled = true;
+
+		if (!vm_eval_capture(wk, handler, 0, ctx->handler_kwargs, ctx->res)) {
+			return false;
+		}
+	} else if (strcmp(get_cstr(wk, ctx->name), "threads") == 0) {
 		LOG_I("dependency threads found");
 
 		*handled = true;
@@ -370,12 +376,12 @@ handle_special_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx, bool
 	} else if (strcmp(get_cstr(wk, ctx->name), "appleframeworks") == 0) {
 		*handled = true;
 		if (!ctx->modules) {
-			interp_error(wk, ctx->err_node, "'appleframeworks' dependency requires the modules keyword");
+			vm_error_at(wk, ctx->err_node, "'appleframeworks' dependency requires the modules keyword");
 			return false;
 		}
 
 		make_obj(wk, ctx->res, obj_dependency);
-		if (machine_system() == machine_system_darwin) {
+		if (host_machine.sys == machine_system_darwin) {
 			struct obj_dependency *dep = get_obj_dependency(wk, *ctx->res);
 			dep->name = make_str(wk, "appleframeworks");
 			dep->flags |= dep_flag_found;
@@ -386,7 +392,7 @@ handle_special_dependency(struct workspace *wk, struct dep_lookup_ctx *ctx, bool
 	} else if (strcmp(get_cstr(wk, ctx->name), "") == 0) {
 		*handled = true;
 		if (ctx->requirement == requirement_required) {
-			interp_error(wk, ctx->err_node, "dependency '' cannot be required");
+			vm_error_at(wk, ctx->err_node, "dependency '' cannot be required");
 			return false;
 		}
 		make_obj(wk, ctx->res, obj_dependency);
@@ -403,7 +409,7 @@ dependency_iter(struct workspace *wk, void *_ctx, obj name)
 	bool handled;
 	struct dep_lookup_ctx *parent_ctx = _ctx;
 	struct dep_lookup_ctx ctx = *parent_ctx;
-	ctx.name = name;
+	parent_ctx->name = ctx.name = name;
 
 	if (!handle_special_dependency(wk, &ctx, &handled)) {
 		return ir_err;
@@ -416,7 +422,6 @@ dependency_iter(struct workspace *wk, void *_ctx, obj name)
 	}
 
 	if (ctx.found) {
-		parent_ctx->name = name;
 		parent_ctx->lib_mode = ctx.lib_mode;
 		parent_ctx->from_cache = ctx.from_cache;
 		parent_ctx->found = true;
@@ -443,9 +448,9 @@ set_dependency_cache_iter(struct workspace *wk, void *_ctx, obj name)
 }
 
 bool
-func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_dependency(struct workspace *wk, obj self, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | obj_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_required,
 		kw_native, // ignored
@@ -465,38 +470,42 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 	struct args_kw akw[] = {
 		[kw_required] = { "required", tc_required_kw },
 		[kw_native] = { "native", obj_bool },
-		[kw_version] = { "version", ARG_TYPE_ARRAY_OF | obj_string },
+		[kw_version] = { "version", TYPE_TAG_LISTIFY | obj_string },
 		[kw_static] = { "static", obj_bool },
-		[kw_modules] = { "modules", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_optional_modules] = { "optional_modules", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_components] = { "components", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_fallback] = { "fallback", ARG_TYPE_ARRAY_OF | obj_string },
+		[kw_modules] = { "modules", TYPE_TAG_LISTIFY | obj_string },
+		[kw_optional_modules] = { "optional_modules", TYPE_TAG_LISTIFY | obj_string },
+		[kw_components] = { "components", TYPE_TAG_LISTIFY | obj_string },
+		[kw_fallback] = { "fallback", TYPE_TAG_LISTIFY | obj_string },
 		[kw_allow_fallback] = { "allow_fallback", obj_bool },
-		[kw_default_options] = { "default_options", ARG_TYPE_ARRAY_OF | obj_string },
+		[kw_default_options] = { "default_options", wk->complex_types.options_dict_or_list },
 		[kw_not_found_message] = { "not_found_message", obj_string },
 		[kw_disabler] = { "disabler", obj_bool },
 		[kw_method] = { "method", obj_string },
 		[kw_include_type] = { "include_type", obj_string },
-		0
+		0,
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	if (!get_obj_array(wk, an[0].val)->len) {
-		interp_error(wk, an[0].node, "no dependency names specified");
+		vm_error_at(wk, an[0].node, "no dependency names specified");
 		return false;
 	}
 
 	enum include_type inc_type = include_type_preserve;
 	if (akw[kw_include_type].set) {
-		if (!coerce_include_type(wk, get_str(wk, akw[kw_include_type].val),
-			akw[kw_include_type].node, &inc_type)) {
+		if (!coerce_include_type(
+			    wk, get_str(wk, akw[kw_include_type].val), akw[kw_include_type].node, &inc_type)) {
 			return false;
 		}
 	}
 
+	// TODO: lookup_method is unused.  This is partially because cmake and
+	// extraframework aren't supported so the default (auto) is the same as
+	// just saying pkgconfig.  It also seems that rarely is the lookup
+	// method specified to be builtin or system in practice.
 	enum dependency_lookup_method lookup_method = dependency_lookup_method_auto;
 	if (akw[kw_method].set) {
 		struct {
@@ -530,16 +539,18 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		}
 
 		if (i == ARRAY_LEN(lookup_method_names)) {
-			interp_error(wk, akw[kw_method].node, "invalid dependency method %o", akw[kw_method].val);
+			vm_error_at(wk, akw[kw_method].node, "invalid dependency method %o", akw[kw_method].val);
 			return false;
 		}
 
 		if (!(lookup_method == dependency_lookup_method_auto
-		      || lookup_method == dependency_lookup_method_pkgconfig
-		      || lookup_method == dependency_lookup_method_builtin
-		      || lookup_method == dependency_lookup_method_system
-		      )) {
-			interp_warning(wk, akw[kw_method].node, "unsupported dependency method %o, falling back to 'auto'", akw[kw_method].val);
+			    || lookup_method == dependency_lookup_method_pkgconfig
+			    || lookup_method == dependency_lookup_method_builtin
+			    || lookup_method == dependency_lookup_method_system)) {
+			vm_warning_at(wk,
+				akw[kw_method].node,
+				"unsupported dependency method %o, falling back to 'auto'",
+				akw[kw_method].val);
 			lookup_method = dependency_lookup_method_auto;
 		}
 	}
@@ -563,6 +574,12 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		} else {
 			lib_mode = dep_lib_mode_shared;
 		}
+	} else {
+		obj prefer_static;
+		get_option_value(wk, current_project(wk), "prefer_static", &prefer_static);
+		if (get_obj_bool(wk, prefer_static)) {
+			lib_mode = dep_lib_mode_static;
+		}
 	}
 
 	/* A fallback is allowed if */
@@ -573,7 +590,8 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		|| (!akw[kw_allow_fallback].set && requirement == requirement_required)
 		/* - allow_fallback is not specified and the fallback keyword is
 		 *   specified with at least one value (i.e. not an empty array) */
-		|| (!akw[kw_allow_fallback].set && akw[kw_fallback].set && get_obj_array(wk, akw[kw_fallback].val)->len);
+		|| (!akw[kw_allow_fallback].set && akw[kw_fallback].set
+			&& get_obj_array(wk, akw[kw_fallback].val)->len);
 
 	uint32_t fallback_err_node = 0;
 	obj fallback = 0;
@@ -588,8 +606,15 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		}
 	}
 
+	struct args_kw handler_kwargs[] = {
+		{ "required", .val = requirement == requirement_required ? obj_bool_true : obj_bool_false },
+		{ "static", .val = lib_mode == dep_lib_mode_static ? obj_bool_true : obj_bool_false },
+		0,
+	};
+
 	struct dep_lookup_ctx ctx = {
 		.res = res,
+		.handler_kwargs = handler_kwargs,
 		.names = an[0].val,
 		.requirement = requirement,
 		.versions = &akw[kw_version],
@@ -630,14 +655,16 @@ func_dependency(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		log_plain("\n");
 
 		if (ctx.requirement == requirement_required) {
-			interp_error(wk, ctx.err_node, "required dependency not found");
+			vm_error_at(wk, ctx.err_node, "required dependency not found");
 			return false;
 		} else {
 			if (ctx.disabler) {
 				*ctx.res = disabler_id;
 			} else {
 				make_obj(wk, ctx.res, obj_dependency);
-				get_obj_dependency(wk, *ctx.res)->name = ctx.name;
+				struct obj_dependency *dep = get_obj_dependency(wk, *ctx.res);
+				dep->name = ctx.name;
+				dep->type = dependency_type_not_found;
 			}
 		}
 	} else if (!str_eql(get_str(wk, ctx.name), &WKSTR(""))) {
@@ -699,9 +726,7 @@ coerce_dependency_sources_iter(struct workspace *wk, void *_ctx, obj val)
 	struct process_dependency_sources_ctx *ctx = _ctx;
 
 	switch (get_obj_type(wk, val)) {
-	case obj_generated_list:
-		obj_array_push(wk, ctx->res, val);
-		break;
+	case obj_generated_list: obj_array_push(wk, ctx->res, val); break;
 	default: {
 		obj res;
 		if (!coerce_files(wk, ctx->err_node, val, &res)) {
@@ -715,7 +740,7 @@ coerce_dependency_sources_iter(struct workspace *wk, void *_ctx, obj val)
 	return ir_cont;
 }
 bool
-func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_declare_dependency(struct workspace *wk, obj _, obj *res)
 {
 	enum kwargs {
 		kw_sources,
@@ -727,27 +752,32 @@ func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *re
 		kw_include_directories,
 		kw_variables,
 		kw_compile_args,
+		kw_objects,
+		kw_extra_files,
 	};
 	struct args_kw akw[] = {
-		[kw_sources] = { "sources", ARG_TYPE_ARRAY_OF | tc_coercible_files | tc_generated_list },
+		[kw_sources] = { "sources", TYPE_TAG_LISTIFY | tc_coercible_files | tc_generated_list },
 		[kw_link_with] = { "link_with", tc_link_with_kw },
 		[kw_link_whole] = { "link_whole", tc_link_with_kw },
-		[kw_link_args] = { "link_args", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_dependencies] = { "dependencies", ARG_TYPE_ARRAY_OF | tc_dependency },
+		[kw_link_args] = { "link_args", TYPE_TAG_LISTIFY | obj_string },
+		[kw_dependencies] = { "dependencies", TYPE_TAG_LISTIFY | tc_dependency },
 		[kw_version] = { "version", obj_string },
-		[kw_include_directories] = { "include_directories", ARG_TYPE_ARRAY_OF | tc_coercible_inc },
+		[kw_include_directories] = { "include_directories", TYPE_TAG_LISTIFY | tc_coercible_inc },
 		[kw_variables] = { "variables", tc_array | tc_dict },
-		[kw_compile_args] = { "compile_args", ARG_TYPE_ARRAY_OF | obj_string },
-		0
+		[kw_compile_args] = { "compile_args", TYPE_TAG_LISTIFY | obj_string },
+		[kw_objects] = { "objects", TYPE_TAG_LISTIFY | tc_file | tc_string },
+		[kw_extra_files] = { "extra_files", TYPE_TAG_LISTIFY | tc_coercible_files }, // ignored
+		0,
 	};
 
-	if (!interp_args(wk, args_node, NULL, NULL, akw)) {
+	if (!pop_args(wk, NULL, akw)) {
 		return false;
 	}
 
 	if (akw[kw_include_directories].set) {
 		obj inc_dirs;
-		if (!coerce_include_dirs(wk, akw[kw_include_directories].node, akw[kw_include_directories].val, false, &inc_dirs)) {
+		if (!coerce_include_dirs(
+			    wk, akw[kw_include_directories].node, akw[kw_include_directories].val, false, &inc_dirs)) {
 			return false;
 		}
 
@@ -757,16 +787,19 @@ func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *re
 	struct obj_dependency *dep;
 	make_obj(wk, res, obj_dependency);
 	dep = get_obj_dependency(wk, *res);
+
+	if (akw[kw_objects].set) {
+		dep->dep.objects = akw[kw_objects].val;
+	}
+
 	build_dep_init(wk, &dep->dep);
 
-	dep->name = make_strf(wk, "%s:declared_dep@%s:%d",
-		get_cstr(wk, current_project(wk)->cfg.name),
-		wk->src->label,
-		get_node(wk->ast, args_node)->line);
+	dep->name = make_strf(wk, "%s:declared_dep", get_cstr(wk, current_project(wk)->cfg.name)),
 	dep->flags |= dep_flag_found;
 	dep->type = dependency_type_declared;
 
-	if (akw[kw_variables].set && !coerce_key_value_dict(wk, akw[kw_variables].node, akw[kw_variables].val, &dep->variables)) {
+	if (akw[kw_variables].set
+		&& !coerce_key_value_dict(wk, akw[kw_variables].node, akw[kw_variables].val, &dep->variables)) {
 		return false;
 	}
 
@@ -796,21 +829,20 @@ func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *re
 	}
 
 	if (akw[kw_link_with].set) {
-		if (!dep_process_link_with(wk, akw[kw_link_with].node,
-			akw[kw_link_with].val, &dep->dep)) {
+		if (!dep_process_link_with(wk, akw[kw_link_with].node, akw[kw_link_with].val, &dep->dep)) {
 			return false;
 		}
 	}
 
 	if (akw[kw_link_whole].set) {
-		if (!dep_process_link_whole(wk, akw[kw_link_whole].node,
-			akw[kw_link_whole].val, &dep->dep)) {
+		if (!dep_process_link_whole(wk, akw[kw_link_whole].node, akw[kw_link_whole].val, &dep->dep)) {
 			return false;
 		}
 	}
 
 	if (akw[kw_include_directories].set) {
-		dep_process_includes(wk, akw[kw_include_directories].val, include_type_preserve, dep->dep.include_directories);
+		dep_process_includes(
+			wk, akw[kw_include_directories].val, include_type_preserve, dep->dep.include_directories);
 	}
 
 	if (akw[kw_dependencies].set) {
@@ -826,11 +858,11 @@ func_declare_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *re
 static bool
 skip_if_present(struct workspace *wk, obj arr, obj val)
 {
-	if (hash_get(&wk->obj_hash, &val)) {
+	if (hash_get(&wk->vm.objects.obj_hash, &val)) {
 		return true;
 	}
 
-	hash_set(&wk->obj_hash, &val, true);
+	hash_set(&wk->vm.objects.obj_hash, &val, true);
 
 	return false;
 }
@@ -850,20 +882,14 @@ dep_process_includes_iter(struct workspace *wk, void *_ctx, obj inc_id)
 	bool new_is_system = inc->is_system;
 
 	switch (ctx->include_type) {
-	case include_type_preserve:
-		break;
-	case include_type_system:
-		new_is_system = true;
-		break;
-	case include_type_non_system:
-		new_is_system = false;
-		break;
+	case include_type_preserve: break;
+	case include_type_system: new_is_system = true; break;
+	case include_type_non_system: new_is_system = false; break;
 	}
 
 	if (inc->is_system != new_is_system) {
 		make_obj(wk, &inc_id, obj_include_directory);
-		struct obj_include_directory *new_inc =
-			get_obj_include_directory(wk, inc_id);
+		struct obj_include_directory *new_inc = get_obj_include_directory(wk, inc_id);
 		*new_inc = *inc;
 		new_inc->is_system = new_is_system;
 	}
@@ -875,10 +901,13 @@ dep_process_includes_iter(struct workspace *wk, void *_ctx, obj inc_id)
 void
 dep_process_includes(struct workspace *wk, obj arr, enum include_type include_type, obj dest)
 {
-	obj_array_foreach_flat(wk, arr, &(struct dep_process_includes_ctx) {
-		.include_type = include_type,
-		.dest = dest,
-	}, dep_process_includes_iter);
+	obj_array_foreach_flat(wk,
+		arr,
+		&(struct dep_process_includes_ctx){
+			.include_type = include_type,
+			.dest = dest,
+		},
+		dep_process_includes_iter);
 }
 
 void
@@ -919,6 +948,10 @@ build_dep_init(struct workspace *wk, struct build_dep *dep)
 	if (!dep->sources) {
 		make_obj(wk, &dep->sources, obj_array);
 	}
+
+	if (!dep->objects) {
+		make_obj(wk, &dep->objects, obj_array);
+	}
 }
 
 static void
@@ -963,18 +996,10 @@ merge_build_deps(struct workspace *wk, struct build_dep *src, struct build_dep *
 	if (dep && src->sources) {
 		obj_array_extend(wk, dest->sources, src->sources);
 	}
-}
 
-static void
-obj_array_dedup_in_place(struct workspace *wk, obj *arr)
-{
-	if (!*arr) {
-		return;
+	if (dep && src->objects) {
+		obj_array_extend(wk, dest->objects, src->objects);
 	}
-
-	obj dedupd;
-	obj_array_dedup(wk, *arr, &dedupd);
-	*arr = dedupd;
 }
 
 static enum iteration_result
@@ -1016,6 +1041,7 @@ dedup_build_dep(struct workspace *wk, struct build_dep *dep)
 	obj_array_dedup_in_place(wk, &dep->rpath);
 	obj_array_dedup_in_place(wk, &dep->order_deps);
 	obj_array_dedup_in_place(wk, &dep->sources);
+	obj_array_dedup_in_place(wk, &dep->objects);
 
 	obj new_link_args;
 	make_obj(wk, &new_link_args, obj_array);
@@ -1056,15 +1082,14 @@ dep_process_link_with_iter(struct workspace *wk, void *_ctx, obj val)
 	}
 
 	switch (t) {
-	case obj_both_libs:
-		val = get_obj_both_libs(wk, val)->dynamic_lib;
+	case obj_both_libs: val = get_obj_both_libs(wk, val)->dynamic_lib;
 	/* fallthrough */
 	case obj_build_target: {
 		struct obj_build_target *tgt = get_obj_build_target(wk, val);
 		const char *path = get_cstr(wk, tgt->build_path);
 
 		if (ctx->link_whole && tgt->type != tgt_static_library) {
-			interp_error(wk, ctx->err_node, "link whole only accepts static libraries");
+			vm_error_at(wk, ctx->err_node, "link whole only accepts static libraries");
 			return ir_err;
 		}
 
@@ -1106,12 +1131,8 @@ dep_process_link_with_iter(struct workspace *wk, void *_ctx, obj val)
 		obj_array_push(wk, dest_link_with, *get_obj_file(wk, val));
 		break;
 	}
-	case obj_string:
-		obj_array_push(wk, dest_link_with, val);
-		break;
-	default:
-		interp_error(wk, ctx->err_node, "invalid type for link_with: '%s'", obj_type_to_s(t));
-		return ir_err;
+	case obj_string: obj_array_push(wk, dest_link_with, val); break;
+	default: vm_error_at(wk, ctx->err_node, "invalid type for link_with: '%s'", obj_type_to_s(t)); return ir_err;
 	}
 
 	return ir_cont;
@@ -1123,12 +1144,15 @@ dep_process_link_with(struct workspace *wk, uint32_t err_node, obj arr, struct b
 	build_dep_init(wk, dest);
 	dest->raw.link_with = arr;
 
-	hash_clear(&wk->obj_hash);
+	hash_clear(&wk->vm.objects.obj_hash);
 
-	if (!obj_array_foreach_flat(wk, arr, &(struct dep_process_link_with_ctx) {
-		.dest = dest,
-		.err_node = err_node,
-	}, dep_process_link_with_iter)) {
+	if (!obj_array_foreach_flat(wk,
+		    arr,
+		    &(struct dep_process_link_with_ctx){
+			    .dest = dest,
+			    .err_node = err_node,
+		    },
+		    dep_process_link_with_iter)) {
 		return false;
 	}
 
@@ -1142,13 +1166,16 @@ dep_process_link_whole(struct workspace *wk, uint32_t err_node, obj arr, struct
 	build_dep_init(wk, dest);
 	dest->raw.link_whole = arr;
 
-	hash_clear(&wk->obj_hash);
+	hash_clear(&wk->vm.objects.obj_hash);
 
-	if (!obj_array_foreach_flat(wk, arr, &(struct dep_process_link_with_ctx) {
-		.dest = dest,
-		.link_whole = true,
-		.err_node = err_node,
-	}, dep_process_link_with_iter)) {
+	if (!obj_array_foreach_flat(wk,
+		    arr,
+		    &(struct dep_process_link_with_ctx){
+			    .dest = dest,
+			    .link_whole = true,
+			    .err_node = err_node,
+		    },
+		    dep_process_link_with_iter)) {
 		return false;
 	}
 
@@ -1183,7 +1210,7 @@ dep_process_deps(struct workspace *wk, obj deps, struct build_dep *dest)
 	build_dep_init(wk, dest);
 	dest->raw.deps = deps;
 
-	hash_clear(&wk->obj_hash);
+	hash_clear(&wk->vm.objects.obj_hash);
 
 	obj_array_foreach(wk, deps, dest, dep_process_deps_iter);
 
