@@ -5,19 +5,24 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <string.h>
 
 #include "coerce.h"
 #include "error.h"
 #include "functions/kernel/install.h"
 #include "install.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "options.h"
+#include "platform/assert.h"
 #include "platform/path.h"
 
+#define install_follow_symlinks_check()                                                      \
+	if (akw[kw_follow_symlinks].set && !get_obj_bool(wk, akw[kw_follow_symlinks].val)) { \
+		LOG_W("follow_symlinks: false is not supported");                            \
+	}
+
 bool
-func_install_subdir(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
+func_install_subdir(struct workspace *wk, obj _, obj *ret)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -27,23 +32,25 @@ func_install_subdir(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 		kw_exclude_directories,
 		kw_exclude_files,
 		kw_strip_directory,
+		kw_follow_symlinks,
 	};
 	struct args_kw akw[] = {
 		[kw_install_dir] = { "install_dir", obj_string, .required = true },
 		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_install_tag] = { "install_tag", obj_string }, // TODO
-		[kw_exclude_directories] = { "exclude_directories", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_exclude_files] = { "exclude_files", ARG_TYPE_ARRAY_OF | obj_string },
+		[kw_exclude_directories] = { "exclude_directories", TYPE_TAG_LISTIFY | obj_string },
+		[kw_exclude_files] = { "exclude_files", TYPE_TAG_LISTIFY | obj_string },
 		[kw_strip_directory] = { "strip_directory", obj_bool },
-		0
+		[kw_follow_symlinks] = { "follow_symlinks", obj_bool },
+		0,
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
-	bool strip_directory = akw[kw_strip_directory].set
-		? get_obj_bool(wk, akw[kw_strip_directory].val)
-		: false;
+	install_follow_symlinks_check();
+
+	bool strip_directory = akw[kw_strip_directory].set ? get_obj_bool(wk, akw[kw_strip_directory].val) : false;
 
 	obj dest = akw[kw_install_dir].val;
 	if (!strip_directory) {
@@ -66,7 +73,7 @@ func_install_subdir(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 	}
 
 	SBUF(path);
-	path_join(wk, &path, get_cstr(wk, current_project(wk)->cwd), get_cstr(wk, an[0].val));
+	path_join(wk, &path, workspace_cwd(wk), get_cstr(wk, an[0].val));
 	obj src = sbuf_into_str(wk, &path);
 
 	struct obj_install_target *tgt;
@@ -102,7 +109,7 @@ install_man_iter(struct workspace *wk, void *_ctx, obj val)
 	--len;
 
 	if (len <= 1 || man.buf[len - 1] != '.' || man.buf[len] < '0' || man.buf[len] > '9') {
-		interp_error(wk, ctx->err_node, "invalid path to man page");
+		vm_error_at(wk, ctx->err_node, "invalid path to man page");
 		return ir_err;
 	}
 
@@ -135,9 +142,9 @@ install_man_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 bool
-func_install_man(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
+func_install_man(struct workspace *wk, obj _, obj *ret)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_coercible_files }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_coercible_files }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_install_dir,
 		kw_install_mode,
@@ -147,9 +154,9 @@ func_install_man(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 		[kw_install_dir] = { "install_dir", obj_string },
 		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_locale] = { "locale", obj_string },
-		0
+		0,
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
@@ -184,7 +191,7 @@ func_install_man(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 }
 
 bool
-func_install_symlink(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
+func_install_symlink(struct workspace *wk, obj _, obj *ret)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -196,9 +203,9 @@ func_install_symlink(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 		[kw_install_dir] = { "install_dir", obj_string, .required = true },
 		[kw_install_tag] = { "install_tag", obj_string }, // TODO
 		[kw_pointing_to] = { "pointing_to", obj_string, .required = true },
-		0
+		0,
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
@@ -233,19 +240,17 @@ install_emptydir_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 bool
-func_install_emptydir(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
+func_install_emptydir(struct workspace *wk, obj _, obj *ret)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | obj_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_install_mode,
 		kw_install_tag,
 	};
-	struct args_kw akw[] = {
-		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
+	struct args_kw akw[] = { [kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_install_tag] = { "install_tag", obj_string }, // TODO
-		0
-	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+		0 };
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
@@ -286,9 +291,9 @@ install_data_rename_iter(struct workspace *wk, void *_ctx, obj val)
 }
 
 bool
-func_install_data(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_install_data(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_file | tc_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_file | tc_string }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_install_dir,
 		kw_install_mode,
@@ -296,24 +301,28 @@ func_install_data(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		kw_rename,
 		kw_sources,
 		kw_preserve_path,
+		kw_follow_symlinks,
 	};
 
 	struct args_kw akw[] = {
 		[kw_install_dir] = { "install_dir", obj_string },
 		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_install_tag] = { "install_tag", obj_string }, // TODO
-		[kw_rename] = { "rename", ARG_TYPE_ARRAY_OF | obj_string },
-		[kw_sources] = { "sources", ARG_TYPE_ARRAY_OF | tc_file | tc_string },
+		[kw_rename] = { "rename", TYPE_TAG_LISTIFY | obj_string },
+		[kw_sources] = { "sources", TYPE_TAG_LISTIFY | tc_file | tc_string },
 		[kw_preserve_path] = { "preserve_path", obj_bool },
-		0
+		[kw_follow_symlinks] = { "follow_symlinks", obj_bool },
+		0,
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
+	install_follow_symlinks_check();
+
 	if (akw[kw_rename].set && akw[kw_preserve_path].set) {
-		interp_error(wk, akw[kw_preserve_path].node, "rename keyword conflicts with preserve_path");
+		vm_error_at(wk, akw[kw_preserve_path].node, "rename keyword conflicts with preserve_path");
 		return false;
 	}
 
@@ -328,7 +337,6 @@ func_install_data(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 		path_join(wk, &buf, get_cstr(wk, install_dir_base), get_cstr(wk, current_project(wk)->cfg.name));
 
 		install_dir = sbuf_into_str(wk, &buf);
-
 	}
 
 	obj sources = an[0].val;
@@ -340,9 +348,8 @@ func_install_data(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 	}
 
 	if (akw[kw_rename].set) {
-		if (get_obj_array(wk, akw[kw_rename].val)->len !=
-		    get_obj_array(wk, sources)->len) {
-			interp_error(wk, akw[kw_rename].node, "number of elements in rename != number of sources");
+		if (get_obj_array(wk, akw[kw_rename].val)->len != get_obj_array(wk, sources)->len) {
+			vm_error_at(wk, akw[kw_rename].node, "number of elements in rename != number of sources");
 			return false;
 		}
 
@@ -360,38 +367,40 @@ func_install_data(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 
 		return obj_array_foreach(wk, coerced, &ctx, install_data_rename_iter);
 	} else {
-		bool preserve_path =
-			akw[kw_preserve_path].set
-			&& get_obj_bool(wk, akw[kw_preserve_path].val);
+		bool preserve_path = akw[kw_preserve_path].set && get_obj_bool(wk, akw[kw_preserve_path].val);
 
-		return push_install_targets(wk, err_node, sources, install_dir,
-			akw[kw_install_mode].val, preserve_path);
+		return push_install_targets(
+			wk, err_node, sources, install_dir, akw[kw_install_mode].val, preserve_path);
 	}
 }
 
 bool
-func_install_headers(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
+func_install_headers(struct workspace *wk, obj _, obj *ret)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_file | tc_string }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_file | tc_string }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_install_dir,
 		kw_install_mode,
 		kw_subdir,
 		kw_preserve_path,
+		kw_follow_symlinks,
 	};
 	struct args_kw akw[] = {
 		[kw_install_dir] = { "install_dir", obj_string },
 		[kw_install_mode] = { "install_mode", tc_install_mode_kw },
 		[kw_subdir] = { "subdir", obj_string },
 		[kw_preserve_path] = { "preserve_path", obj_bool },
-		0
+		[kw_follow_symlinks] = { "follow_symlinks", obj_bool },
+		0,
 	};
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
+	install_follow_symlinks_check();
+
 	if (akw[kw_install_dir].set && akw[kw_subdir].set) {
-		interp_error(wk, akw[kw_subdir].node, "subdir may not be set if install_dir is set");
+		vm_error_at(wk, akw[kw_subdir].node, "subdir may not be set if install_dir is set");
 		return false;
 	}
 
@@ -411,10 +420,7 @@ func_install_headers(struct workspace *wk, obj _, uint32_t args_node, obj *ret)
 		install_dir = install_dir_base;
 	}
 
-	bool preserve_path =
-		akw[kw_preserve_path].set
-		&& get_obj_bool(wk, akw[kw_preserve_path].val);
+	bool preserve_path = akw[kw_preserve_path].set && get_obj_bool(wk, akw[kw_preserve_path].val);
 
-	return push_install_targets(wk, an[0].node, an[0].val,
-		install_dir, akw[kw_install_mode].val, preserve_path);
+	return push_install_targets(wk, an[0].node, an[0].val, install_dir, akw[kw_install_mode].val, preserve_path);
 }
