@@ -8,9 +8,9 @@
 #include <string.h>
 
 #include "error.h"
-#include "functions/common.h"
+#include "lang/func_lookup.h"
 #include "functions/kernel/options.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
 
@@ -34,7 +34,7 @@ build_option_type_from_s(struct workspace *wk, uint32_t node, uint32_t name, enu
 		}
 	}
 
-	interp_error(wk, node, "invalid option type '%s'", get_cstr(wk, name));
+	vm_error_at(wk, node, "invalid option type '%s'", get_cstr(wk, name));
 	return false;
 }
 
@@ -44,16 +44,12 @@ validate_option_name(struct workspace *wk, uint32_t err_node, obj name)
 	uint32_t i;
 	const struct str *s = get_str(wk, name);
 	for (i = 0; i < s->len; ++i) {
-		if (('a' <= s->s[i] && s->s[i] <= 'z')
-		    || ('A' <= s->s[i] && s->s[i] <= 'Z')
-		    || ('0' <= s->s[i] && s->s[i] <= '9')
-		    || (s->s[i] == '-')
-		    || (s->s[i] == '_')
-		    ) {
+		if (('a' <= s->s[i] && s->s[i] <= 'z') || ('A' <= s->s[i] && s->s[i] <= 'Z')
+			|| ('0' <= s->s[i] && s->s[i] <= '9') || (s->s[i] == '-') || (s->s[i] == '_')) {
 			continue;
 		}
 
-		interp_error(wk, err_node, "option name may not contain '%c'", s->s[i]);
+		vm_error_at(wk, err_node, "option name may not contain '%c'", s->s[i]);
 		return false;
 	}
 
@@ -61,7 +57,7 @@ validate_option_name(struct workspace *wk, uint32_t err_node, obj name)
 }
 
 bool
-func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_option(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -77,29 +73,37 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		kw_kind = kwargs_count,
 	};
 
-	struct args_kw akw[] = {
-		[kw_type] = { "type", obj_string },
+	// TODO: this winds up creating 4 typeinfo objects every time you call
+	// option, it'd be nice to have some caching
+	type_tag deprecated_type = make_complex_type(wk,
+		complex_type_or,
+		tc_string | tc_bool,
+		make_complex_type(wk,
+			complex_type_or,
+			make_complex_type(wk, complex_type_nested, tc_dict, tc_string),
+			make_complex_type(wk, complex_type_nested, tc_array, tc_string)));
+
+	struct args_kw akw[] = { [kw_type] = { "type", obj_string },
 		[kw_value] = { "value", tc_any },
 		[kw_description] = { "description", obj_string },
 		[kw_choices] = { "choices", obj_array },
 		[kw_max] = { "max", obj_number },
 		[kw_min] = { "min", obj_number },
 		[kw_yield] = { "yield", obj_bool },
-		[kw_deprecated] = { "deprecated", tc_string | tc_bool | tc_array | tc_dict },
+		[kw_deprecated] = { "deprecated", deprecated_type },
 		[kw_kind] = { 0 },
-		0
-	};
+		0 };
 
 	if (initializing_builtin_options) {
-		akw[kw_kind] = (struct args_kw) { "kind", tc_string };
+		akw[kw_kind] = (struct args_kw){ "kind", tc_string };
 	}
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	if (!akw[kw_type].set) {
-		interp_error(wk, args_node, "missing required keyword 'type'");
+		vm_error(wk, "missing required keyword 'type'");
 		return false;
 	}
 
@@ -127,22 +131,20 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 	uint32_t i;
 	for (i = 0; i < kwargs_count; ++i) {
 		switch (keyword_validity[type][i]) {
-		case kw_opt:
-			break;
+		case kw_opt: break;
 		case kw_inv:
 			if (akw[i].set) {
-				interp_error(wk, akw[i].node, "invalid keyword for option type");
+				vm_error_at(wk, akw[i].node, "invalid keyword for option type");
 				return false;
 			}
 			break;
 		case kw_req:
 			if (!akw[i].set) {
-				interp_error(wk, args_node, "missing keyword '%s' for option type", akw[i].key);
+				vm_error(wk, "missing keyword '%s' for option type", akw[i].key);
 				return false;
 			}
 			break;
-		default:
-			assert(false && "unreachable");
+		default: assert(false && "unreachable");
 		}
 	}
 
@@ -151,16 +153,14 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		val = akw[kw_value].val;
 	} else {
 		switch (type) {
-		case op_string:
-			val = make_str(wk, "");
-			break;
+		case op_string: val = make_str(wk, ""); break;
 		case op_boolean:
 			make_obj(wk, &val, obj_bool);
 			set_obj_bool(wk, val, true);
 			break;
 		case op_combo:
 			if (!get_obj_array(wk, akw[kw_choices].val)->len) {
-				interp_error(wk, akw[kw_choices].node, "combo option with no choices");
+				vm_error_at(wk, akw[kw_choices].node, "combo option with no choices");
 				return false;
 			}
 
@@ -177,8 +177,7 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 			make_obj(wk, &val, obj_feature_opt);
 			set_obj_feature_opt(wk, val, feature_opt_auto);
 			break;
-		default:
-			UNREACHABLE_RETURN;
+		default: UNREACHABLE_RETURN;
 		}
 	}
 
@@ -192,24 +191,7 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 	o->choices = akw[kw_choices].val;
 	o->yield = akw[kw_yield].set && get_obj_bool(wk, akw[kw_yield].val);
 	o->description = akw[kw_description].val;
-
-	if (akw[kw_deprecated].set) {
-		switch (get_obj_type(wk, akw[kw_deprecated].val)) {
-		case obj_array:
-			typecheck_array(wk, akw[kw_deprecated].node, akw[kw_deprecated].val, obj_string);
-			break;
-		case obj_dict:
-			typecheck_dict(wk, akw[kw_deprecated].node, akw[kw_deprecated].val, obj_string);
-			break;
-		case obj_string:
-		case obj_bool:
-			break;
-		default:
-			UNREACHABLE;
-		}
-
-		o->deprecated = akw[kw_deprecated].val;
-	}
+	o->deprecated = akw[kw_deprecated].val;
 
 	if (akw[kw_kind].set) {
 		if (str_eql(&WKSTR("default"), get_str(wk, akw[kw_kind].val))) {
@@ -217,7 +199,7 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		} else if (str_eql(&WKSTR("prefixed_dir"), get_str(wk, akw[kw_kind].val))) {
 			o->kind = build_option_kind_prefixed_dir;
 		} else {
-			interp_error(wk, akw[kw_kind].node, "invalid kind: %o", akw[kw_kind].val);
+			vm_error_at(wk, akw[kw_kind].node, "invalid kind: %o", akw[kw_kind].val);
 			return false;
 		}
 	}
@@ -233,7 +215,7 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 		opts = wk->global_opts;
 	}
 
-	if (!create_option(wk, args_node, opts, opt, val)) {
+	if (!create_option(wk, opts, opt, val)) {
 		return false;
 	}
 
@@ -241,17 +223,17 @@ func_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
 }
 
 bool
-func_get_option(struct workspace *wk, obj rcvr, uint32_t args_node, obj *res)
+func_get_option(struct workspace *wk, obj self, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	obj opt;
 	if (!get_option(wk, current_project(wk), get_str(wk, an[0].val), &opt)) {
-		interp_error(wk, an[0].node, "undefined option");
+		vm_error_at(wk, an[0].node, "undefined option");
 		return false;
 	}
 
