@@ -7,7 +7,7 @@
 
 #include "functions/kernel/subproject.h"
 #include "functions/string.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
 #include "platform/filesystem.h"
@@ -15,8 +15,12 @@
 #include "wrap.h"
 
 static bool
-subproject_prepare(struct workspace *wk, struct sbuf *cwd_buf, const char **cwd,
-	struct sbuf *build_dir_buf, const char **build_dir, bool required,
+subproject_prepare(struct workspace *wk,
+	struct sbuf *cwd_buf,
+	const char **cwd,
+	struct sbuf *build_dir_buf,
+	const char **build_dir,
+	bool required,
 	bool *found)
 {
 	if (!fs_dir_exists(*cwd)) {
@@ -81,8 +85,12 @@ wrap_done:
 }
 
 bool
-subproject(struct workspace *wk, obj name, enum requirement_type req, struct args_kw *default_options,
-	struct args_kw *versions, obj *res)
+subproject(struct workspace *wk,
+	obj name,
+	enum requirement_type req,
+	struct args_kw *default_options,
+	struct args_kw *versions,
+	obj *res)
 {
 	// don't re-evaluate the same subproject
 	if (obj_dict_index(wk, wk->subprojects, name, res)) {
@@ -98,10 +106,16 @@ subproject(struct workspace *wk, obj name, enum requirement_type req, struct arg
 	SBUF(cwd);
 	SBUF(build_dir);
 
-	path_join(wk, &cwd, get_cstr(wk, current_project(wk)->source_root), get_cstr(wk, current_project(wk)->subprojects_dir));
+	path_join(wk,
+		&cwd,
+		get_cstr(wk, current_project(wk)->source_root),
+		get_cstr(wk, current_project(wk)->subprojects_dir));
 	path_push(wk, &cwd, subproj_name);
 
-	path_join(wk, &build_dir, get_cstr(wk, current_project(wk)->build_root), get_cstr(wk, current_project(wk)->subprojects_dir));
+	path_join(wk,
+		&build_dir,
+		get_cstr(wk, current_project(wk)->build_root),
+		get_cstr(wk, current_project(wk)->subprojects_dir));
 	path_push(wk, &build_dir, subproj_name);
 
 	uint32_t subproject_id = 0;
@@ -111,8 +125,8 @@ subproject(struct workspace *wk, obj name, enum requirement_type req, struct arg
 	SBUF(sp_cwd_buf);
 	SBUF(sp_build_dir_buf);
 
-	if (!subproject_prepare(wk, &sp_cwd_buf, &sp_cwd, &sp_build_dir_buf,
-		&sp_build_dir, req == requirement_required, &found)) {
+	if (!subproject_prepare(
+		    wk, &sp_cwd_buf, &sp_cwd, &sp_build_dir_buf, &sp_build_dir, req == requirement_required, &found)) {
 		return false;
 	}
 
@@ -121,8 +135,7 @@ subproject(struct workspace *wk, obj name, enum requirement_type req, struct arg
 	}
 
 	if (default_options && default_options->set) {
-		if (!parse_and_set_default_options(wk, default_options->node,
-			default_options->val, name, true)) {
+		if (!parse_and_set_default_options(wk, default_options->node, default_options->val, name, true)) {
 			return false;
 		}
 	}
@@ -132,19 +145,21 @@ subproject(struct workspace *wk, obj name, enum requirement_type req, struct arg
 	}
 
 	if (versions && versions->set) {
-		struct project *subp = darr_get(&wk->projects, subproject_id);
+		struct project *subp = arr_get(&wk->projects, subproject_id);
 
 		bool compare_result;
-		if (!version_compare(wk, versions->node, get_str(wk, subp->cfg.version),
-			versions->val, &compare_result)) {
+		if (!version_compare(
+			    wk, versions->node, get_str(wk, subp->cfg.version), versions->val, &compare_result)) {
 			goto not_found;
 		}
 
 		if (!compare_result) {
 			if (req == requirement_required) {
-				interp_error(wk, versions->node,
+				vm_error_at(wk,
+					versions->node,
 					"subproject version mismatch; wanted %o, got %o",
-					versions->val, subp->cfg.version);
+					versions->val,
+					subp->cfg.version);
 				goto not_found;
 			}
 		}
@@ -165,7 +180,7 @@ subproject(struct workspace *wk, obj name, enum requirement_type req, struct arg
 	return true;
 not_found:
 	if (subproject_id) {
-		struct project *proj = darr_get(&wk->projects, subproject_id);
+		struct project *proj = arr_get(&wk->projects, subproject_id);
 		proj->not_ok = true;
 	}
 
@@ -173,7 +188,7 @@ not_found:
 }
 
 bool
-func_subproject(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_subproject(struct workspace *wk, obj _, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
 	enum kwargs {
@@ -183,13 +198,13 @@ func_subproject(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 
 	};
 	struct args_kw akw[] = {
-		[kw_default_options] = { "default_options", ARG_TYPE_ARRAY_OF | obj_string },
+		[kw_default_options] = { "default_options", wk->complex_types.options_dict_or_list },
 		[kw_required] = { "required", tc_required_kw },
-		[kw_version] = { "version", ARG_TYPE_ARRAY_OF | obj_string },
-		0
+		[kw_version] = { "version", TYPE_TAG_LISTIFY | obj_string },
+		0,
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
