@@ -7,36 +7,37 @@
 
 #include "compat.h"
 
+#include "args.h"
+#include "backend/common_args.h"
 #include "coerce.h"
 #include "compilers.h"
 #include "error.h"
 #include "functions/build_target.h"
-#include "functions/common.h"
 #include "functions/meson.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
 #include "platform/path.h"
 #include "version.h"
 
 static bool
-func_meson_get_compiler(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_get_compiler(struct workspace *wk, obj _, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	enum kwargs { kw_native, };
-	struct args_kw akw[] = {
-		[kw_native] = { "native", obj_bool },
-		0
+	enum kwargs {
+		kw_native,
 	};
+	struct args_kw akw[] = { [kw_native] = { "native", obj_bool }, 0 };
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
 	enum compiler_language l;
 	if (!s_to_compiler_language(get_cstr(wk, an[0].val), &l)
-	    || !obj_dict_geti(wk, current_project(wk)->compilers, l, res)) {
-		interp_error(wk, an[0].node, "no compiler found for '%s'", get_cstr(wk, an[0].val));
+		|| !obj_dict_geti(wk, current_project(wk)->compilers, l, res)) {
+		vm_error_at(wk, an[0].node, "no compiler found for '%s'", get_cstr(wk, an[0].val));
 		return false;
 	}
 
@@ -44,9 +45,9 @@ func_meson_get_compiler(struct workspace *wk, obj _, uint32_t args_node, obj *re
 }
 
 static bool
-func_meson_project_name(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_project_name(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -55,20 +56,37 @@ func_meson_project_name(struct workspace *wk, obj _, uint32_t args_node, obj *re
 }
 
 static bool
-func_meson_project_license(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_project_license(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
 	*res = current_project(wk)->cfg.license;
+	if (!*res) {
+		make_obj(wk, res, obj_array);
+	}
 	return true;
 }
 
 static bool
-func_meson_project_version(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_project_license_files(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
+		return false;
+	}
+
+	*res = current_project(wk)->cfg.license_files;
+	if (!*res) {
+		make_obj(wk, res, obj_array);
+	}
+	return true;
+}
+
+static bool
+func_meson_project_version(struct workspace *wk, obj _, obj *res)
+{
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -77,9 +95,9 @@ func_meson_project_version(struct workspace *wk, obj _, uint32_t args_node, obj
 }
 
 static bool
-func_meson_version(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_version(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -88,9 +106,9 @@ func_meson_version(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 }
 
 static bool
-func_meson_current_source_dir(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_current_source_dir(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -99,9 +117,9 @@ func_meson_current_source_dir(struct workspace *wk, obj _, uint32_t args_node, o
 }
 
 static bool
-func_meson_current_build_dir(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_current_build_dir(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -110,9 +128,9 @@ func_meson_current_build_dir(struct workspace *wk, obj _, uint32_t args_node, ob
 }
 
 static bool
-func_meson_project_source_root(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_project_source_root(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -121,9 +139,9 @@ func_meson_project_source_root(struct workspace *wk, obj _, uint32_t args_node,
 }
 
 static bool
-func_meson_project_build_root(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_project_build_root(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -132,9 +150,9 @@ func_meson_project_build_root(struct workspace *wk, obj _, uint32_t args_node, o
 }
 
 static bool
-func_meson_global_source_root(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_global_source_root(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -143,9 +161,9 @@ func_meson_global_source_root(struct workspace *wk, obj _, uint32_t args_node, o
 }
 
 static bool
-func_meson_global_build_root(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_global_build_root(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -154,9 +172,25 @@ func_meson_global_build_root(struct workspace *wk, obj _, uint32_t args_node, ob
 }
 
 static bool
-func_meson_is_subproject(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_build_options(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
+		return false;
+	}
+
+	obj options = regenerate_build_command(wk, true);
+
+	// remove the build directory from options
+	obj_array_pop(wk, options);
+
+	*res = join_args_shell(wk, options);
+	return true;
+}
+
+static bool
+func_meson_is_subproject(struct workspace *wk, obj _, obj *res)
+{
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -166,9 +200,9 @@ func_meson_is_subproject(struct workspace *wk, obj _, uint32_t args_node, obj *r
 }
 
 static bool
-func_meson_backend(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_backend(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -177,9 +211,9 @@ func_meson_backend(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 }
 
 static bool
-func_meson_is_cross_build(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_is_cross_build(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -189,9 +223,9 @@ func_meson_is_cross_build(struct workspace *wk, obj _, uint32_t args_node, obj *
 }
 
 static bool
-func_meson_is_unity(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_is_unity(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -201,20 +235,16 @@ func_meson_is_unity(struct workspace *wk, obj _, uint32_t args_node, obj *res)
 }
 
 static bool
-func_meson_override_dependency(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_override_dependency(struct workspace *wk, obj _, obj *res)
 {
 	struct args_norm an[] = { { obj_string }, { obj_dependency }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_static,
 		kw_native, // ignored
 	};
-	struct args_kw akw[] = {
-		[kw_static] = { "static", obj_bool },
-		[kw_native] = { "native", obj_bool },
-		0
-	};
+	struct args_kw akw[] = { [kw_static] = { "static", obj_bool }, [kw_native] = { "native", obj_bool }, 0 };
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
@@ -228,12 +258,8 @@ func_meson_override_dependency(struct workspace *wk, obj _, uint32_t args_node,
 		}
 	} else {
 		switch (get_option_default_library(wk)) {
-		case tgt_static_library:
-			override_dict = wk->dep_overrides_static;
-			break;
-		default:
-			override_dict = wk->dep_overrides_dynamic;
-			break;
+		case tgt_static_library: override_dict = wk->dep_overrides_static; break;
+		default: override_dict = wk->dep_overrides_dynamic; break;
 		}
 	}
 
@@ -242,15 +268,13 @@ func_meson_override_dependency(struct workspace *wk, obj _, uint32_t args_node,
 }
 
 static bool
-func_meson_override_find_program(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_override_find_program(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = {
-		{ obj_string },
-		{ tc_file | tc_external_program | tc_build_target | tc_custom_target },
-		ARG_TYPE_NULL
-	};
+	type_tag tc_allowed = tc_file | tc_external_program | tc_build_target | tc_custom_target
+			      | tc_python_installation;
+	struct args_norm an[] = { { obj_string }, { tc_allowed }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
@@ -270,10 +294,8 @@ func_meson_override_find_program(struct workspace *wk, obj _, uint32_t args_node
 		obj_array_push(wk, override, ver);
 		break;
 	case obj_external_program:
-		override = an[1].val;
-		break;
-	default:
-		UNREACHABLE;
+	case obj_python_installation: override = an[1].val; break;
+	default: UNREACHABLE;
 	}
 
 	obj_dict_set(wk, wk->find_program_overrides, an[0].val, override);
@@ -338,15 +360,24 @@ process_script_commandline_iter(struct workspace *wk, void *_ctx, obj val)
 	}
 	//fallthrough
 	case obj_external_program:
-	case obj_file:
-		if (!coerce_executable(wk, ctx->node, val, &str)) {
+	case obj_python_installation:
+	case obj_file: {
+		obj args;
+		if (!coerce_executable(wk, ctx->node, val, &str, &args)) {
 			return ir_err;
 		}
+
+		if (args) {
+			obj_array_push(wk, ctx->arr, str);
+			obj_array_extend_nodup(wk, ctx->arr, args);
+			return ir_cont;
+		}
+
 		break;
+	}
 	default:
 type_error:
-		interp_error(wk, ctx->node, "invalid type for script commandline '%s'",
-			obj_type_to_s(t));
+		vm_error_at(wk, ctx->node, "invalid type for script commandline '%s'", obj_type_to_s(t));
 		return ir_err;
 	}
 
@@ -357,20 +388,22 @@ cont:
 }
 
 static bool
-func_meson_add_install_script(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_add_install_script(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_exe }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_exe }, ARG_TYPE_NULL };
 	enum kwargs {
 		kw_install_tag, // ignored
 		kw_skip_if_destdir, // ignored
+		kw_dry_run,
 	};
 	struct args_kw akw[] = {
 		[kw_install_tag] = { "install_tag", obj_string },
 		[kw_skip_if_destdir] = { "skip_if_destdir", obj_bool },
-		0
+		[kw_dry_run] = { "dry_run", obj_bool },
+		0,
 	};
 
-	if (!interp_args(wk, args_node, an, NULL, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
@@ -385,15 +418,24 @@ func_meson_add_install_script(struct workspace *wk, obj _, uint32_t args_node, o
 		return false;
 	}
 
-	obj_array_push(wk, wk->install_scripts, ctx.arr);
+	if (!akw[kw_dry_run].set) {
+		make_obj(wk, &akw[kw_dry_run].val, obj_bool);
+		set_obj_bool(wk, akw[kw_dry_run].val, false);
+	}
+
+	obj install_script;
+	make_obj(wk, &install_script, obj_array);
+	obj_array_push(wk, install_script, akw[kw_dry_run].val);
+	obj_array_push(wk, install_script, ctx.arr);
+	obj_array_push(wk, wk->install_scripts, install_script);
 	return true;
 }
 
 static bool
-func_meson_add_postconf_script(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_add_postconf_script(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_exe }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_exe }, ARG_TYPE_NULL };
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
@@ -411,11 +453,11 @@ func_meson_add_postconf_script(struct workspace *wk, obj _, uint32_t args_node,
 }
 
 static bool
-func_meson_add_dist_script(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_add_dist_script(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { ARG_TYPE_GLOB | tc_exe }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { TYPE_TAG_GLOB | tc_exe }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
@@ -435,19 +477,18 @@ func_meson_add_dist_script(struct workspace *wk, obj _, uint32_t args_node, obj
 }
 
 static bool
-func_meson_get_cross_property(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_get_cross_property(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	struct args_norm ao[] = { { tc_any }, ARG_TYPE_NULL };
+	struct args_norm an[] = { { obj_string }, { tc_any, .optional = true }, ARG_TYPE_NULL };
 
-	if (!interp_args(wk, args_node, an, ao, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
-	if (ao[0].set) {
-		*res = ao[0].val;
+	if (an[1].set) {
+		*res = an[1].val;
 	} else {
-		interp_error(wk, an[0].node, "TODO: get cross property");
+		vm_error_at(wk, an[0].node, "TODO: get cross property");
 		return false;
 	}
 
@@ -455,24 +496,22 @@ func_meson_get_cross_property(struct workspace *wk, obj _, uint32_t args_node, o
 }
 
 static bool
-func_meson_get_external_property(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_get_external_property(struct workspace *wk, obj _, obj *res)
 {
-	struct args_norm an[] = { { obj_string }, ARG_TYPE_NULL };
-	struct args_norm ao[] = { { tc_any }, ARG_TYPE_NULL };
-	enum kwargs { kw_native, };
-	struct args_kw akw[] = {
-		[kw_native] = { "native", obj_bool },
-		0
+	struct args_norm an[] = { { obj_string }, { tc_any, .optional = true }, ARG_TYPE_NULL };
+	enum kwargs {
+		kw_native,
 	};
+	struct args_kw akw[] = { [kw_native] = { "native", obj_bool }, 0 };
 
-	if (!interp_args(wk, args_node, an, ao, akw)) {
+	if (!pop_args(wk, an, akw)) {
 		return false;
 	}
 
-	if (ao[0].set) {
-		*res = ao[0].val;
+	if (an[1].set) {
+		*res = an[1].val;
 	} else {
-		interp_error(wk, an[0].node, "TODO: get external property");
+		vm_error_at(wk, an[0].node, "TODO: get external property");
 		return false;
 	}
 
@@ -480,9 +519,9 @@ func_meson_get_external_property(struct workspace *wk, obj _, uint32_t args_node
 }
 
 static bool
-func_meson_can_run_host_binaries(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_can_run_host_binaries(struct workspace *wk, obj _, obj *res)
 {
-	if (!interp_args(wk, args_node, NULL, NULL, NULL)) {
+	if (!pop_args(wk, NULL, NULL)) {
 		return false;
 	}
 
@@ -492,22 +531,23 @@ func_meson_can_run_host_binaries(struct workspace *wk, obj _, uint32_t args_node
 }
 
 static bool
-func_meson_add_devenv(struct workspace *wk, obj _, uint32_t args_node, obj *res)
+func_meson_add_devenv(struct workspace *wk, obj _, obj *res)
 {
 	struct args_norm an[] = { { tc_any }, ARG_TYPE_NULL };
-	if (!interp_args(wk, args_node, an, NULL, NULL)) {
+	if (!pop_args(wk, an, NULL)) {
 		return false;
 	}
 
 	return true;
 }
 
-const struct func_impl_name impl_tbl_meson[] = {
+const struct func_impl impl_tbl_meson[] = {
 	{ "add_devenv", func_meson_add_devenv },
 	{ "add_dist_script", func_meson_add_dist_script },
 	{ "add_install_script", func_meson_add_install_script },
 	{ "add_postconf_script", func_meson_add_postconf_script },
 	{ "backend", func_meson_backend, tc_string },
+	{ "build_options", func_meson_build_options, tc_string },
 	{ "build_root", func_meson_global_build_root, tc_string },
 	{ "can_run_host_binaries", func_meson_can_run_host_binaries, tc_bool },
 	{ "current_build_dir", func_meson_current_build_dir, tc_string },
@@ -525,6 +565,7 @@ const struct func_impl_name impl_tbl_meson[] = {
 	{ "override_find_program", func_meson_override_find_program },
 	{ "project_build_root", func_meson_project_build_root, tc_string },
 	{ "project_license", func_meson_project_license, tc_string },
+	{ "project_license_files", func_meson_project_license_files, tc_string },
 	{ "project_name", func_meson_project_name, tc_string },
 	{ "project_source_root", func_meson_project_source_root, tc_string },
 	{ "project_version", func_meson_project_version, tc_string },
@@ -532,3 +573,72 @@ const struct func_impl_name impl_tbl_meson[] = {
 	{ "version", func_meson_version, tc_string },
 	{ NULL, NULL },
 };
+
+static enum iteration_result
+compiler_dict_to_str_dict_iter(struct workspace *wk, void *_ctx, obj k, obj v)
+{
+	obj_dict_set(wk, *(obj *)_ctx, make_str(wk, compiler_language_to_s(k)), v);
+	return ir_cont;
+}
+
+static obj
+compiler_dict_to_str_dict(struct workspace *wk, obj d)
+{
+	obj r;
+	make_obj(wk, &r, obj_dict);
+	obj_dict_foreach(wk, d, &r, compiler_dict_to_str_dict_iter);
+
+	return r;
+}
+
+static bool
+func_meson_project(struct workspace *wk, obj _, obj *res)
+{
+	if (!pop_args(wk, NULL, NULL)) {
+		return false;
+	}
+
+	struct project *proj = current_project(wk);
+
+	make_obj(wk, res, obj_dict);
+	obj_dict_set(wk, *res, make_str(wk, "opts"), proj->opts);
+	obj_dict_set(wk, *res, make_str(wk, "compilers"), compiler_dict_to_str_dict(wk, proj->compilers));
+	obj_dict_set(wk, *res, make_str(wk, "args"), compiler_dict_to_str_dict(wk, proj->args));
+	obj_dict_set(wk, *res, make_str(wk, "link_args"), compiler_dict_to_str_dict(wk, proj->link_args));
+	return true;
+}
+
+static bool
+func_meson_register_dependency_handler(struct workspace *wk, obj _, obj *res)
+{
+	struct args_norm an[] = {
+		{ tc_string },
+		{ tc_capture },
+		ARG_TYPE_NULL,
+	};
+
+	if (!pop_args(wk, an, NULL)) {
+		return false;
+	}
+
+	obj_dict_set(wk, wk->dependency_handlers, an[0].val, an[1].val);
+	return true;
+}
+
+static bool
+func_meson_argv0(struct workspace *wk, obj _, obj *res)
+{
+	if (!pop_args(wk, 0, 0)) {
+		return false;
+	}
+
+	*res = make_str(wk, wk->argv0);
+	return true;
+}
+
+const struct func_impl impl_tbl_meson_internal[] = {
+	{ "project", func_meson_project, tc_dict },
+	{ "register_dependency_handler", func_meson_register_dependency_handler },
+	{ "argv0", func_meson_argv0, tc_string },
+	{ NULL, NULL },
+};
