@@ -5,10 +5,9 @@
 
 #include "compat.h"
 
-#include <assert.h>
-
 #include "guess.h"
 #include "log.h"
+#include "platform/assert.h"
 
 bool
 guess_version(struct workspace *wk, const char *src, obj *res)
@@ -16,6 +15,10 @@ guess_version(struct workspace *wk, const char *src, obj *res)
 	uint32_t dots = 0, ver_len = 0, new_len, new_dots;
 	const char *p, *ver = NULL;
 
+	if (!src) {
+		return false;
+	}
+
 	for (p = src; *p; ++p) {
 		new_len = 0;
 		new_dots = 0;
