@@ -8,39 +8,20 @@
 #include <string.h>
 
 #include "backend/output.h"
+#include "embedded.h"
 #include "error.h"
-#include "lang/interpreter.h"
+#include "lang/typecheck.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "options.h"
 #include "platform/mem.h"
 #include "platform/path.h"
 
-bool
-get_obj_id(struct workspace *wk, const char *name, obj *res, uint32_t proj_id)
-{
-	uint64_t *idp;
-	struct project *proj = darr_get(&wk->projects, proj_id);
-
-	if ((idp = hash_get_str(&proj->scope, name))) {
-		*res = *idp;
-		return true;
-	} else if ((idp = hash_get_str(&wk->scope, name))) {
-		*res = *idp;
-		return true;
-	} else {
-		return false;
-	}
-}
-
 struct project *
-make_project(struct workspace *wk, uint32_t *id, const char *subproject_name,
-	const char *cwd, const char *build_dir)
+make_project(struct workspace *wk, uint32_t *id, const char *subproject_name, const char *cwd, const char *build_dir)
 {
-	*id = darr_push(&wk->projects, &(struct project){ 0 });
-	struct project *proj = darr_get(&wk->projects, *id);
-
-	hash_init_str(&proj->scope, 128);
+	*id = arr_push(&wk->projects, &(struct project){ 0 });
+	struct project *proj = arr_get(&wk->projects, *id);
 
 	make_obj(wk, &proj->args, obj_dict);
 	make_obj(wk, &proj->compilers, obj_dict);
@@ -67,112 +48,72 @@ make_project(struct workspace *wk, uint32_t *id, const char *subproject_name,
 	proj->build_dir = make_str(wk, build_dir);
 	proj->build_root = proj->build_dir;
 
+	proj->scope_stack = wk->vm.behavior.scope_stack_dup(wk, wk->vm.default_scope_stack);
+
 	return proj;
 }
 
 struct project *
 current_project(struct workspace *wk)
 {
-	return darr_get(&wk->projects, wk->cur_project);
+	return wk->projects.len ? arr_get(&wk->projects, wk->cur_project) : 0;
 }
 
 void
 workspace_init_bare(struct workspace *wk)
 {
-	*wk = (struct workspace){
-		.interp_node = interp_node,
-		.assign_variable = assign_variable,
-		.unassign_variable = unassign_variable,
-		.get_variable = get_obj_id,
-		.eval_project_file = eval_project_file,
-	};
+	*wk = (struct workspace){ 0 };
+	vm_init(wk);
+	stack_init(&wk->stack, 4096);
+	complex_types_init(wk, &wk->complex_types);
 
-#ifdef TRACY_ENABLE
 	{
+#ifdef TRACY_ENABLE
 		static bool first = true;
-
 		if (first) {
 			wk->tracy.is_master_workspace = true;
 		}
-
 		first = false;
-	}
 #endif
+	}
+}
 
-	bucket_array_init(&wk->chrs, 4096, 1);
-	bucket_array_init(&wk->objs, 1024, sizeof(struct obj_internal));
+static bool
+workspace_eval_startup_file(struct workspace *wk, const char *script)
+{
+	obj _;
+	bool ret;
+	const char *src;
 
-	const struct {
-		uint32_t item_size;
-		uint32_t bucket_size;
-	} sizes[] = {
-		[obj_number] = { sizeof(int64_t), 1024 },
-		[obj_string] = { sizeof(struct str), 1024 },
-		[obj_compiler] = { sizeof(struct obj_compiler), 4 },
-		[obj_array] = { sizeof(struct obj_array), 2048 },
-		[obj_dict] = { sizeof(struct obj_dict), 512 },
-		[obj_build_target] = { sizeof(struct obj_build_target), 16 },
-		[obj_custom_target] = { sizeof(struct obj_custom_target), 16 },
-		[obj_subproject] = { sizeof(struct obj_subproject), 16 },
-		[obj_dependency] = { sizeof(struct obj_dependency), 16 },
-		[obj_external_program] = { sizeof(struct obj_external_program), 32 },
-		[obj_run_result] = { sizeof(struct obj_run_result), 32 },
-		[obj_configuration_data] = { sizeof(struct obj_configuration_data), 16 },
-		[obj_test] = { sizeof(struct obj_test), 64 },
-		[obj_module] = { sizeof(struct obj_module), 16 },
-		[obj_install_target] = { sizeof(struct obj_install_target), 128 },
-		[obj_environment] = { sizeof(struct obj_environment), 4 },
-		[obj_include_directory] = { sizeof(struct obj_include_directory), 16 },
-		[obj_option] = { sizeof(struct obj_option), 32 },
-		[obj_generator] = { sizeof(struct obj_generator), 16 },
-		[obj_generated_list] = { sizeof(struct obj_generated_list), 16 },
-		[obj_alias_target] = { sizeof(struct obj_alias_target), 4 },
-		[obj_both_libs] = { sizeof(struct obj_both_libs), 4 },
-		[obj_typeinfo] = { sizeof(struct obj_typeinfo), 4 },
-		[obj_source_set] = { sizeof(struct obj_source_set), 4 },
-		[obj_source_configuration] = { sizeof(struct obj_source_configuration), 4 },
-	};
-
-	uint32_t i;
-	for (i = _obj_aos_start; i < obj_type_count; ++i) {
-		bucket_array_init(&wk->obj_aos[i - _obj_aos_start], sizes[i].bucket_size, sizes[i].item_size);
+	if (!(src = embedded_get(script))) {
+		LOG_E("embedded script %s not found", script);
+		return false;
 	}
 
-	obj id;
-	make_obj(wk, &id, obj_null);
-	assert(id == 0);
+	stack_push(&wk->stack, wk->vm.lang_mode, language_extended);
+	stack_push(&wk->stack, wk->vm.scope_stack, wk->vm.behavior.scope_stack_dup(wk, wk->vm.default_scope_stack));
 
-	hash_init(&wk->obj_hash, 128, sizeof(obj));
+	ret = eval(wk, &(struct source){ .src = src, .label = script, .len = strlen(src) }, eval_mode_default, &_);
+
+	stack_pop(&wk->stack, wk->vm.scope_stack);
+	stack_pop(&wk->stack, wk->vm.lang_mode);
+	return ret;
 }
 
 void
-workspace_init(struct workspace *wk)
+workspace_init_runtime(struct workspace *wk)
 {
-	workspace_init_bare(wk);
+	machine_init();
 
 	wk->argv0 = "dummy";
 	wk->build_root = "dummy";
 
-	obj id;
-	make_obj(wk, &id, obj_disabler);
-	assert(id == disabler_id);
-
 	SBUF(source_root);
-	path_cwd(wk, &source_root);
+	path_copy_cwd(wk, &source_root);
 	wk->source_root = get_cstr(wk, sbuf_into_str(wk, &source_root));
 
-	darr_init(&wk->projects, 16, sizeof(struct project));
-	darr_init(&wk->option_overrides, 32, sizeof(struct option_override));
-	darr_init(&wk->source_data, 4, sizeof(struct source_data));
-	hash_init_str(&wk->scope, 32);
-
-	make_obj(wk, &id, obj_meson);
-	hash_set_str(&wk->scope, "meson", id);
-
-	make_obj(wk, &id, obj_machine);
-	hash_set_str(&wk->scope, "host_machine", id);
-	hash_set_str(&wk->scope, "build_machine", id);
-	hash_set_str(&wk->scope, "target_machine", id);
+	arr_init(&wk->projects, 16, sizeof(struct project));
+	arr_init(&wk->option_overrides, 32, sizeof(struct option_override));
 
 	make_obj(wk, &wk->binaries, obj_dict);
 	make_obj(wk, &wk->host_machine, obj_dict);
@@ -188,63 +129,54 @@ workspace_init(struct workspace *wk)
 	make_obj(wk, &wk->find_program_overrides, obj_dict);
 	make_obj(wk, &wk->global_opts, obj_dict);
 	make_obj(wk, &wk->compiler_check_cache, obj_dict);
+	make_obj(wk, &wk->dependency_handlers, obj_dict);
+}
 
+void
+workspace_init_startup_files(struct workspace *wk)
+{
 	if (!init_global_options(wk)) {
 		UNREACHABLE;
 	}
+
+#ifdef MUON_BOOTSTRAPPED
+	const char *startup_files[] = {
+		"dependencies.meson",
+	};
+
+	for (uint32_t i = 0; i < ARRAY_LEN(startup_files); ++i) {
+		if (!workspace_eval_startup_file(wk, startup_files[i])) {
+			LOG_W("script %s failed to load", startup_files[i]);
+		}
+	}
+#endif
+}
+
+void
+workspace_init(struct workspace *wk)
+{
+	workspace_init_bare(wk);
+	workspace_init_runtime(wk);
+	workspace_init_startup_files(wk);
 }
 
 void
 workspace_destroy_bare(struct workspace *wk)
 {
-	bucket_array_destroy(&wk->chrs);
-
-	bucket_array_destroy(&wk->objs);
-
-	uint32_t i;
-	struct bucket_array *str_ba = &wk->obj_aos[obj_string - _obj_aos_start];
-	for (i = 0; i < str_ba->len; ++i) {
-		struct str *s = bucket_array_get(str_ba, i);
-		if (s->flags & str_flag_big) {
-			z_free((void *)s->s);
-		}
-	}
-
-	for (i = _obj_aos_start; i < obj_type_count; ++i) {
-		bucket_array_destroy(&wk->obj_aos[i - _obj_aos_start]);
-	}
-
-	hash_destroy(&wk->obj_hash);
+	vm_destroy(wk);
+	stack_destroy(&wk->stack);
 }
 
 void
 workspace_destroy(struct workspace *wk)
 {
-	uint32_t i;
-	struct project *proj;
-	for (i = 0; i < wk->projects.len; ++i) {
-		proj = darr_get(&wk->projects, i);
-
-		hash_destroy(&proj->scope);
-	}
-
-	for (i = 0; i < wk->source_data.len; ++i) {
-		struct source_data *sdata = darr_get(&wk->source_data, i);
-
-		source_data_destroy(sdata);
-	}
-
-	darr_destroy(&wk->projects);
-	darr_destroy(&wk->option_overrides);
-	darr_destroy(&wk->source_data);
-	hash_destroy(&wk->scope);
-
+	arr_destroy(&wk->projects);
+	arr_destroy(&wk->option_overrides);
 	workspace_destroy_bare(wk);
 }
 
 bool
-workspace_setup_paths(struct workspace *wk, const char *build, const char *argv0,
-	uint32_t argc, char *const argv[])
+workspace_setup_paths(struct workspace *wk, const char *build, const char *argv0, uint32_t argc, char *const argv[])
 {
 	SBUF(build_root);
 	path_make_absolute(wk, &build_root, build);
@@ -319,7 +251,7 @@ workspace_print_summaries(struct workspace *wk, FILE *out)
 	uint32_t i;
 	struct project *proj;
 	for (i = 0; i < wk->projects.len; ++i) {
-		proj = darr_get(&wk->projects, i);
+		proj = arr_get(&wk->projects, i);
 		if (proj->not_ok) {
 			continue;
 		}
@@ -345,7 +277,7 @@ workspace_add_regenerate_deps_iter(struct workspace *wk, void *_ctx, obj v)
 	SBUF(path);
 	const char *s = get_cstr(wk, v);
 	if (!path_is_absolute(s)) {
-		path_join(wk, &path, get_cstr(wk, current_project(wk)->cwd), s);
+		path_join(wk, &path, workspace_cwd(wk), s);
 		v = sbuf_into_str(wk, &path);
 		s = get_cstr(wk, v);
 	}
@@ -366,9 +298,18 @@ void
 workspace_add_regenerate_deps(struct workspace *wk, obj obj_or_arr)
 {
 	if (get_obj_type(wk, obj_or_arr) == obj_array) {
-		obj_array_foreach(wk, obj_or_arr, NULL,
-			workspace_add_regenerate_deps_iter);
+		obj_array_foreach(wk, obj_or_arr, NULL, workspace_add_regenerate_deps_iter);
 	} else {
 		workspace_add_regenerate_deps_iter(wk, NULL, obj_or_arr);
 	}
 }
+
+const char *
+workspace_cwd(struct workspace *wk)
+{
+	if (wk->vm.lang_mode == language_internal) {
+		return path_cwd();
+	} else {
+		return get_cstr(wk, current_project(wk)->cwd);
+	}
+}
