@@ -8,7 +8,6 @@
 
 #include <errno.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
 
 #include "buf_size.h"
@@ -25,15 +24,15 @@ const char *log_level_clr[log_level_count] = {
 
 const char *log_level_name[log_level_count] = {
 	[log_error] = "error",
-	[log_warn]  = "warning",
-	[log_info]  = "info",
+	[log_warn] = "warning",
+	[log_info] = "info",
 	[log_debug] = "debug",
 };
 
-static const char *log_level_shortname[log_level_count] = {
+const char *log_level_shortname[log_level_count] = {
 	[log_error] = "err ",
-	[log_warn]  = "warn ",
-	[log_info]  = "",
+	[log_warn] = "warn ",
+	[log_info] = "",
 	[log_debug] = "dbg ",
 };
 
@@ -43,7 +42,9 @@ static struct {
 	uint32_t filter;
 	bool initialized, clr;
 	const char *prefix;
-} log_cfg = { .level = log_info, };
+} log_cfg = {
+	.level = log_info,
+};
 
 bool
 log_should_print(enum log_level lvl)
@@ -63,30 +64,42 @@ log_set_prefix(const char *prefix)
 	log_cfg.prefix = prefix;
 }
 
+uint32_t
+log_print_prefix(enum log_level lvl, char *buf, uint32_t size)
+{
+	uint32_t len = 0;
+	assert(log_cfg.initialized);
+
+	if (log_cfg.prefix) {
+		len += snprintf(&buf[len], size - len, "%s ", log_cfg.prefix);
+	}
+
+	if (*log_level_shortname[lvl]) {
+		if (log_cfg.clr) {
+			len += snprintf(&buf[len],
+				BUF_SIZE_4k - len,
+				"\033[%sm%s\033[0m",
+				log_level_clr[lvl],
+				log_level_shortname[lvl]);
+		} else {
+			len = strlen(log_level_shortname[lvl]);
+			strncpy(buf, log_level_shortname[lvl], BUF_SIZE_4k);
+		}
+	}
+
+	return len;
+}
+
 void
 log_print(bool nl, enum log_level lvl, const char *fmt, ...)
 {
 	static char buf[BUF_SIZE_4k + 3];
 
 	if (log_should_print(lvl)) {
-		uint32_t len = 0;
+		uint32_t len = log_print_prefix(lvl, buf, BUF_SIZE_4k);
 
 		assert(log_cfg.initialized);
 
-		if (log_cfg.prefix) {
-			len += snprintf(&buf[len], BUF_SIZE_4k - len, "%s ", log_cfg.prefix);
-		}
-
-		if (*log_level_shortname[lvl]) {
-			if (log_cfg.clr) {
-				len += snprintf(&buf[len], BUF_SIZE_4k - len, "\033[%sm%s\033[0m",
-					log_level_clr[lvl], log_level_shortname[lvl]);
-			} else {
-				len = strlen(log_level_shortname[lvl]);
-				strncpy(buf, log_level_shortname[lvl], BUF_SIZE_4k);
-			}
-		}
-
 		va_list ap;
 		va_start(ap, fmt);
 		len += vsnprintf(&buf[len], BUF_SIZE_4k - len, fmt, ap);
