@@ -5,12 +5,12 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <string.h>
 
 #include "error.h"
 #include "formats/ini.h"
 #include "machine_file.h"
+#include "platform/assert.h"
 #include "platform/filesystem.h"
 #include "platform/mem.h"
 
@@ -55,33 +55,40 @@ struct machine_file_parse_ctx {
 };
 
 static bool
-machine_file_parse_cb(void *_ctx, struct source *src, const char *_sect,
-	const char *k, const char *v, uint32_t line)
+machine_file_parse_cb(void *_ctx,
+	struct source *src,
+	const char *_sect,
+	const char *k,
+	const char *v,
+	struct source_location location)
 {
 	struct machine_file_parse_ctx *ctx = _ctx;
 	enum machine_file_section sect;
 
-	if (!mfile_lookup(_sect, machine_file_section_names,
-		machine_file_section_count, &sect)) {
+	if (!mfile_lookup(_sect, machine_file_section_names, machine_file_section_count, (uint32_t *)&sect)) {
 		if (!k) {
-			error_messagef(src, line, 1, log_error, "invalid section '%s'", _sect);
+			error_messagef(src, location, log_error, "invalid section '%s'", _sect);
 		}
 		return false;
 	} else if (!k) {
-		hash_clear(&current_project(ctx->wk)->scope);
+		/* TODO: hash_clear(&current_project(ctx->wk)->scope); */
 		return true;
 	}
 
 	if (!_sect) {
-		error_messagef(src, line, 1, log_error, "key not under any section");
+		error_messagef(src, location, log_error, "key not under any section");
 		return false;
 	}
 
-	struct source val_src = { .label = k, .src = v, .len = strlen(v), };
+	struct source val_src = {
+		.label = k,
+		.src = v,
+		.len = strlen(v),
+	};
 
 	obj res;
 	if (!eval(ctx->wk, &val_src, eval_mode_default, &res)) {
-		error_messagef(src, line, 1, log_error, "failed to parse value");
+		error_messagef(src, location, log_error, "failed to parse value");
 		return false;
 	}
 
@@ -89,9 +96,9 @@ machine_file_parse_cb(void *_ctx, struct source *src, const char *_sect,
 	/* obj_to_s(ctx->wk, res, buf, 2048); */
 
 	if (sect == mfile_section_constants) {
-		hash_set_str(&ctx->wk->scope, k, res);
+		/* TODO: obj_dict_set(ctx->wk, ctx->wk->global_scope, make_str(ctx->wk, k), res); */
 	} else {
-		hash_set_str(&current_project(ctx->wk)->scope, k, res);
+		/* TODO: hash_set_str(&current_project(ctx->wk)->scope, k, res); */
 
 		obj cloned, objkey, dest_dict;
 		if (!obj_clone(ctx->wk, ctx->dest_wk, res, &cloned)) {
@@ -106,12 +113,8 @@ machine_file_parse_cb(void *_ctx, struct source *src, const char *_sect,
 
 			dest_dict = ctx->dest_wk->binaries;
 			break;
-		case mfile_section_host_machine:
-			dest_dict = ctx->dest_wk->host_machine;
-			break;
-		default:
-			assert(false && "todo");
-			return false;
+		case mfile_section_host_machine: dest_dict = ctx->dest_wk->host_machine; break;
+		default: assert(false && "todo"); return false;
 		}
 
 		obj_dict_set(ctx->dest_wk, dest_dict, objkey, cloned);
@@ -128,12 +131,15 @@ machine_file_parse(struct workspace *dest_wk, const char *path)
 	struct workspace wk;
 	workspace_init(&wk);
 
-	wk.lang_mode = language_internal;
+	wk.vm.lang_mode = language_internal;
 
 	uint32_t proj_id;
 	make_project(&wk, &proj_id, "dummy", wk.source_root, wk.build_root);
 
-	struct machine_file_parse_ctx ctx = { .wk = &wk, .dest_wk = dest_wk, };
+	struct machine_file_parse_ctx ctx = {
+		.wk = &wk,
+		.dest_wk = dest_wk,
+	};
 
 	struct source src;
 	char *buf = NULL;
@@ -149,5 +155,4 @@ ret:
 	fs_source_destroy(&src);
 	workspace_destroy(&wk);
 	return ret;
-
 }
