@@ -12,6 +12,7 @@
 
 #include "args.h"
 #include "backend/backend.h"
+#include "backend/output.h"
 #include "cmd_install.h"
 #include "cmd_test.h"
 #include "embedded.h"
@@ -19,15 +20,16 @@
 #include "external/libcurl.h"
 #include "external/libpkgconf.h"
 #include "external/samurai.h"
-#include "functions/common.h"
 #include "lang/analyze.h"
+#include "lang/compiler.h"
 #include "lang/fmt.h"
-#include "lang/interpreter.h"
+#include "lang/func_lookup.h"
 #include "lang/serial.h"
 #include "machine_file.h"
 #include "meson_opts.h"
 #include "options.h"
 #include "opts.h"
+#include "platform/init.h"
 #include "platform/mem.h"
 #include "platform/path.h"
 #include "platform/run_cmd.h"
@@ -38,7 +40,7 @@
 static bool
 ensure_in_build_dir(void)
 {
-	if (!fs_dir_exists("muon-private")) {
+	if (!fs_dir_exists(output_path.private_dir)) {
 		LOG_E("this subcommand must be run from a build directory");
 		return false;
 	}
@@ -77,28 +79,25 @@ cmd_exe(uint32_t argc, uint32_t argi, char *const argv[])
 		const char *environment;
 		const char *args;
 		const char *const *cmd;
+		const char *remove_before_running;
 	} opts = { 0 };
 
-	OPTSTART("f:c:e:a:") {
-		case 'f':
-			opts.feed = optarg;
-			break;
-		case 'c':
-			opts.capture = optarg;
-			break;
-		case 'e':
-			opts.environment = optarg;
-			break;
-		case 'a':
-			opts.args = optarg;
-			break;
-	} OPTEND(argv[argi],
+	OPTSTART("f:c:e:a:R:") {
+	case 'f': opts.feed = optarg; break;
+	case 'c': opts.capture = optarg; break;
+	case 'e': opts.environment = optarg; break;
+	case 'a': opts.args = optarg; break;
+	case 'R': opts.remove_before_running = optarg; break;
+	}
+	OPTEND(argv[argi],
 		" <cmd> [arg1[ arg2[...]]]",
 		"  -f <file> - feed file to input\n"
 		"  -c <file> - capture output to file\n"
 		"  -e <file> - load environment from data file\n"
-		"  -a <file> - load arguments from data file\n",
-		NULL, -1)
+		"  -a <file> - load arguments from data file\n"
+		"  -R <file> - remove file if it exists before executing the command\n",
+		NULL,
+		-1)
 
 	if (argi >= argc && !opts.args) {
 		LOG_E("missing command");
@@ -111,6 +110,12 @@ cmd_exe(uint32_t argc, uint32_t argi, char *const argv[])
 	opts.cmd = (const char *const *)&argv[argi];
 	++argi;
 
+	if (opts.remove_before_running && fs_exists(opts.remove_before_running)) {
+		if (!fs_remove(opts.remove_before_running)) {
+			return false;
+		}
+	}
+
 	bool ret = false;
 	struct run_cmd_ctx ctx = { 0 };
 	ctx.stdin_path = opts.feed;
@@ -120,8 +125,7 @@ cmd_exe(uint32_t argc, uint32_t argi, char *const argv[])
 	}
 
 	struct workspace wk;
-	bool initialized_workspace = false,
-	     allocated_argv = false;
+	bool initialized_workspace = false, allocated_argv = false;
 
 	const char *envstr = NULL;
 	uint32_t envc = 0;
@@ -189,113 +193,142 @@ cmd_check(uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct {
 		const char *filename;
-		bool print_ast;
-		uint32_t parse_mode;
+		bool print_ast, print_dis;
+		enum vm_compile_mode compile_mode;
 	} opts = { 0 };
 
-	OPTSTART("pP") {
-		case 'p':
-			opts.print_ast = true;
-			break;
-		case 'P':
-			opts.print_ast = true;
-			opts.parse_mode = pm_keep_formatting;
-			break;
-	} OPTEND(argv[argi],
+	OPTSTART("pdm:") {
+	case 'p': opts.print_ast = true; break;
+	case 'd': opts.print_dis = true; break;
+	case 'm': {
+		const char *p, *chars = "xf";
+		for (p = optarg; *p; ++p) {
+			if (!strchr(chars, *p)) {
+				LOG_E("invalid mode '%c', must be one of %s", *p, chars);
+				return false;
+			}
+
+			switch (*p) {
+			case 'x': opts.compile_mode |= vm_compile_mode_language_extended; break;
+			case 'f': opts.compile_mode |= vm_compile_mode_fmt; break;
+			default: UNREACHABLE;
+			}
+		}
+		break;
+	}
+	}
+	OPTEND(argv[argi],
 		" <filename>",
 		"  -p - print parsed ast\n"
-		"  -P - print parsed formatting ast\n",
-		NULL, 1)
+		"  -d - print dissasembly\n"
+		"  -m <mode> - parse with parse mode <mode>\n",
+		NULL,
+		1)
 
 	opts.filename = argv[argi];
 
 	bool ret = false;
 
-	struct source src = { 0 };
-	struct ast ast = { 0 };
-	struct source_data sdata = { 0 };
+	struct workspace wk;
+	workspace_init_bare(&wk);
 
-	if (!fs_read_entire_file(opts.filename, &src)) {
-		goto ret;
-	}
+	arr_push(&wk.vm.src, &(struct source){ 0 });
+	struct source *src = arr_get(&wk.vm.src, 0);
 
-	if (!parser_parse(NULL, &ast, &sdata, &src, opts.parse_mode)) {
+	if (!fs_read_entire_file(opts.filename, src)) {
 		goto ret;
 	}
 
 	if (opts.print_ast) {
-		print_ast(&ast);
+		struct node *n;
+		if (!(n = parse(&wk, src, opts.compile_mode))) {
+			goto ret;
+		}
+
+		if (opts.compile_mode & vm_compile_mode_fmt) {
+			print_fmt_ast(&wk, n);
+		} else {
+			print_ast(&wk, n);
+		}
+	} else {
+		uint32_t _entry;
+		if (!vm_compile(&wk, src, opts.compile_mode, &_entry)) {
+			goto ret;
+		}
+
+		if (opts.print_dis) {
+			vm_dis(&wk);
+		}
 	}
 
 	ret = true;
 ret:
-	fs_source_destroy(&src);
-	ast_destroy(&ast);
-	source_data_destroy(&sdata);
+	fs_source_destroy(src);
+	workspace_destroy(&wk);
 	return ret;
 }
 
 static bool
 cmd_analyze(uint32_t argc, uint32_t argi, char *const argv[])
 {
-	struct analyze_opts opts = {
+	struct az_opts opts = {
 		.replay_opts = error_diagnostic_store_replay_include_sources,
-		.enabled_diagnostics = analyze_diagnostic_unused_variable
-				       | analyze_diagnostic_dead_code,
+		.enabled_diagnostics = az_diagnostic_unused_variable | az_diagnostic_dead_code
+				       | az_diagnostic_redirect_script_error,
 	};
 
-	OPTSTART("luqO:W:i:") {
-		case 'i':
-			opts.internal_file = optarg;
-			break;
-		case 'l':
-			opts.subdir_error = true;
-			opts.replay_opts &= ~error_diagnostic_store_replay_include_sources;
-			break;
-		case 'O':
-			opts.file_override = optarg;
-			break;
-		case 'q':
-			opts.replay_opts |= error_diagnostic_store_replay_errors_only;
-			break;
-		case 'W': {
-			bool enable = true;
-			const char *name = optarg;
-			if (str_startswith(&WKSTR(optarg), &WKSTR("no-"))) {
-				enable = false;
-				name += 3;
-			}
-
-			if (strcmp(name, "list") == 0) {
-				analyze_print_diagnostic_names();
-				return true;
-			} else if (strcmp(name, "error") == 0) {
-				opts.replay_opts |= error_diagnostic_store_replay_werror;
-			} else {
-				enum analyze_diagnostic d;
-				if (!analyze_diagnostic_name_to_enum(name, &d)) {
-					LOG_E("invalid diagnostic name '%s'", name);
-					return false;
-				}
-
-				if (enable) {
-					opts.enabled_diagnostics |= d;
-				} else {
-					opts.enabled_diagnostics &= ~d;
-				}
-			}
-			break;
+	OPTSTART("luqO:W:i:td:") {
+	case 'i': opts.internal_file = optarg; break;
+	case 'l':
+		opts.subdir_error = true;
+		opts.replay_opts &= ~error_diagnostic_store_replay_include_sources;
+		break;
+	case 'O': opts.file_override = optarg; break;
+	case 'q': opts.replay_opts |= error_diagnostic_store_replay_errors_only; break;
+	case 't': opts.eval_trace = true; break;
+	case 'd': opts.get_definition_for = optarg; break;
+	case 'W': {
+		bool enable = true;
+		const char *name = optarg;
+		if (str_startswith(&WKSTR(optarg), &WKSTR("no-"))) {
+			enable = false;
+			name += 3;
 		}
-	} OPTEND(argv[argi], "",
+
+		if (strcmp(name, "list") == 0) {
+			az_print_diagnostic_names();
+			return true;
+		} else if (strcmp(name, "error") == 0) {
+			opts.replay_opts |= error_diagnostic_store_replay_werror;
+		} else {
+			enum az_diagnostic d;
+			if (!az_diagnostic_name_to_enum(name, &d)) {
+				LOG_E("invalid diagnostic name '%s'", name);
+				return false;
+			}
+
+			if (enable) {
+				opts.enabled_diagnostics |= d;
+			} else {
+				opts.enabled_diagnostics &= ~d;
+			}
+		}
+		break;
+	}
+	}
+	OPTEND(argv[argi],
+		"",
 		"  -l - optimize output for editor linter plugins\n"
 		"  -q - only report errors\n"
 		"  -O <path> - read project file with matching path from stdin\n"
 		"  -i <path> - analyze the single file <path> in internal mode\n"
+		"  -t - print a tree of all meson source files that are evaluated\n"
+		"  -d <var> - print the location of the definition of <var>\n"
 		"  -W [no-]<diagnostic> - enable or disable diagnostics\n"
 		"  -W list - list available diagnostics\n"
-		"  -W error - turn all warnings into errors\n"
-		,
-		NULL, 0)
+		"  -W error - turn all warnings into errors\n",
+		NULL,
+		0)
 
 	if (opts.internal_file && opts.file_override) {
 		LOG_E("-i and -O are mutually exclusive");
@@ -322,17 +355,15 @@ cmd_options(uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct list_options_opts opts = { 0 };
 	OPTSTART("am") {
-		case 'a':
-			opts.list_all = true;
-			break;
-		case 'm':
-			opts.only_modified = true;
-			break;
-	} OPTEND(argv[argi], "",
+	case 'a': opts.list_all = true; break;
+	case 'm': opts.only_modified = true; break;
+	}
+	OPTEND(argv[argi],
+		"",
 		"  -a - list all options\n"
-		"  -m - list only modified options\n"
-		,
-		NULL, 0)
+		"  -m - list only modified options\n",
+		NULL,
+		0)
 
 	return list_options(&opts);
 }
@@ -341,15 +372,19 @@ static bool
 cmd_summary(uint32_t argc, uint32_t argi, char *const argv[])
 {
 	OPTSTART("") {
-	} OPTEND(argv[argi], "", "", NULL, 0)
+	}
+	OPTEND(argv[argi], "", "", NULL, 0)
 
 	if (!ensure_in_build_dir()) {
 		return false;
 	}
 
+	SBUF_manual(path);
+	path_join(0, &path, output_path.private_dir, output_path.summary);
+
 	bool ret = false;
 	struct source src = { 0 };
-	if (!fs_read_entire_file("muon-private/summary.txt", &src)) {
+	if (!fs_read_entire_file(path.buf, &src)) {
 		goto ret;
 	}
 
@@ -357,6 +392,7 @@ cmd_summary(uint32_t argc, uint32_t argi, char *const argv[])
 
 	ret = true;
 ret:
+	sbuf_destroy(&path);
 	fs_source_destroy(&src);
 	return ret;
 }
@@ -373,9 +409,11 @@ cmd_info(uint32_t argc, uint32_t argi, char *const argv[])
 	};
 
 	OPTSTART("") {
-	} OPTEND(argv[argi], "", "", commands, -1);
+	}
+	OPTEND(argv[argi], "", "", commands, -1);
 
-	cmd_func cmd = NULL;;
+	cmd_func cmd = NULL;
+	;
 	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
 		return false;
 	}
@@ -394,10 +432,8 @@ cmd_subprojects_check_wrap(uint32_t argc, uint32_t argi, char *const argv[])
 	} opts = { 0 };
 
 	OPTSTART("") {
-	} OPTEND(argv[argi],
-		" <filename>",
-		"",
-		NULL, 1)
+	}
+	OPTEND(argv[argi], " <filename>", "", NULL, 1)
 
 	opts.filename = argv[argi];
 
@@ -453,7 +489,8 @@ cmd_subprojects_download(uint32_t argc, uint32_t argi, char *const argv[])
 	bool res = false;
 
 	OPTSTART("") {
-	} OPTEND(argv[argi], " <list of subprojects>", "", NULL, -1)
+	}
+	OPTEND(argv[argi], " <list of subprojects>", "", NULL, -1)
 
 	SBUF_manual(path);
 	path_make_absolute(NULL, &path, cmd_subprojects_subprojects_dir);
@@ -501,12 +538,9 @@ cmd_subprojects(uint32_t argc, uint32_t argi, char *const argv[])
 	};
 
 	OPTSTART("d:") {
-		case 'd':
-			cmd_subprojects_subprojects_dir = optarg;
-			break;
-	} OPTEND(argv[0], "",
-		"  -d <directory> - use an alternative subprojects directory\n",
-		commands, -1)
+	case 'd': cmd_subprojects_subprojects_dir = optarg; break;
+	}
+	OPTEND(argv[0], "", "  -d <directory> - use an alternative subprojects directory\n", commands, -1)
 
 	cmd_func cmd;
 	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
@@ -515,19 +549,47 @@ cmd_subprojects(uint32_t argc, uint32_t argi, char *const argv[])
 
 	return cmd(argc, argi, argv);
 }
-
 static bool
-eval_internal(const char *filename, bool embedded, const char *argv0, char *const argv[], uint32_t argc)
+cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
 {
+	struct workspace wk;
+	workspace_init_bare(&wk);
+
+	const char *filename;
+	bool embedded = false;
+
+	OPTSTART("esb:") {
+	case 'e': embedded = true; break;
+	case 's': {
+		wk.vm.disable_fuzz_unsafe_functions = true;
+		break;
+	}
+	case 'b': {
+		vm_dbg_push_breakpoint(&wk, optarg);
+		break;
+	}
+	}
+	OPTEND(argv[argi],
+		" <filename> [args]",
+		"  -e - lookup <filename> as an embedded script\n"
+		"  -s - disable functions that are unsafe to be called at random\n",
+		NULL,
+		-1)
+
+	if (argi >= argc) {
+		LOG_E("missing required filename argument");
+		return false;
+	}
+
+	filename = argv[argi];
+
 	bool ret = false;
 
 	struct source src = { 0 };
-	bool src_allocd = false;
 
-	struct workspace wk;
-	workspace_init(&wk);
+	machine_init();
 
-	wk.lang_mode = language_internal;
+	wk.vm.lang_mode = language_internal;
 
 	if (embedded) {
 		if (!(src.src = embedded_get(filename))) {
@@ -541,19 +603,15 @@ eval_internal(const char *filename, bool embedded, const char *argv0, char *cons
 		if (!fs_read_entire_file(filename, &src)) {
 			goto ret;
 		}
-		src_allocd = true;
 	}
 
-	uint32_t proj_id;
-	make_project(&wk, &proj_id, "dummy", wk.source_root, wk.build_root);
-
 	{ // populate argv array
 		obj argv_obj;
 		make_obj(&wk, &argv_obj, obj_array);
-		hash_set_str(&wk.scope, "argv", argv_obj);
+		wk.vm.behavior.assign_variable(&wk, "argv", argv_obj, 0, assign_local);
 
 		uint32_t i;
-		for (i = 0; i < argc; ++i) {
+		for (i = argi; i < argc; ++i) {
 			obj_array_push(&wk, argv_obj, make_str(&wk, argv[i]));
 		}
 	}
@@ -565,47 +623,16 @@ eval_internal(const char *filename, bool embedded, const char *argv0, char *cons
 
 	ret = true;
 ret:
-	if (src_allocd) {
-		fs_source_destroy(&src);
-	}
 	workspace_destroy(&wk);
 	return ret;
 }
 
-static bool
-cmd_eval(uint32_t argc, uint32_t argi, char *const argv[])
-{
-	const char *filename;
-	bool embedded = false;
-
-	OPTSTART("es") {
-		case 'e':
-			embedded = true;
-			break;
-		case 's':
-			disable_fuzz_unsafe_functions = true;
-			break;
-	} OPTEND(argv[argi], " <filename> [args]",
-		"  -e - lookup <filename> as an embedded script\n"
-		"  -s - disable functions that are unsafe to be called at random\n",
-		NULL, -1)
-
-	if (argi >= argc) {
-		LOG_E("missing required filename argument");
-		return false;
-	}
-
-	filename = argv[argi];
-
-	return eval_internal(filename, embedded, argv[0], &argv[argi], argc - argi);
-}
-
 static bool
 cmd_repl(uint32_t argc, uint32_t argi, char *const argv[])
 {
 	struct workspace wk;
 	workspace_init(&wk);
-	wk.lang_mode = language_internal;
+	wk.vm.lang_mode = language_internal;
 
 	obj id;
 	make_project(&wk, &id, "dummy", wk.source_root, wk.build_root);
@@ -620,7 +647,8 @@ static bool
 cmd_dump_signatures(uint32_t argc, uint32_t argi, char *const argv[])
 {
 	OPTSTART("") {
-	} OPTEND(argv[argi], "", "", NULL, 0)
+	}
+	OPTEND(argv[argi], "", "", NULL, 0)
 
 	struct workspace wk;
 	workspace_init(&wk);
@@ -649,9 +677,11 @@ cmd_internal(uint32_t argc, uint32_t argi, char *const argv[])
 	};
 
 	OPTSTART("") {
-	} OPTEND(argv[argi], "", "", commands, -1);
+	}
+	OPTEND(argv[argi], "", "", commands, -1);
 
-	cmd_func cmd = NULL;;
+	cmd_func cmd = NULL;
+	;
 	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
 		return false;
 	}
@@ -663,7 +693,7 @@ cmd_internal(uint32_t argc, uint32_t argi, char *const argv[])
 static bool
 cmd_samu(uint32_t argc, uint32_t argi, char *const argv[])
 {
-	return muon_samu(argc - argi, (char **)&argv[argi]);
+	return samu_main(argc - argi, (char **)&argv[argi], 0);
 }
 
 static bool
@@ -676,59 +706,62 @@ cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
 		test_opts.print_summary = true;
 	}
 
-	OPTSTART("s:d:Sfj:lvRe:") {
-		case 'l':
-			test_opts.list = true;
-			break;
-		case 'e':
-			test_opts.setup = optarg;
-			break;
-		case 's':
-			if (test_opts.suites_len > MAX_CMDLINE_TEST_SUITES) {
-				LOG_E("too many -s options (max: %d)", MAX_CMDLINE_TEST_SUITES);
-				return false;
-			}
-			test_opts.suites[test_opts.suites_len] = optarg;
-			++test_opts.suites_len;
-			break;
-		case 'd':
-			if (strcmp(optarg, "auto") == 0) {
-				test_opts.display = test_display_auto;
-			} else if (strcmp(optarg, "dots") == 0) {
-				test_opts.display = test_display_dots;
-			} else if (strcmp(optarg, "bar") == 0) {
-				test_opts.display = test_display_bar;
-			} else {
-				LOG_E("invalid progress display mode '%s'", optarg);
-				return false;
-			}
-			break;
-		case 'f':
-			test_opts.fail_fast = true;
-			break;
-		case 'S':
-			test_opts.print_summary = true;
-			break;
-		case 'j': {
-			char *endptr;
-			unsigned long n = strtoul(optarg, &endptr, 10);
-
-			if (n > UINT32_MAX || !*optarg || *endptr) {
-				LOG_E("invalid number of jobs: %s", optarg);
-				return false;
-			}
-
-			test_opts.jobs = n;
-			break;
+	OPTSTART("s:d:Sfj:lvRe:o:") {
+	case 'l': test_opts.list = true; break;
+	case 'e': test_opts.setup = optarg; break;
+	case 's':
+		if (test_opts.suites_len > MAX_CMDLINE_TEST_SUITES) {
+			LOG_E("too many -s options (max: %d)", MAX_CMDLINE_TEST_SUITES);
+			return false;
 		}
-		case 'v':
-			++test_opts.verbosity;
-			break;
-		case 'R':
-			test_opts.no_rebuild = true;
-			break;
-	} OPTEND(argv[argi], "",
+		test_opts.suites[test_opts.suites_len] = optarg;
+		++test_opts.suites_len;
+		break;
+	case 'd':
+		if (strcmp(optarg, "auto") == 0) {
+			test_opts.display = test_display_auto;
+		} else if (strcmp(optarg, "dots") == 0) {
+			test_opts.display = test_display_dots;
+		} else if (strcmp(optarg, "bar") == 0) {
+			test_opts.display = test_display_bar;
+		} else {
+			LOG_E("invalid progress display mode '%s'", optarg);
+			return false;
+		}
+		break;
+	case 'o':
+		if (strcmp(optarg, "term") == 0) {
+			test_opts.output = test_output_term;
+		} else if (strcmp(optarg, "html") == 0) {
+			test_opts.output = test_output_html;
+		} else if (strcmp(optarg, "json") == 0) {
+			test_opts.output = test_output_json;
+		} else {
+			LOG_E("invalid progress display mode '%s'", optarg);
+			return false;
+		}
+		break;
+	case 'f': test_opts.fail_fast = true; break;
+	case 'S': test_opts.print_summary = true; break;
+	case 'j': {
+		char *endptr;
+		unsigned long n = strtoul(optarg, &endptr, 10);
+
+		if (n > UINT32_MAX || !*optarg || *endptr) {
+			LOG_E("invalid number of jobs: %s", optarg);
+			return false;
+		}
+
+		test_opts.jobs = n;
+		break;
+	}
+	case 'v': ++test_opts.verbosity; break;
+	case 'R': test_opts.no_rebuild = true; break;
+	}
+	OPTEND(argv[argi],
+		" [test [test [...]]]",
 		"  -d <mode> - change progress display mode (auto|dots|bar)\n"
+		"  -o <mode> - set output mode (term|html|json)\n"
 		"  -e <setup> - use test setup <setup>\n"
 		"  -f - fail fast; exit after first failure\n"
 		"  -j <jobs> - set the number of test workers\n"
@@ -737,27 +770,36 @@ cmd_test(uint32_t argc, uint32_t argi, char *const argv[])
 		"  -S - print a summary with elapsed time\n"
 		"  -s <suite> - only run items in <suite>, may be passed multiple times\n"
 		"  -v - increase verbosity, may be passed twice\n",
-		NULL, 0)
+		NULL,
+		-1)
 
 	if (!ensure_in_build_dir()) {
 		return false;
 	}
 
+	test_opts.tests = &argv[argi];
+	test_opts.tests_len = argc - argi;
+
 	return tests_run(&test_opts, argv[0]);
 }
 
 static bool
 cmd_install(uint32_t argc, uint32_t argi, char *const argv[])
 {
-	struct install_options opts = { 0 };
+	struct install_options opts = {
+		.destdir = getenv("DESTDIR"),
+	};
 
-	OPTSTART("n") {
-		case 'n':
-			opts.dry_run = true;
-			break;
-	} OPTEND(argv[argi], "",
-		"  -n - dry run\n",
-		NULL, 0)
+	OPTSTART("nd:") {
+	case 'n': opts.dry_run = true; break;
+	case 'd': opts.destdir = optarg; break;
+	}
+	OPTEND(argv[argi],
+		"",
+		"  -n - dry run\n"
+		"  -d <destdir> - set destdir\n",
+		NULL,
+		0)
 
 	if (!ensure_in_build_dir()) {
 		return false;
@@ -772,47 +814,51 @@ cmd_setup(uint32_t argc, uint32_t argi, char *const argv[])
 	TracyCZoneAutoS;
 	bool res = false;
 	struct workspace wk;
-	workspace_init(&wk);
+	workspace_init_bare(&wk);
+	workspace_init_runtime(&wk);
 
 	uint32_t original_argi = argi + 1;
 
-	OPTSTART("D:c:b") {
-		case 'D':
-			if (!parse_and_set_cmdline_option(&wk, optarg)) {
-				goto ret;
-			}
-			break;
-		case 'c': {
-			FILE *f;
-			if (!(f = fs_fopen(optarg, "rb"))) {
-				goto ret;
-			} else if (!serial_load(&wk, &wk.compiler_check_cache, f)) {
-				LOG_E("failed to load compiler check cache");
-				goto ret;
-			} else if (!fs_fclose(f)) {
-				goto ret;
-			}
-			break;
+	OPTSTART("D:c:b:") {
+	case 'D':
+		if (!parse_and_set_cmdline_option(&wk, optarg)) {
+			goto ret;
 		}
-		case 'b':
-			wk.dbg.break_on_err = true;
-			break;
-	} OPTEND(argv[argi],
+		break;
+	case 'c': {
+		FILE *f;
+		if (!(f = fs_fopen(optarg, "rb:"))) {
+			goto ret;
+		} else if (!serial_load(&wk, &wk.compiler_check_cache, f)) {
+			LOG_E("failed to load compiler check cache");
+			goto ret;
+		} else if (!fs_fclose(f)) {
+			goto ret;
+		}
+		break;
+	}
+	case 'b': {
+		vm_dbg_push_breakpoint(&wk, optarg);
+		break;
+	}
+	}
+	OPTEND(argv[argi],
 		" <build dir>",
 		"  -D <option>=<value> - set project options\n"
 		"  -c <compiler_check_cache.dat> - path to compiler check cache dump\n"
-		"  -b - break on errors\n",
-		NULL, 1)
+		"  -b <breakpoint> - set breakpoint\n",
+		NULL,
+		1)
 
 	const char *build = argv[argi];
 	++argi;
 
-	if (!workspace_setup_paths(&wk, build, argv[0],
-		argc - original_argi,
-		&argv[original_argi])) {
+	if (!workspace_setup_paths(&wk, build, argv[0], argc - original_argi, &argv[original_argi])) {
 		goto ret;
 	}
 
+	workspace_init_startup_files(&wk);
+
 	uint32_t project_id;
 	if (!eval_project(&wk, NULL, wk.source_root, wk.build_root, &project_id)) {
 		goto ret;
@@ -845,24 +891,23 @@ cmd_format(uint32_t argc, uint32_t argi, char *const argv[])
 	struct {
 		char *const *filenames;
 		const char *cfg_path;
-		bool in_place, check_only;
+		bool in_place, check_only, editorconfig;
 	} opts = { 0 };
 
-	OPTSTART("ic:q") {
-		case 'i':
-			opts.in_place = true;
-			break;
-		case 'c':
-			opts.cfg_path = optarg;
-			break;
-		case 'q':
-			opts.check_only = true;
-			break;
-	} OPTEND(argv[argi], " <file>[ <file>[...]]",
+	OPTSTART("ic:qe") {
+	case 'i': opts.in_place = true; break;
+	case 'c': opts.cfg_path = optarg; break;
+	case 'q': opts.check_only = true; break;
+	case 'e': opts.editorconfig = true; break;
+	}
+	OPTEND(argv[argi],
+		" <file>[ <file>[...]]",
 		"  -q - exit with 1 if files would be modified by muon fmt\n"
 		"  -i - format files in-place\n"
-		"  -c <muon_fmt.ini> - read configuration from muon_fmt.ini\n",
-		NULL, -1)
+		"  -c <muon_fmt.ini> - read configuration from muon_fmt.ini\n"
+		"  -e - try to read configuration from .editorconfig\n",
+		NULL,
+		-1)
 
 	if (opts.in_place && opts.check_only) {
 		LOG_E("-q and -i are mutually exclusive");
@@ -898,7 +943,7 @@ cmd_format(uint32_t argc, uint32_t argi, char *const argv[])
 			out = stdout;
 		}
 
-		fmt_ret = fmt(&src, out, opts.cfg_path, opts.check_only);
+		fmt_ret = fmt(&src, out, opts.cfg_path, opts.check_only, opts.editorconfig);
 cont:
 		if (opened_out) {
 			fs_fclose(out);
@@ -921,8 +966,7 @@ cmd_version(uint32_t argc, uint32_t argi, char *const argv[])
 		muon_version.version,
 		*muon_version.vcs_tag ? "-" : "",
 		muon_version.vcs_tag,
-		muon_version.meson_compat
-		);
+		muon_version.meson_compat);
 
 	const struct {
 		const char *name;
@@ -1001,26 +1045,27 @@ cmd_main(uint32_t argc, uint32_t argi, char *argv[])
 	SBUF_manual(argv0);
 
 	OPTSTART("vC:") {
-		case 'v':
-			log_set_lvl(log_debug);
-			break;
-		case 'C': {
-			// fix argv0 here since if it is a relative path it will be
-			// wrong after chdir
-			if (!path_is_basename(argv[0])) {
-				path_make_absolute(NULL, &argv0, argv[0]);
-				argv[0] = argv0.buf;
-			}
-
-			if (!path_chdir(optarg)) {
-				return false;
-			}
-			break;
+	case 'v': log_set_lvl(log_debug); break;
+	case 'C': {
+		// fix argv0 here since if it is a relative path it will be
+		// wrong after chdir
+		if (!path_is_basename(argv[0])) {
+			path_make_absolute(NULL, &argv0, argv[0]);
+			argv[0] = argv0.buf;
 		}
-	} OPTEND(argv[0], "",
+
+		if (!path_chdir(optarg)) {
+			return false;
+		}
+		break;
+	}
+	}
+	OPTEND(argv[0],
+		"",
 		"  -v - turn on debug messages\n"
 		"  -C <path> - chdir to path\n",
-		commands, -1)
+		commands,
+		-1)
 
 	cmd_func cmd;
 	if (!find_cmd(commands, &cmd, argc, argi, argv, false)) {
@@ -1037,6 +1082,8 @@ ret:
 int
 main(int argc, char *argv[])
 {
+	platform_init();
+
 	log_init();
 	log_set_lvl(log_info);
 
@@ -1050,7 +1097,7 @@ main(int argc, char *argv[])
 	{
 		SBUF(basename);
 		path_basename(NULL, &basename, argv[0]);
-		meson_compat = strcmp(basename.buf, "meson") == 0 && strcmp(argv[1], "internal") != 0;
+		meson_compat = strcmp(basename.buf, "meson") == 0 && (argc < 2 || strcmp(argv[1], "internal") != 0);
 		sbuf_destroy(&basename);
 	}
 
