@@ -12,9 +12,9 @@
 #include "backend/output.h"
 #include "embedded.h"
 #include "error.h"
-#include "lang/interpreter.h"
+#include "lang/object_iterators.h"
 #include "lang/serial.h"
-#include "lang/workspace.h"
+#include "lang/typecheck.h"
 #include "log.h"
 #include "options.h"
 #include "platform/filesystem.h"
@@ -31,18 +31,18 @@ static const char *build_option_type_to_s[build_option_type_count] = {
 	[op_feature] = "feature",
 };
 
-static bool set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val,
-	enum option_value_source source, bool coerce);
+static bool
+set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum option_value_source source, bool coerce);
 
 static bool
-parse_config_string(struct workspace *wk, const struct str *ss, struct option_override *oo)
+parse_config_string(struct workspace *wk, const struct str *ss, struct option_override *oo, bool key_only)
 {
 	if (str_has_null(ss)) {
 		LOG_E("option cannot contain NUL");
 		return false;
 	}
 
-	struct str subproject = { 0 }, key = { 0 }, val, cur = { 0 };
+	struct str subproject = { 0 }, key = { 0 }, val = { 0 }, cur = { 0 };
 
 	cur.s = ss->s;
 	bool reading_key = true, have_subproject = false;
@@ -73,18 +73,32 @@ parse_config_string(struct workspace *wk, const struct str *ss, struct option_ov
 		++cur.len;
 	}
 
-	val = cur;
+	if (!reading_key) {
+		val = cur;
+	} else {
+		key = cur;
+	}
 
-	if (!key.len) {
-		LOG_E("expected '=' in option '%s'", ss->s);
-		return false;
-	} else if (have_subproject && !subproject.len) {
+	if (have_subproject && !subproject.len) {
 		LOG_E("missing subproject in option '%s'", ss->s);
 		return false;
 	}
 
+	if (!key.len) {
+		LOG_E("missing key in option '%s'", ss->s);
+		return false;
+	} else if (key_only && val.len) {
+		LOG_E("unexpected '=' in option '%s'", ss->s);
+		return false;
+	} else if (!key_only && !val.len) {
+		LOG_E("expected '=' in option '%s'", ss->s);
+		return false;
+	}
+
 	oo->name = make_strn(wk, key.s, key.len);
-	oo->val = make_strn(wk, val.s, val.len);
+	if (!key_only) {
+		oo->val = make_strn(wk, val.s, val.len);
+	}
 	if (have_subproject) {
 		oo->proj = make_strn(wk, subproject.s, subproject.len);
 	}
@@ -123,7 +137,7 @@ check_invalid_subproject_option(struct workspace *wk)
 	bool found, ret = true;
 
 	for (i = 0; i < wk->option_overrides.len; ++i) {
-		oo = darr_get(&wk->option_overrides, i);
+		oo = arr_get(&wk->option_overrides, i);
 		if (!oo->proj || oo->source < option_value_source_commandline) {
 			continue;
 		}
@@ -131,7 +145,7 @@ check_invalid_subproject_option(struct workspace *wk)
 		found = false;
 
 		for (j = 1; j < wk->projects.len; ++j) {
-			proj = darr_get(&wk->projects, j);
+			proj = arr_get(&wk->projects, j);
 			if (proj->not_ok) {
 				continue;
 			}
@@ -153,26 +167,8 @@ check_invalid_subproject_option(struct workspace *wk)
 	return ret;
 }
 
-struct check_array_opt_ctx {
-	obj choices;
-	uint32_t node;
-};
-
-static enum iteration_result
-check_array_opt_iter(struct workspace *wk, void *_ctx, obj val)
-{
-	struct check_array_opt_ctx *ctx = _ctx;
-
-	if (!obj_array_in(wk, ctx->choices, val)) {
-		interp_error(wk, ctx->node, "array element %o is not one of %o", val, ctx->choices);
-		return ir_err;
-	}
-
-	return ir_cont;
-}
-
 static bool
-coerce_feature_opt(struct workspace *wk, uint32_t node, const struct str *val, obj *res)
+coerce_feature_opt(struct workspace *wk, const struct str *val, obj *res)
 {
 	enum feature_opt_state f;
 
@@ -183,7 +179,7 @@ coerce_feature_opt(struct workspace *wk, uint32_t node, const struct str *val, o
 	} else if (str_eql(val, &WKSTR("disabled"))) {
 		f = feature_opt_disabled;
 	} else {
-		interp_error(wk, node, "unable to coerce '%s' into a feature", val->s);
+		vm_error(wk, "unable to coerce '%s' into a feature", val->s);
 		return false;
 	}
 
@@ -196,7 +192,6 @@ struct check_deprecated_option_ctx {
 	struct obj_option *opt;
 	obj *val;
 	obj sval;
-	uint32_t err_node;
 };
 
 static enum iteration_result
@@ -208,7 +203,7 @@ check_deprecated_option_iter(struct workspace *wk, void *_ctx, obj old, obj new)
 	case op_array: {
 		uint32_t idx;
 		if (obj_array_index_of(wk, *ctx->val, old, &idx)) {
-			interp_warning(wk, ctx->err_node, "option value %o is deprecated", old);
+			vm_warning(wk, "option value %o is deprecated", old);
 
 			if (new) {
 				obj_array_set(wk, *ctx->val, idx, new);
@@ -218,7 +213,7 @@ check_deprecated_option_iter(struct workspace *wk, void *_ctx, obj old, obj new)
 	}
 	default:
 		if (str_eql(get_str(wk, ctx->sval), get_str(wk, old))) {
-			interp_warning(wk, ctx->err_node, "option value %o is deprecated", old);
+			vm_warning(wk, "option value %o is deprecated", old);
 
 			if (new) {
 				*ctx->val = new;
@@ -230,30 +225,28 @@ check_deprecated_option_iter(struct workspace *wk, void *_ctx, obj old, obj new)
 }
 
 static bool
-check_deprecated_option(struct workspace *wk, uint32_t err_node,
-	struct obj_option *opt, obj sval, obj *val)
+check_deprecated_option(struct workspace *wk, struct obj_option *opt, obj sval, obj *val)
 {
 	struct check_deprecated_option_ctx ctx = {
 		.val = val,
 		.sval = sval,
 		.opt = opt,
-		.err_node = err_node
 	};
 
 	switch (get_obj_type(wk, opt->deprecated)) {
 	case obj_bool:
 		if (get_obj_bool(wk, opt->deprecated)) {
-			interp_warning(wk, err_node, "option %o is deprecated", ctx.opt->name);
+			vm_warning(wk, "option %o is deprecated", ctx.opt->name);
 		}
 		break;
 	case obj_string: {
 		struct project *cur_proj = current_project(wk);
 
-		interp_warning(wk, err_node, "option %o is deprecated to %o", opt->name, opt->deprecated);
+		vm_warning(wk, "option %o is deprecated to %o", opt->name, opt->deprecated);
 
 		obj newopt;
 		if (get_option(wk, cur_proj, get_str(wk, opt->deprecated), &newopt)) {
-			set_option(wk, err_node, newopt, sval, option_value_source_deprecated_rename, true);
+			set_option(wk, 0, newopt, sval, option_value_source_deprecated_rename, true);
 		} else {
 			struct option_override oo = {
 				.proj = current_project(wk)->cfg.name,
@@ -262,23 +255,20 @@ check_deprecated_option(struct workspace *wk, uint32_t err_node,
 				.source = option_value_source_deprecated_rename,
 			};
 
-			darr_push(&wk->option_overrides, &oo);
+			arr_push(&wk->option_overrides, &oo);
 		}
 		break;
 	}
 	case obj_dict:
-	case obj_array:
-		obj_iterable_foreach(wk, opt->deprecated, &ctx, check_deprecated_option_iter);
-		break;
-	default:
-		UNREACHABLE;
+	case obj_array: obj_iterable_foreach(wk, opt->deprecated, &ctx, check_deprecated_option_iter); break;
+	default: UNREACHABLE;
 	}
 
 	return true;
 }
 
 static bool
-coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *opt, obj sval, obj *res)
+coerce_option_override(struct workspace *wk, struct obj_option *opt, obj sval, obj *res)
 {
 	const struct str *val = get_str(wk, sval);
 	*res = 0;
@@ -298,7 +288,7 @@ coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *o
 	}
 
 	if (opt->deprecated) {
-		if (!check_deprecated_option(wk, node, opt, sval, res)) {
+		if (!check_deprecated_option(wk, opt, sval, res)) {
 			return false;
 		}
 
@@ -314,9 +304,7 @@ coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *o
 
 	switch (opt->type) {
 	case op_combo:
-	case op_string:
-		*res = sval;
-		break;
+	case op_string: *res = sval; break;
 	case op_boolean: {
 		bool b;
 		if (str_eql(val, &WKSTR("true"))) {
@@ -324,7 +312,7 @@ coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *o
 		} else if (str_eql(val, &WKSTR("false"))) {
 			b = false;
 		} else {
-			interp_error(wk, node, "unable to coerce '%s' into a boolean", val->s);
+			vm_error(wk, "unable to coerce '%s' into a boolean", val->s);
 			return false;
 		}
 
@@ -335,10 +323,10 @@ coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *o
 	case op_integer: {
 		int64_t num;
 		char *endptr;
-		num = strtol(val->s, &endptr, 10);
+		num = strtoll(val->s, &endptr, 10);
 
 		if (!val->len || *endptr) {
-			interp_error(wk, node, "unable to coerce '%s' into a number", val->s);
+			vm_error(wk, "unable to coerce '%s' into a number", val->s);
 			return false;
 		}
 
@@ -350,22 +338,20 @@ coerce_option_override(struct workspace *wk, uint32_t node, struct obj_option *o
 		// do nothing, array values were already coerced above
 		break;
 	}
-	case op_feature:
-		return coerce_feature_opt(wk, node, val, res);
-	default:
-		UNREACHABLE;
+	case op_feature: return coerce_feature_opt(wk, val, res);
+	default: UNREACHABLE;
 	}
 
 	return true;
 }
 
 static bool
-typecheck_opt(struct workspace *wk, uint32_t err_node, obj val, enum build_option_type type, obj *res)
+typecheck_opt(struct workspace *wk, uint32_t node, obj val, enum build_option_type type, obj name, obj *res)
 {
 	enum obj_type expected_type;
 
 	if (type == op_feature && get_obj_type(wk, val) == obj_string) {
-		if (!coerce_feature_opt(wk, err_node, get_str(wk, val), res)) {
+		if (!coerce_feature_opt(wk, get_str(wk, val), res)) {
 			return false;
 		}
 
@@ -379,11 +365,13 @@ typecheck_opt(struct workspace *wk, uint32_t err_node, obj val, enum build_optio
 	case op_combo: expected_type = obj_string; break;
 	case op_integer: expected_type = obj_number; break;
 	case op_array: expected_type = obj_array; break;
-	default:
-		UNREACHABLE_RETURN;
+	default: UNREACHABLE_RETURN;
 	}
 
-	if (!typecheck(wk, err_node, val, expected_type)) {
+	char msg[256];
+	snprintf(msg, sizeof(msg), "expected type %%s for option %s, got %%s", get_cstr(wk, name));
+
+	if (!typecheck_custom(wk, node, val, expected_type, msg)) {
 		return false;
 	}
 
@@ -428,14 +416,16 @@ prefix_dir_opts(struct workspace *wk)
 	obj prefix;
 	get_option_value(wk, NULL, "prefix", &prefix);
 
-	return obj_dict_foreach(wk, wk->global_opts, &(struct prefix_dir_opts_ctx) {
-		.prefix = get_str(wk, prefix),
-	}, prefix_dir_opts_iter);
+	return obj_dict_foreach(wk,
+		wk->global_opts,
+		&(struct prefix_dir_opts_ctx){
+			.prefix = get_str(wk, prefix),
+		},
+		prefix_dir_opts_iter);
 }
 
 static void
-extend_array_option(struct workspace *wk, obj opt, obj new_val,
-	enum option_value_source source)
+extend_array_option(struct workspace *wk, obj opt, obj new_val, enum option_value_source source)
 {
 	struct obj_option *o = get_obj_option(wk, opt);
 
@@ -448,8 +438,7 @@ extend_array_option(struct workspace *wk, obj opt, obj new_val,
 }
 
 static bool
-set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val,
-	enum option_value_source source, bool coerce)
+set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val, enum option_value_source source, bool coerce)
 {
 	struct obj_option *o = get_obj_option(wk, opt);
 
@@ -488,25 +477,25 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val,
 	o->source = source;
 
 	if (get_obj_type(wk, o->deprecated) == obj_bool && get_obj_bool(wk, o->deprecated)) {
-		interp_warning(wk, node, "option %o is deprecated", o->name);
+		vm_warning_at(wk, node, "option %o is deprecated", o->name);
 	}
 
 	if (coerce) {
 		obj coerced;
-		if (!coerce_option_override(wk, node, o, new_val, &coerced)) {
+		if (!coerce_option_override(wk, o, new_val, &coerced)) {
 			return false;
 		}
 		new_val = coerced;
 	}
 
-	if (!typecheck_opt(wk, node, new_val, o->type, &new_val)) {
+	if (!typecheck_opt(wk, node, new_val, o->type, o->name, &new_val)) {
 		return false;
 	}
 
 	switch (o->type) {
 	case op_combo: {
 		if (!obj_array_in(wk, o->choices, new_val)) {
-			interp_error(wk, node, "'%o' is not one of %o", new_val, o->choices);
+			vm_error_at(wk, node, "'%o' is not one of %o", new_val, o->choices);
 			return false;
 		}
 		break;
@@ -514,34 +503,33 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val,
 	case op_integer: {
 		int64_t num = get_obj_number(wk, new_val);
 
-		if ((o->max && num > get_obj_number(wk, o->max))
-		    || (o->min && num < get_obj_number(wk, o->min)) ) {
-			interp_error(wk, node, "value %" PRId64 " is out of range (%" PRId64 "..%" PRId64 ")",
+		if ((o->max && num > get_obj_number(wk, o->max)) || (o->min && num < get_obj_number(wk, o->min))) {
+			vm_error_at(wk,
+				node,
+				"value %" PRId64 " is out of range (%" PRId64 "..%" PRId64 ")",
 				get_obj_number(wk, new_val),
 				(o->min ? get_obj_number(wk, o->min) : INT64_MIN),
-				(o->max ? get_obj_number(wk, o->max) : INT64_MAX)
-				);
+				(o->max ? get_obj_number(wk, o->max) : INT64_MAX));
 			return false;
 		}
 		break;
 	}
 	case op_array: {
 		if (o->choices) {
-			if (!obj_array_foreach(wk, new_val, &(struct check_array_opt_ctx) {
-					.choices = o->choices,
-					.node = node,
-				}, check_array_opt_iter)) {
-				return false;
+			obj val;
+			obj_array_for(wk, new_val, val) {
+				if (!obj_array_in(wk, o->choices, val)) {
+					vm_error_at(wk, node, "array element %o is not one of %o", val, o->choices);
+					return false;
+				}
 			}
 		}
 		break;
 	}
 	case op_string:
 	case op_feature:
-	case op_boolean:
-		break;
-	default:
-		UNREACHABLE_RETURN;
+	case op_boolean: break;
+	default: UNREACHABLE_RETURN;
 	}
 
 	o->val = new_val;
@@ -549,9 +537,9 @@ set_option(struct workspace *wk, uint32_t node, obj opt, obj new_val,
 }
 
 bool
-create_option(struct workspace *wk, uint32_t node, obj opts, obj opt, obj val)
+create_option(struct workspace *wk, obj opts, obj opt, obj val)
 {
-	if (!set_option(wk, node, opt, val, option_value_source_default, false)) {
+	if (!set_option(wk, 0, opt, val, option_value_source_default, false)) {
 		return false;
 	}
 
@@ -568,12 +556,11 @@ create_option(struct workspace *wk, uint32_t node, obj opts, obj opt, obj val)
 	}
 
 	const struct str *name = get_str(wk, o->name);
-	if (str_has_null(name)
-	    || strchr(name->s, ':')) {
-		interp_error(wk, node, "invalid option name %o", o->name);
+	if (str_has_null(name) || strchr(name->s, ':')) {
+		vm_error(wk, "invalid option name %o", o->name);
 		return false;
 	} else if (get_option(wk, proj, name, &_)) {
-		interp_error(wk, node, "duplicate option %o", o->name);
+		vm_error(wk, "duplicate option %o", o->name);
 		return false;
 	}
 
@@ -582,8 +569,7 @@ create_option(struct workspace *wk, uint32_t node, obj opts, obj opt, obj val)
 }
 
 bool
-get_option_overridable(struct workspace *wk, const struct project *proj, obj overrides,
-	const struct str *name, obj *res)
+get_option_overridable(struct workspace *wk, const struct project *proj, obj overrides, const struct str *name, obj *res)
 {
 	if (overrides && obj_dict_index_strn(wk, overrides, name->s, name->len, res)) {
 		return true;
@@ -597,8 +583,7 @@ get_option_overridable(struct workspace *wk, const struct project *proj, obj ove
 }
 
 bool
-get_option(struct workspace *wk, const struct project *proj,
-	const struct str *name, obj *res)
+get_option(struct workspace *wk, const struct project *proj, const struct str *name, obj *res)
 {
 	return get_option_overridable(wk, proj, 0, name, res);
 }
@@ -649,13 +634,11 @@ set_compile_opt_from_env(struct workspace *wk, const char *name, const char *fla
 	}
 
 	if ((flags = getenv(flags)) && *flags) {
-		extend_array_option(wk, opt, str_split(wk, &WKSTR(flags), NULL),
-			option_value_source_environment);
+		extend_array_option(wk, opt, str_split(wk, &WKSTR(flags), NULL), option_value_source_environment);
 	}
 
 	if ((extra = getenv(extra)) && *extra) {
-		extend_array_option(wk, opt, str_split(wk, &WKSTR(extra), NULL),
-			option_value_source_environment);
+		extend_array_option(wk, opt, str_split(wk, &WKSTR(extra), NULL), option_value_source_environment);
 	}
 }
 
@@ -681,24 +664,24 @@ init_builtin_options(struct workspace *wk, const char *script, const char *fallb
 		opts = fallback;
 	}
 
-	enum language_mode old_mode = wk->lang_mode;
-	wk->lang_mode = language_opts;
+	enum language_mode old_mode = wk->vm.lang_mode;
+	wk->vm.lang_mode = language_opts;
 	obj _;
 	initializing_builtin_options = true;
-	bool ret = eval_str(wk, opts, eval_mode_default, &_);
+	bool ret = eval_str_label(wk, script, opts, eval_mode_default, &_);
 	initializing_builtin_options = false;
-	wk->lang_mode = old_mode;
+	wk->vm.lang_mode = old_mode;
 	return ret;
 }
 
 static bool
 init_per_project_options(struct workspace *wk)
 {
-	return init_builtin_options(wk, "per_project_options.meson",
+	return init_builtin_options(wk,
+		"per_project_options.meson",
 		"option('default_library', type: 'string', value: 'static')\n"
 		"option('warning_level', type: 'string', value: '3')\n"
-		"option('c_std', type: 'string', value: 'c99')\n"
-		);
+		"option('c_std', type: 'string', value: 'c99')\n");
 }
 
 static enum iteration_result
@@ -717,12 +700,12 @@ set_yielding_project_options_iter(struct workspace *wk, void *_ctx, obj _k, obj
 
 	po = get_obj_option(wk, parent_opt);
 	if (po->type != o->type) {
-		interp_warning(wk, 0,
+		vm_warning_at(wk,
+			0,
 			"option %o cannot yield to parent option due to a type mismatch (%s != %s)",
 			o->name,
 			build_option_type_to_s[po->type],
-			build_option_type_to_s[o->type]
-			);
+			build_option_type_to_s[o->type]);
 		return ir_cont;
 	}
 
@@ -743,43 +726,60 @@ setup_project_options(struct workspace *wk, const char *cwd)
 		return true;
 	}
 
-	SBUF(meson_opts);
-	path_join(wk, &meson_opts, cwd, "meson_options.txt");
+	const char *option_file_names[] = {
+		"meson.options",
+		"meson_options.txt",
+	};
 
-	if (fs_file_exists(meson_opts.buf)) {
-		enum language_mode old_mode = wk->lang_mode;
-		wk->lang_mode = language_opts;
-		if (!wk->eval_project_file(wk, meson_opts.buf, false)) {
+	SBUF(meson_opts);
+
+	bool exists = false;
+	uint32_t i;
+	for (i = 0; i < ARRAY_LEN(option_file_names); ++i) {
+		path_join(wk, &meson_opts, cwd, option_file_names[i]);
+
+		if (fs_file_exists(meson_opts.buf)) {
+			exists = true;
+			break;
+		}
+	}
+
+	if (exists) {
+		enum language_mode old_mode = wk->vm.lang_mode;
+		wk->vm.lang_mode = language_opts;
+		if (!wk->vm.behavior.eval_project_file(wk, meson_opts.buf, false)) {
+			wk->vm.lang_mode = old_mode;
 			return false;
 		}
-		wk->lang_mode = old_mode;
+		wk->vm.lang_mode = old_mode;
 	}
 
 	bool is_master_project = wk->cur_project == 0;
 
 	if (!is_master_project) {
-		if (!obj_dict_foreach(wk, current_project(wk)->opts,
-			darr_get(&wk->projects, 0),
-			set_yielding_project_options_iter)) {
+		if (!obj_dict_foreach(wk,
+			    current_project(wk)->opts,
+			    arr_get(&wk->projects, 0),
+			    set_yielding_project_options_iter)) {
 			return false;
 		}
 	}
 
 	bool ret = true;
-	uint32_t i;
 	struct option_override *oo;
 
 	for (i = 0; i < wk->option_overrides.len; ++i) {
-		oo = darr_get(&wk->option_overrides, i);
+		oo = arr_get(&wk->option_overrides, i);
 
-		if (!subproj_name_matches(wk, get_cstr(wk, current_project(wk)->subproject_name), get_cstr(wk, oo->proj))) {
+		if (!subproj_name_matches(
+			    wk, get_cstr(wk, current_project(wk)->subproject_name), get_cstr(wk, oo->proj))) {
 			continue;
 		}
 
 		const struct str *name = get_str(wk, oo->name);
 		obj opt;
 		if (obj_dict_index_strn(wk, current_project(wk)->opts, name->s, name->len, &opt)
-		    || (is_master_project && obj_dict_index_strn(wk, wk->global_opts, name->s, name->len, &opt))) {
+			|| (is_master_project && obj_dict_index_strn(wk, wk->global_opts, name->s, name->len, &opt))) {
 			if (!set_option(wk, 0, opt, oo->val, oo->source, !oo->obj_value)) {
 				ret = false;
 			}
@@ -797,35 +797,41 @@ setup_project_options(struct workspace *wk, const char *cwd)
 bool
 init_global_options(struct workspace *wk)
 {
-	if (!init_builtin_options(wk, "global_options.meson",
-		"option('buildtype', type: 'string', value: 'debugoptimized')\n"
-		"option('prefix', type: 'string', value: '/usr/local')\n"
-		"option('bindir', type: 'string', value: 'bin')\n"
-		"option('mandir', type: 'string', value: 'share/man')\n"
-		"option('datadir', type: 'string', value: 'share')\n"
-		"option('libdir', type: 'string', value: 'lib')\n"
-		"option('includedir', type: 'string', value: 'include')\n"
-		"option('wrap_mode', type: 'string', value: 'nopromote')\n"
-		"option('force_fallback_for', type: 'array', value: [])\n"
-		"option('pkg_config_path', type: 'string', value: '')\n"
-		"option('c_args', type: 'array', value: [])\n"
-		"option('c_link_args', type: 'array', value: [])\n"
-		"option('werror', type: 'boolean', value: false)\n"
+	if (!init_builtin_options(wk,
+		    "global_options.meson",
+		    "option('buildtype', type: 'string', value: 'debugoptimized')\n"
+		    "option('prefix', type: 'string', value: '/usr/local')\n"
+		    "option('bindir', type: 'string', value: 'bin')\n"
+		    "option('mandir', type: 'string', value: 'share/man')\n"
+		    "option('datadir', type: 'string', value: 'share')\n"
+		    "option('libdir', type: 'string', value: 'lib')\n"
+		    "option('includedir', type: 'string', value: 'include')\n"
+		    "option('wrap_mode', type: 'string', value: 'nopromote')\n"
+		    "option('force_fallback_for', type: 'array', value: [])\n"
+		    "option('pkg_config_path', type: 'string', value: '')\n"
+		    "option('c_args', type: 'array', value: [])\n"
+		    "option('c_link_args', type: 'array', value: [])\n"
+		    "option('werror', type: 'boolean', value: false)\n"
+		    "option('prefer_static', type: 'boolean', value: false)\n"
 
-		"option('env.CC', type: 'array', value: ['cc'])\n"
-		"option('env.NINJA', type: 'array', value: ['ninja'])\n"
-		)) {
+		    "option('env.CC', type: 'array', value: ['cc'])\n"
+		    "option('env.NINJA', type: 'array', value: ['ninja'])\n"
+		    "option('env.AR', type: 'array', value: ['ar'])\n"
+		    "option('env.LD', type: 'array', value: ['ld'])\n")) {
 		return false;
 	}
 
 	set_binary_from_env(wk, "CC", "env.CC");
 	set_binary_from_env(wk, "NINJA", "env.NINJA");
+	set_binary_from_env(wk, "AR", "env.AR");
+	set_binary_from_env(wk, "LD", "env.LD");
 	set_compile_opt_from_env(wk, "c_args", "CFLAGS", "CPPFLAGS");
 	set_compile_opt_from_env(wk, "c_link_args", "CFLAGS", "LDFLAGS");
 
 #ifdef MUON_BOOTSTRAPPED
 	set_binary_from_env(wk, "CXX", "env.CXX");
 	set_binary_from_env(wk, "OBJC", "env.OBJC");
+	set_binary_from_env(wk, "OBJCPP", "env.OBJCPP");
 	set_binary_from_env(wk, "NASM", "env.NASM");
 	set_compile_opt_from_env(wk, "cpp_args", "CXXFLAGS", "CPPFLAGS");
 	set_compile_opt_from_env(wk, "cpp_link_args", "CXXFLAGS", "LDFLAGS");
@@ -839,138 +845,171 @@ bool
 parse_and_set_cmdline_option(struct workspace *wk, char *lhs)
 {
 	struct option_override oo = { .source = option_value_source_commandline };
-	if (!parse_config_string(wk, &WKSTR(lhs), &oo)) {
+	if (!parse_config_string(wk, &WKSTR(lhs), &oo, false)) {
 		return false;
 	}
 
-	darr_push(&wk->option_overrides, &oo);
+	arr_push(&wk->option_overrides, &oo);
 	return true;
 }
 
-struct parse_and_set_default_options_ctx {
+enum parse_and_set_option_flag {
+	parse_and_set_option_flag_override = 1 << 0,
+	parse_and_set_option_flag_for_subproject = 1 << 1,
+	parse_and_set_option_flag_have_value = 1 << 2,
+};
+
+struct parse_and_set_option_params {
 	uint32_t node;
 	obj project_name;
-	bool for_subproject;
+	obj opt;
+	obj value;
+	obj opts;
+	enum parse_and_set_option_flag flags;
 };
 
-static enum iteration_result
-parse_and_set_default_options_iter(struct workspace *wk, void *_ctx, obj v)
+static bool
+parse_and_set_option(struct workspace *wk, const struct parse_and_set_option_params *params)
 {
-	struct parse_and_set_default_options_ctx *ctx = _ctx;
+	struct option_override oo = { 0 };
 
-	struct option_override oo = { .source = option_value_source_default_options };
-	if (!parse_config_string(wk, get_str(wk, v), &oo)) {
-		interp_error(wk, ctx->node, "invalid option string");
-		return ir_err;
-	}
-
-	bool oo_for_subproject = true;
-	if (!oo.proj) {
-		oo_for_subproject = false;
-		oo.proj = ctx->project_name;
-	}
-
-	if (ctx->for_subproject || oo_for_subproject) {
-		oo.source = option_value_source_subproject_default_options;
-		darr_push(&wk->option_overrides, &oo);
-		return ir_cont;
-	}
-
-	obj opt;
-	if (get_option(wk, current_project(wk), get_str(wk, oo.name), &opt)) {
-		if (!set_option(wk, ctx->node, opt, oo.val, option_value_source_default_options, true)) {
-			return ir_err;
-		}
+	if (params->flags & parse_and_set_option_flag_override) {
+		oo.source = option_value_source_override_options;
 	} else {
-		LLOG_E("invalid option: ");
-		log_option_override(wk, &oo);
-		log_plain("\n");
-		return ir_err;
+		oo.source = option_value_source_default_options;
 	}
 
-	return ir_cont;
-}
+	bool key_only = false;
+	if (params->flags & parse_and_set_option_flag_have_value) {
+		key_only = true;
+		oo.val = params->value;
+		oo.obj_value = true;
+	}
 
-bool
-parse_and_set_default_options(struct workspace *wk, uint32_t err_node,
-	obj arr, obj project_name, bool for_subproject)
-{
-	struct parse_and_set_default_options_ctx ctx = {
-		.node = err_node,
-		.project_name = project_name,
-		.for_subproject = for_subproject,
-	};
-
-	if (!obj_array_foreach(wk, arr, &ctx, parse_and_set_default_options_iter)) {
+	if (!parse_config_string(wk, get_str(wk, params->opt), &oo, key_only)) {
+		vm_error_at(wk, params->node, "invalid option string");
 		return false;
 	}
 
-	return true;
-}
+	if (params->flags & parse_and_set_option_flag_override) {
+		if (oo.proj) {
+			vm_error_at(wk, params->node, "subproject options may not be set in override_options");
+			return false;
+		}
+	} else {
+		bool oo_for_subproject = true;
+		if (!oo.proj) {
+			oo_for_subproject = false;
+			oo.proj = params->project_name;
+		}
 
-struct parse_and_set_override_options_ctx {
-	uint32_t node;
-	obj opts;
-};
-
-static enum iteration_result
-parse_and_set_override_options_iter(struct workspace *wk, void *_ctx, obj v)
-{
-	struct parse_and_set_override_options_ctx *ctx = _ctx;
-
-	struct option_override oo = { .source = option_value_source_default_options };
-	if (!parse_config_string(wk, get_str(wk, v), &oo)) {
-		interp_error(wk, ctx->node, "invalid option string");
-		return ir_err;
-	}
-
-	if (oo.proj) {
-		interp_error(wk, ctx->node, "subproject options may not be set in override_options");
-		return ir_err;
+		if ((params->flags & parse_and_set_option_flag_for_subproject) || oo_for_subproject) {
+			oo.source = option_value_source_subproject_default_options;
+			arr_push(&wk->option_overrides, &oo);
+			return true;
+		}
 	}
 
 	obj opt;
 	if (!get_option(wk, current_project(wk), get_str(wk, oo.name), &opt)) {
-		interp_error(wk, ctx->node, "invalid option %o in override_options", oo.name);
-		return ir_err;
-	}
-
-	obj newopt;
-	make_obj(wk, &newopt, obj_option);
-	struct obj_option *o = get_obj_option(wk, newopt);
-	*o = *get_obj_option(wk, opt);
-
-	if (!set_option(wk, ctx->node, newopt, oo.val, option_value_source_override_options, true)) {
-		return ir_err;
-	}
-
-	if (obj_dict_in(wk, ctx->opts, o->name)) {
-		interp_error(wk, ctx->node, "duplicate option %o in override_options", oo.name);
-		return ir_err;
-	}
-
-	obj_dict_set(wk, ctx->opts, o->name, newopt);
-	return ir_cont;
-}
-
-bool
-parse_and_set_override_options(struct workspace *wk, uint32_t err_node,
-	obj arr, obj *res)
-{
-	struct parse_and_set_override_options_ctx ctx = {
-		.node = err_node,
-	};
-
-	make_obj(wk, &ctx.opts, obj_dict);
-
-	if (!obj_array_foreach(wk, arr, &ctx, parse_and_set_override_options_iter)) {
+		vm_error_at(wk, params->node, "invalid option %o", oo.name);
 		return false;
 	}
 
-	*res = ctx.opts;
+	if (params->flags & parse_and_set_option_flag_override) {
+		obj newopt;
+		make_obj(wk, &newopt, obj_option);
+		struct obj_option *o = get_obj_option(wk, newopt);
+		*o = *get_obj_option(wk, opt);
+		opt = newopt;
+	}
+
+	if (!set_option(wk, params->node, opt, oo.val, oo.source, !oo.obj_value)) {
+		return false;
+	}
+
+	if (params->flags & parse_and_set_option_flag_override) {
+		if (obj_dict_in(wk, params->opts, oo.name)) {
+			vm_error_at(wk, params->node, "duplicate option %o in override_options", oo.name);
+			return false;
+		}
+
+		obj_dict_set(wk, params->opts, oo.name, opt);
+	}
+
 	return true;
 }
 
+static bool
+parse_and_set_options(struct workspace *wk, struct parse_and_set_option_params *params, obj opts)
+{
+	enum obj_type t = get_obj_type(wk, opts);
+
+	switch (t) {
+	case obj_dict: {
+		params->flags |= parse_and_set_option_flag_have_value;
+
+		obj k, v;
+		obj_dict_for(wk, opts, k, v) {
+			params->opt = k;
+			params->value = v;
+			if (!parse_and_set_option(wk, params)) {
+				return false;
+			}
+		}
+		break;
+	}
+	case obj_string:
+		params->opt = opts;
+		if (!parse_and_set_option(wk, params)) {
+			return false;
+		}
+		break;
+	case obj_array: {
+		obj v;
+		obj_array_for(wk, opts, v) {
+			params->opt = v;
+			if (!parse_and_set_option(wk, params)) {
+				return false;
+			}
+		}
+		break;
+	}
+	default: UNREACHABLE;
+	}
+
+	return true;
+}
+
+bool
+parse_and_set_default_options(struct workspace *wk, uint32_t err_node, obj opts, obj project_name, bool for_subproject)
+{
+	struct parse_and_set_option_params params = {
+		.node = err_node,
+		.project_name = project_name,
+	};
+
+	if (for_subproject) {
+		params.flags |= parse_and_set_option_flag_for_subproject;
+	}
+
+	return parse_and_set_options(wk, &params, opts);
+}
+
+bool
+parse_and_set_override_options(struct workspace *wk, uint32_t err_node, obj opts, obj *res)
+{
+	struct parse_and_set_option_params params = {
+		.node = err_node,
+		.flags = parse_and_set_option_flag_override,
+	};
+
+	make_obj(wk, &params.opts, obj_dict);
+	*res = params.opts;
+
+	return parse_and_set_options(wk, &params, opts);
+}
+
 /* helper functions */
 
 enum wrap_mode
@@ -1048,12 +1087,12 @@ make_option_choices_iter(struct workspace *wk, void *_ctx, obj val)
 		clr = ctx->sel_clr;
 	}
 
-	str_app(wk, ctx->res, clr);
-	str_appn(wk, ctx->res, ss->s, ss->len);
-	str_app(wk, ctx->res, ctx->no_clr);
+	str_app(wk, &ctx->res, clr);
+	str_appn(wk, &ctx->res, ss->s, ss->len);
+	str_app(wk, &ctx->res, ctx->no_clr);
 
 	if (ctx->i < ctx->len - 1) {
-		str_app(wk, ctx->res, "|");
+		str_app(wk, &ctx->res, "|");
 	}
 
 	++ctx->i;
@@ -1110,9 +1149,7 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 	} else {
 		make_obj(wk, &choices, obj_array);
 		switch (opt->type) {
-		case op_string:
-			obj_array_push(wk, choices, make_str(wk, "string"));
-			break;
+		case op_string: obj_array_push(wk, choices, make_str(wk, "string")); break;
 		case op_boolean:
 			obj_array_push(wk, choices, make_str(wk, "true"));
 			obj_array_push(wk, choices, make_str(wk, "false"));
@@ -1124,18 +1161,19 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 			obj_array_push(wk, choices, make_str(wk, "disabled"));
 			obj_array_push(wk, choices, make_str(wk, "auto"));
 			make_obj(wk, &selected, obj_array);
-			obj_array_push(wk, selected, make_str(wk, (char *[]){
-				[feature_opt_enabled] = "enabled",
-				[feature_opt_disabled] = "disabled",
-				[feature_opt_auto] = "auto",
-			}[get_obj_feature_opt(wk, opt->val)]));
+			obj_array_push(wk,
+				selected,
+				make_str(wk,
+					(char *[]){
+						[feature_opt_enabled] = "enabled",
+						[feature_opt_disabled] = "disabled",
+						[feature_opt_auto] = "auto",
+					}[get_obj_feature_opt(wk, opt->val)]));
 			break;
 		case op_combo:
 		case op_array:
-		case op_integer:
-			break;
-		default:
-			UNREACHABLE;
+		case op_integer: break;
+		default: UNREACHABLE;
 		}
 	}
 
@@ -1156,13 +1194,16 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 	switch (opt->type) {
 	case op_boolean:
 	case op_combo:
-	case op_feature:
-		obj_fprintf(wk, log_file(), "<%s>", get_cstr(wk, choices));
-		break;
+	case op_feature: obj_fprintf(wk, log_file(), "<%s>", get_cstr(wk, choices)); break;
 	case op_string: {
 		const struct str *def = get_str(wk, opt->val);
-		obj_fprintf(wk, log_file(), "<%s>, default: %s%s%s", get_cstr(wk, choices),
-			sel_clr, def->len ? def->s : "''", no_clr);
+		obj_fprintf(wk,
+			log_file(),
+			"<%s>, default: %s%s%s",
+			get_cstr(wk, choices),
+			sel_clr,
+			def->len ? def->s : "''",
+			no_clr);
 		break;
 	}
 	case op_integer:
@@ -1186,8 +1227,7 @@ list_options_iter(struct workspace *wk, void *_ctx, obj key, obj val)
 			obj_fprintf(wk, log_file(), " where value in %s", get_cstr(wk, choices));
 		}
 		break;
-	default:
-		UNREACHABLE;
+	default: UNREACHABLE;
 	}
 
 	if (opt->source != option_value_source_default) {
@@ -1208,10 +1248,10 @@ list_options(const struct list_options_opts *list_opts)
 	bool ret = false;
 	struct workspace wk = { 0 };
 	workspace_init(&wk);
-	wk.lang_mode = language_opts;
+	wk.vm.lang_mode = language_opts;
 
-	darr_push(&wk.projects, &(struct project){ 0 });
-	struct project *proj = darr_get(&wk.projects, 0);
+	arr_push(&wk.projects, &(struct project){ 0 });
+	struct project *proj = arr_get(&wk.projects, 0);
 	make_obj(&wk, &proj->opts, obj_dict);
 
 	if (fs_file_exists("meson.build")) {
@@ -1219,7 +1259,7 @@ list_options(const struct list_options_opts *list_opts)
 		path_make_absolute(&wk, &meson_opts, "meson_options.txt");
 
 		if (fs_file_exists(meson_opts.buf)) {
-			if (!wk.eval_project_file(&wk, meson_opts.buf, false)) {
+			if (!wk.vm.behavior.eval_project_file(&wk, meson_opts.buf, false)) {
 				goto ret;
 			}
 		}
