@@ -5,20 +5,20 @@
 
 #include "compat.h"
 
-#include <assert.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "error.h"
+#include "platform/assert.h"
 #include "platform/mem.h"
 #include "tracy.h"
 
 #ifdef TRACY_ENABLE
 #include <sys/resource.h>
 
-#define PlotRSS  \
-	struct rusage usage; \
+#define PlotRSS                         \
+	struct rusage usage;            \
 	getrusage(RUSAGE_SELF, &usage); \
 	TracyCPlot("maxrss", ((double)usage.ru_maxrss / 1024.0));
 #else
@@ -62,13 +62,13 @@ z_realloc(void *ptr, size_t size)
 {
 	assert(size);
 	void *ret;
+	TracyCFree(ptr);
 	ret = realloc(ptr, size);
 
 	if (!ret) {
 		error_unrecoverable("realloc failed: %s", strerror(errno));
 	}
 
-	TracyCFree(ptr);
 	TracyCAlloc(ret, size);
 	PlotRSS;
 	return ret;
@@ -78,7 +78,13 @@ void
 z_free(void *ptr)
 {
 	assert(ptr);
-	free(ptr);
 	TracyCFree(ptr);
+	free(ptr);
 	PlotRSS;
 }
+
+uint32_t
+bswap_32(uint32_t x)
+{
+	return x >> 24 | (x >> 8 & 0xff00) | (x << 8 & 0xff0000) | x << 24;
+}
