@@ -5,15 +5,14 @@
 
 #include "compat.h"
 
-#include <string.h>
 #include <stdlib.h>
+#include <string.h>
 
 #include "error.h"
 #include "external/libarchive.h"
 #include "external/libcurl.h"
 #include "formats/ini.h"
 #include "lang/eval.h"
-#include "lang/interpreter.h"
 #include "lang/workspace.h"
 #include "log.h"
 #include "platform/filesystem.h"
@@ -66,24 +65,28 @@ lookup_wrap_str(const char *s, const char *strs[], uint32_t len, uint32_t *res)
 
 struct wrap_parse_ctx {
 	struct wrap wrap;
-	uint32_t field_lines[wrap_fields_count];
+	struct source_location wrap_field_source_location[wrap_fields_count];
 	enum wrap_type section;
 	bool have_type;
 };
 
 static bool
-wrap_parse_cb(void *_ctx, struct source *src, const char *sect,
-	const char *k, const char *v, uint32_t line)
+wrap_parse_cb(void *_ctx,
+	struct source *src,
+	const char *sect,
+	const char *k,
+	const char *v,
+	struct source_location location)
 {
 	uint32_t res;
 	struct wrap_parse_ctx *ctx = _ctx;
 
 	if (!sect) {
-		error_messagef(src, line, 1, log_error, "key not under wrap section");
+		error_messagef(src, location, log_error, "key not under wrap section");
 		return false;
 	} else if (!k) {
 		if (!lookup_wrap_str(sect, wrap_type_section_header, wrap_type_count, &res)) {
-			error_messagef(src, line, 1, log_error, "invalid section '%s'", sect);
+			error_messagef(src, location, log_error, "invalid section '%s'", sect);
 			return false;
 		}
 
@@ -95,7 +98,7 @@ wrap_parse_cb(void *_ctx, struct source *src, const char *sect,
 		}
 
 		if (ctx->have_type) {
-			error_messagef(src, line, 1, log_error, "conflicting wrap types");
+			error_messagef(src, location, log_error, "conflicting wrap types");
 			return false;
 		}
 
@@ -109,15 +112,15 @@ wrap_parse_cb(void *_ctx, struct source *src, const char *sect,
 	assert(k && v);
 
 	if (!lookup_wrap_str(k, wrap_field_names, wrap_fields_count, &res)) {
-		error_messagef(src, line, 1, log_error, "invalid key \"%s\"", k);
+		error_messagef(src, location, log_error, "invalid key \"%s\"", k);
 		return false;
 	} else if (ctx->wrap.fields[res]) {
-		error_messagef(src, line, 1, log_error, "duplicate key \"%s\"", k);
+		error_messagef(src, location, log_error, "duplicate key \"%s\"", k);
 		return false;
 	}
 
 	ctx->wrap.fields[res] = v;
-	ctx->field_lines[res] = line;
+	ctx->wrap_field_source_location[res] = location;
 	return true;
 }
 
@@ -130,18 +133,24 @@ struct wrap_parse_provides_ctx {
 	enum wrap_type section;
 	obj add_provides_tgt;
 	struct source *src;
-	uint32_t line;
+	struct source_location location;
 };
 
 static void
-wrap_check_provide_duplication(struct workspace *wk,
-	struct wrap_parse_provides_ctx *ctx,
-	obj provides, obj key, obj val)
+wrap_check_provide_duplication(struct workspace *wk, struct wrap_parse_provides_ctx *ctx, obj provides, obj key, obj val)
 {
 	obj oldval;
 	if (obj_dict_index(wk, provides, key, &oldval)) {
-		error_messagef(ctx->src, ctx->line, 1, log_warn,
-			"previous provide for %o from %o, is being overridden by %o", key, oldval, val);
+		static char buf[1024];
+		obj_snprintf(wk,
+			buf,
+			ARRAY_LEN(buf),
+			"previous provide for %o from %o, is being overridden by %o",
+			key,
+			oldval,
+			val);
+
+		error_message(ctx->src, ctx->location, log_warn, buf);
 	}
 }
 
@@ -157,12 +166,16 @@ wrap_parse_provides_cb_add_provides_iter(struct workspace *wk, void *_ctx, obj v
 }
 
 static bool
-wrap_parse_provides_cb(void *_ctx, struct source *src, const char *sect,
-	const char *k, const char *v, uint32_t line)
+wrap_parse_provides_cb(void *_ctx,
+	struct source *src,
+	const char *sect,
+	const char *k,
+	const char *v,
+	struct source_location location)
 {
 	struct wrap_parse_provides_ctx *ctx = _ctx;
 	ctx->src = src;
-	ctx->line = line;
+	ctx->location = location;
 
 	if (!sect) {
 		UNREACHABLE_RETURN;
@@ -170,7 +183,7 @@ wrap_parse_provides_cb(void *_ctx, struct source *src, const char *sect,
 
 	if (!k) {
 		enum wrap_type res;
-		if (!lookup_wrap_str(sect, wrap_type_section_header, wrap_type_count, &res)) {
+		if (!lookup_wrap_str(sect, wrap_type_section_header, wrap_type_count, (uint32_t *)&res)) {
 			UNREACHABLE_RETURN;
 		}
 
@@ -183,10 +196,10 @@ wrap_parse_provides_cb(void *_ctx, struct source *src, const char *sect,
 	}
 
 	if (!*k) {
-		error_messagef(src, line, 1, log_error, "empty provides key \"%s\"", k);
+		error_messagef(src, location, log_error, "empty provides key \"%s\"", k);
 		return false;
 	} else if (!*v) {
-		error_messagef(src, line, 1, log_error, "empty provides value \"%s\"", v);
+		error_messagef(src, location, log_error, "empty provides value \"%s\"", v);
 		return false;
 	}
 
@@ -199,8 +212,9 @@ wrap_parse_provides_cb(void *_ctx, struct source *src, const char *sect,
 
 	if (ctx->add_provides_tgt) {
 		if (!obj_array_foreach(ctx->wk,
-			str_split_strip(ctx->wk, &WKSTR(v), &WKSTR(","), NULL),
-			ctx, wrap_parse_provides_cb_add_provides_iter)) {
+			    str_split_strip(ctx->wk, &WKSTR(v), &WKSTR(","), NULL),
+			    ctx,
+			    wrap_parse_provides_cb_add_provides_iter)) {
 			return false;
 		}
 	} else {
@@ -246,8 +260,7 @@ checksum(const uint8_t *file_buf, uint64_t len, const char *sha256)
 }
 
 static bool
-checksum_extract(const char *buf, size_t len, const char *sha256,
-	const char *dest_dir)
+checksum_extract(const char *buf, size_t len, const char *sha256, const char *dest_dir)
 {
 	if (sha256 && !checksum((const uint8_t *)buf, len, sha256)) {
 		return false;
@@ -259,8 +272,7 @@ checksum_extract(const char *buf, size_t len, const char *sha256,
 }
 
 static bool
-fetch_checksum_extract(const char *src, const char *dest, const char *sha256,
-	const char *dest_dir)
+fetch_checksum_extract(const char *src, const char *dest, const char *sha256, const char *dest_dir)
 {
 	bool res = false;
 	uint8_t *dlbuf = NULL;
@@ -284,8 +296,11 @@ ret:
 }
 
 static bool
-wrap_download_or_check_packagefiles(const char *filename, const char *url,
-	const char *hash, const char *subprojects, const char *dest_dir,
+wrap_download_or_check_packagefiles(const char *filename,
+	const char *url,
+	const char *hash,
+	const char *subprojects,
+	const char *dest_dir,
 	bool download)
 {
 	bool res = false;
@@ -347,16 +362,12 @@ validate_wrap(struct wrap_parse_ctx *ctx, const char *file)
 	uint32_t i;
 
 	enum req { invalid, required, optional };
-	enum req field_req[wrap_fields_count] = {
-		[wf_directory] = optional,
+	enum req field_req[wrap_fields_count] = { [wf_directory] = optional,
 		[wf_patch_directory] = optional,
 		[wf_diff_files] = optional,
-		[wf_wrapdb_version] = optional
-	};
+		[wf_wrapdb_version] = optional };
 
-	if (ctx->wrap.fields[wf_patch_url]
-	    || ctx->wrap.fields[wf_patch_filename]
-	    || ctx->wrap.fields[wf_patch_hash]) {
+	if (ctx->wrap.fields[wf_patch_url] || ctx->wrap.fields[wf_patch_filename] || ctx->wrap.fields[wf_patch_hash]) {
 		field_req[wf_patch_url] = optional;
 		field_req[wf_patch_filename] = required;
 		field_req[wf_patch_hash] = optional;
@@ -382,26 +393,28 @@ validate_wrap(struct wrap_parse_ctx *ctx, const char *file)
 		field_req[wf_clone_recursive] = optional;
 		field_req[wf_push_url] = optional;
 		break;
-	default:
-		assert(false && "unreachable");
-		return false;
+	default: assert(false && "unreachable"); return false;
 	}
 
 	bool valid = true;
 
 	for (i = 0; i < wrap_fields_count; ++i) {
 		switch (field_req[i]) {
-		case optional:
-			break;
+		case optional: break;
 		case required:
 			if (!ctx->wrap.fields[i]) {
-				error_messagef(&ctx->wrap.src, 1, 1, log_error, "missing field '%s'", wrap_field_names[i]);
+				error_messagef(&ctx->wrap.src,
+					(struct source_location){ 1, 1 },
+					log_error,
+					"missing field '%s'",
+					wrap_field_names[i]);
 				valid = false;
 			}
 			break;
 		case invalid:
 			if (ctx->wrap.fields[i]) {
-				error_messagef(&ctx->wrap.src, ctx->field_lines[i], 1, log_error, "invalid field");
+				error_messagef(
+					&ctx->wrap.src, ctx->wrap_field_source_location[i], log_error, "invalid field");
 				valid = false;
 			}
 			break;
@@ -441,15 +454,12 @@ wrap_parse(const char *wrap_file, struct wrap *wrap)
 
 	*wrap = ctx.wrap;
 
-	sbuf_init(&wrap->dest_dir, wrap->dest_dir_buf,
-		ARRAY_LEN(wrap->dest_dir_buf), sbuf_flag_overflow_alloc);
-	sbuf_init(&wrap->name, wrap->name_buf,
-		ARRAY_LEN(wrap->name_buf), sbuf_flag_overflow_alloc);
+	sbuf_init(&wrap->dest_dir, wrap->dest_dir_buf, ARRAY_LEN(wrap->dest_dir_buf), sbuf_flag_overflow_alloc);
+	sbuf_init(&wrap->name, wrap->name_buf, ARRAY_LEN(wrap->name_buf), sbuf_flag_overflow_alloc);
 
 	path_basename(NULL, &wrap->name, wrap_file);
 
-	const struct str name = { .s = wrap->name.buf, .len = wrap->name.len },
-			 ext = WKSTR(".wrap");
+	const struct str name = { .s = wrap->name.buf, .len = wrap->name.len }, ext = WKSTR(".wrap");
 
 	assert(str_endswith(&name, &ext));
 
@@ -475,7 +485,7 @@ ret:
 }
 
 static bool
-wrap_run_cmd(const char *cmd, const char *const argv[], const char *chdir, const char *feed)
+wrap_run_cmd(const char *const argv[], const char *chdir, const char *feed)
 {
 	struct run_cmd_ctx cmd_ctx = {
 		.flags = run_cmd_ctx_flag_dont_capture,
@@ -518,17 +528,17 @@ wrap_apply_diff_files(struct wrap *wrap, const char *subprojects, const char *de
 			L("applying diff_file '%s'", diff_path.buf);
 
 			if (patch_cmd_found) {
-				if (!wrap_run_cmd("patch", (const char *const[]) {
-					"patch", "-p1",
-					NULL
-				}, wrap->dest_dir.buf, diff_path.buf)) {
+				if (!wrap_run_cmd((const char *const[]){ "patch", "-p1", NULL },
+					    wrap->dest_dir.buf,
+					    diff_path.buf)) {
 					goto ret;
 				}
 			} else {
-				if (!wrap_run_cmd("git", (const char *const[]) {
-					"git", "--work-tree", ".", "apply", "-p1", diff_path.buf,
-					NULL
-				}, wrap->dest_dir.buf, NULL)) {
+				if (!wrap_run_cmd(
+					    (const char *const[]){
+						    "git", "--work-tree", ".", "apply", "-p1", diff_path.buf, NULL },
+					    wrap->dest_dir.buf,
+					    NULL)) {
 					goto ret;
 				}
 			}
@@ -572,14 +582,12 @@ wrap_apply_patch(struct wrap *wrap, const char *subprojects, bool download)
 	}
 
 	if (filename) {
-		if (!wrap_download_or_check_packagefiles(
-			filename,
-			wrap->fields[wf_patch_url],
-			wrap->fields[wf_patch_hash],
-			subprojects,
-			dest_dir,
-			download
-			)) {
+		if (!wrap_download_or_check_packagefiles(filename,
+			    wrap->fields[wf_patch_url],
+			    wrap->fields[wf_patch_hash],
+			    subprojects,
+			    dest_dir,
+			    download)) {
 			return false;
 		}
 	}
@@ -596,17 +604,17 @@ wrap_apply_patch(struct wrap *wrap, const char *subprojects, bool download)
 static bool
 wrap_handle_git(struct wrap *wrap, const char *subprojects)
 {
-	if (!wrap_run_cmd("git", (const char *const[]) {
-		"git", "clone", wrap->fields[wf_url], wrap->dest_dir.buf,
-		NULL
-	}, NULL, NULL)) {
+	if (!wrap_run_cmd((const char *const[]){ "git", "clone", wrap->fields[wf_url], wrap->dest_dir.buf, NULL },
+		    NULL,
+		    NULL)) {
 		return false;
 	}
 
-	if (!wrap_run_cmd("git", (const char *const[]) {
-		"git", "-c", "advice.detachedHead=false", "checkout", wrap->fields[wf_revision], "--",
-		NULL
-	}, wrap->dest_dir.buf, NULL)) {
+	if (!wrap_run_cmd(
+		    (const char *const[]){
+			    "git", "-c", "advice.detachedHead=false", "checkout", wrap->fields[wf_revision], "--", NULL },
+		    wrap->dest_dir.buf,
+		    NULL)) {
 		return false;
 	}
 
@@ -624,20 +632,16 @@ wrap_handle_file(struct wrap *wrap, const char *subprojects, bool download)
 		dest = subprojects;
 	}
 
-	if (!fs_dir_exists(dest)) {
-		if (!fs_mkdir(dest)) {
-			return false;
-		}
+	if (!fs_mkdir(dest, true)) {
+		return false;
 	}
 
-	return wrap_download_or_check_packagefiles(
-		wrap->fields[wf_source_filename],
+	return wrap_download_or_check_packagefiles(wrap->fields[wf_source_filename],
 		wrap->fields[wf_source_url],
 		wrap->fields[wf_source_hash],
 		subprojects,
 		dest,
-		download
-		);
+		download);
 }
 
 bool
@@ -673,9 +677,7 @@ wrap_handle(const char *wrap_file, const char *subprojects, struct wrap *wrap, b
 			goto ret;
 		}
 		break;
-	default:
-		assert(false && "unreachable");
-		goto ret;
+	default: assert(false && "unreachable"); goto ret;
 	}
 
 	if (!wrap_apply_patch(wrap, subprojects, download)) {
