@@ -11,7 +11,7 @@
 #include <stdio.h>
 
 static bool
-embed(const char *path, const char *basename)
+embed(const char *path, const char *embedded_name)
 {
 	int c;
 	uint32_t i = 0;
@@ -22,10 +22,11 @@ embed(const char *path, const char *basename)
 		return false;
 	}
 
-	printf("{ .name = \"%s\", .src = (char []){\n", basename);
+	printf("{ .name = \"%s\", .src = (char []){\n", embedded_name);
 
 	while ((c = fgetc(f)) != EOF) {
-		printf("0x%x, ", c);
+		// output signed char
+		printf("%hhd, ", (signed char)c);
 
 		if ((i % 14) == 0) {
 			fputs("\n", stdout);
@@ -45,9 +46,8 @@ main(int argc, char *const argv[])
 	assert(((argc - 1) & 1) == 0 && "you must pass an even number of arguments");
 
 	printf("uint32_t embedded_len = %d;\n"
-		"\n"
-		"static struct embedded_file embedded[] = {\n"
-		,
+	       "\n"
+	       "static struct embedded_file embedded[] = {\n",
 		(argc - 1) / 2);
 
 	uint32_t i;
